; editor tab setting: 8
;
;                           D                88888         //
;                           D               8     8       //
;                           D               8     8      //    EEEEE
;           PPPPPP     DDDDDD    PPPPPP      88888      //    E     E
;           P     P   D     D    P     P    8     8    //     EEEEEE
;           P     P   D     D    P     P    8     8   //      E
;           PPPPPP     DDDDDD    PPPPPP      88888   //        EEEEE
;           P                    P
;           P                    P
;           P                    P
;
;
;                              MEMORY ADDRESS
;-------------------------------------------------------------          -------
;|    EMA    |           |           |           |           |          | RUN |
;-------------------------------------------------------------          -------
;  O   O   O | O   O   O | O   O   O | O   O   O | O   O   O |             O
;            |           |           |           |           |
;            | O   O   O | O   O   O | O   O   O | O   O   O |
;            |-----------|-----------|-----------|-----------|
;            |   |   |   |   |   |   |   |   |   |   |   |   |STATE -----
;            |-----------|-----------|-----------|-----------|           \
;            |   |   |   |   |   |   |   |   |   |   |   |   |STATUS ---  \
;            |-----------|-----------|-----------|-----------|          \ _
;            |   |   |   |   |   |   |   |   |   |   |   |   | AC  ----  /  \
;            |   |   |   |   |   |   |   |   |   |   |   |   | MD  ---  |o   |
;            |   |   |   |   |   |   |   |   |   |   |   |   | MQ  ----  \_ /
;            |___|___|___|___|___|___|___|___|___|___|___|___| BUS ----/  
;
;     SW       0   1   2   3   4   5   6   7   8   9  10  11
;    |--|    |---+---+---+---+---+---+---+---+---+---+---+---|   |--+--|  |--+- - - -
;    |==|    |===|===|===|===|===|===|===|===|===|===|===|===|   |==|==|  |==|= : : :
;    |--|    |---+---+---+---+---+---+---+---+---+---+---+---|   |--+--+  |--+- - - -
;
;
;  SWITCHES and KEYS 
; -------------------
;  [s] SW		- when this switch is up, the line on the OMNIBUS called SW
;			  is high; when the switch is down, the line is low. This
;			  switch is used by special peripheral controls, for example
;			  the Bootstrap Loader.
;  [s] SWITCH REGISTER	- these 12 switches provide a means of communication between
;	(12 switches)	  operator and machine. They allow a 12-bit word to be input.
;			  When the switch is up is designates a binary 1, switch down
;			  is a binary 0. These switches are used during manual functs
;			  or under program control.
;
;
;            ADDR EXTD        START   EXAM HALT SING      DEP
;            LOAD ADDR     CLEAR CONT           STEP
;                 LOAD
;           |----+----+    +----+----+----+----+----+    +----+
;           |====|====|    |====|====|====|====|====|    |====|
;           |----|----+    +----+----+----+----+----+    +----+
;
;
;  [k] ADDR LOAD	- this key loads the contents of the Switch Register into the
;			  PC and forces Fetch to be set (no Major States while the
;			  Load Address Key is depressed).
;  [k] EXTD ADDR LOAD	- this key loads the contents of SR6-11 into the Data Field and
;			  Instruction Field registers of the Memory Extension Control.
;			  SR9-11 goes to Data Field 0-1-2. SR6-8 goes to Instruction
;			  Field 0-1-2.
;  [k] CLEAR		- this key issues an Initialize Pulse, clearing the AC, LINK,
;			  Interrupt system, and I/O Flags.
;  [k] CONT		- this key resumes the computer program by issueing a Memory
;			  Start and setting the Run flip-flop. The word stored at the
;			  address currently held by the PC is taken as the first
;			  instruction.
;  [k] EXAM		- puts the contents of the core memory at the address specified
;			  by the contents of the PC into the MB and PC. Then the
;			  contents of the PC is incremented by one to allow examination
;			  of the contents of sequential core memory addresses by
;			  repeating the operation of the EXAMine key.
;  [s] HALT		- this switch clears the Run flip-flop and causes the machine
;			  to stop at TS1 of the next Fetch cycle. This switch is also
;			  used for single instruction stepping. HALT is active when the
;			  switch is in the "down" position.
;  [s] SING STEP	- this switch clears the Run flip-flop and causes the machine
;			  to stop at TS1 of the next Fetch cycle. Thereafter, repeated
;			  depressing of the CONTinue keys steps the program one cycle
;			  at a time, so that the contents of registers can be observed
;			  in each state. SING STEP is active when the switch is in the
;			  "down" position.
;  [k] DEP		- loads the contents of the SR into the MB and core memory at
;			  the address given by the current contents of the PC. Then the
;			  contents of the PC is incremented by one. This allows storing
;			  of information in sequential memory addresses by repeated
;			  operation of the DEPosit key.
;
; There are two modes "single step", and "single instruction".  Each is set by pushing
; down the appropriate switch. IIRC, single instruction mode stops before(?) the fetch
; cycle of each instruction. (That would be after the execute cycle of the previous
; instruction.)
; Singe *step* mode stops after the fetch, after the defer, and after the execute.
; Not sure how it interacts with data break and other special cycles.
;
;
;  LAMPS
; -------
;  Memory Address  indicates the contents of the memory address that will be accessed
;		   next.
;  EMA		   indicates which Extended Memory Field is being accessed
;		   IF0-2 (in STATUS) are the three EMA bits for a fetch or a direct
;		   memory reference. DF0-2 are the three EMA bits for an indirect
;		   memory reference. The three EMA bits control the actual memory
;		   reference being performed at any given instant.  "F" and "D" cycles
;		   always IF, "E" cycle memory accesses depend on whether the previous
;		   cycle was a "D" or not.
;  RUN		   when lit means the machine's timing is enabled and capable of
;		   executing instructions.
;
;
;   _________________________________________________________________
;   |  F    D    E   IR0  IR1  IR2  MD   DATA  SW  PAUSE  BRK   BRK  |
;   |                               DIR  CONT             PROG       | STATE
;   +-==============================================================-+
;   | LNK  GT   INT   NO  ION   UM  IF0  IF1   IF2  DF0   DF1   DF2  |
;   |           BUS  INT                                             | STATUS
;   +-==============================================================-+
;   |  0    1    2    3    4    5    6    7    8    9    10     11   | AC,MD,MQ,BUS
;   ------------------------------------------------------------------
;
;   note: only "PAUSE" and "GT" are not implemented at this point.
;	  "BRK PROG" and "BRK" are implemented when the IDE disk software is added.
;
;
;
;  INDICATOR SELECTOR SWITCH (ISS)
; ====================================================================================
;
;  STATE (STE)	- indicates an individual function for each bit:
;
;   0 - Fetch	  is ON for a FETCH cycle, reading an instruction from memory
;   1 - Defer	  is ON for a DEFER cycle, reading (and possibly modifying)
;		  an indirect address from memory
;   2 - Execute	  is ON for an EXECUTE cycle, executing the requested instruction
;   3 - IR 0
;   4 - IR 1	  (instruction register) latch the opcode of the instruction being
;   5 - IR 2	  executed
;   6 - MD DIR	  indicates whether memory will be re-written with the value in MB
;		  (essentially a read), or a new value will be written from an
;		  internal register (ie, a write). Note that all PDP-8 memory cycles
;		  are read-write, as core requires the data be re-written.
;		  Implemented for ISZ, DCA, and JMS (thus ignoring that read from core
;		  requires a write operation to correct the destructive read operation
;		  of core memory). ISZ, DCA, and JMS do always a write operation, that
;		  is inherent to the type of instruction execution.
;   7 - DATA CONT signals an addition is required during a read-modify-write operation
;		  It is turned on in the (special) case when an instruction does an
;		  indirect access to the lower words at 010-017 in page-0 that exhibit
;		  pre-increment auto-index, and during ISZ execute cycles.
;		  These 2 previous mentioned condition are implemented.
;		  DATA CONT is also turned on during data break cycles to increment
;		  addresses and decrement counters (not implemented),  and maybe (???)
;		  during JMS execute cycles too (not sure about that: not implemented)
;   8 - SW	  SW reflects the state of the SW switch
;   9 - Pause	  is set when an I/O instruction is being executed. It is reset
;		  automatically at the end of the cycle unless the peripheral requests
;		  another cycle for the I/O (not yet implemented ...?)
;  10 - BRK PROG  BRK (in) PROG indicates a data break cycle (DMA) has been requested
;		  (only implemented when IDE disk is attached -- soon!).
;  11 - BRK	  BRK (cycle) means a requested data break (DMA ACK) is in progress
;		  (only implemented when IDE disk is attached -- soon!).
;
;
;  STATUS (STS)	- indicates an individual function for each bit:
;
;   0 - LNK	  LINK is the link (carry) bit
;   1 - GT	  Greater Than flag
;   2 - INT BUS	  Interrupt Bus, indicates that an I/O device has requested an IRQ
;   3 - NO INT	  No Interrupt Allowed, indicates that no interrupt is allowed just
;		  now, even if interrupts are nominally enabled
;   4 - ION	  Interrupt On, indicates interrupts are enabled
;   5 - UM	  User Mode, indicates that the processor is in user mode (IOT, HLT,
;		  OSR generate a trap) - Time Share option
;   6 - IF0
;   7 - IF1	  Instruction Field, these are the 3 EMA bits for a fetch or a direct
;   8 - IF2	  memory reference
;   9 - DF0
;  10 - DF1	  Data Field, the 3 EMA bits for an indirect memory reference
;  11 - DF2
;
;
;  AC		- indicates bits 0-11 of the accumulator AC at TS1
;  MD		- indicates accessed data just written or rewritten into memory
;  MQ		- indicates contents of MQ (Multiplier Quotient) register at TS1
;  BUS		- indicates bits 0-11 of the (OMNIBUS) I/O data lines. It is the
;		  12-bit value being OR-ed or loaded into AC during programmed I/O,
;		  for example the console I/O. DMA is different and uses MA/MD.
;
; One example: if address 0200 contains the instruction "TAD I 10", the addresses 0200
; and 10 are accessed with IF, and the memory to be added to AC is accessed with DF.
;
;
;
;  Real Console implementation
; -----------------------------
;
;    LED's : 15 - memory address         switches : 15,  keys: 6
;            12 - data register(s)       selector :  6 positions
;             1 - RUN LED                key      :  1 position "LOCK" (& PWR on/off)
;   ----------------------------------------------------------------------------------
;       output bits  : 28                   input bits  : 28
;       output ports : 4                    input ports : 4
;
;      ASSIGNMENT of OUTPUT PINs           ASSIGNMENT of INPUT PINs 
;
;      outport #0 = address low 8 bits      inport #0 = switch 4 - 11
;      outport #1 = address high 7 bits     inport #1 = switch 0 - 3 and SW
;      outport #2 = data low 8 bits                     SW, HALT, SING-STEP
;      outport #3 = data high 4 bits        inport #2 = 6 bit ISS knob
;                   and RUN LED             inport #3 = ADDR-LOAD, EXTD-ADDR-LOAD
;                                                       CLEAR, CONT, EXAM, DEP
;
;  Connection of the switches and keys to the I/O Board
; ------------------------------------------------------
;  I used switches and keys that have *3* positions. Only in the "up" and "down"
;  position a contact is closed. In the middle position none of the 2 contacts
;  is closed. Using 3-position switches gives the possibility to replicate the
;  position of the switches and keys in the best manner. With 3-position switches
;  the logical "up" position is the physical *middle* position of the lever; the
;  physical up position is not used/connected. See also http://www.pdp-11.nl and
;  click the link [Homebrew 'PDP-11'] in the menu at the left, then the link
;  [spin off] at the right top.
;
;  SW, HALT, SING-STEP switch, SR switches:     ADDR-LOAD, EXTD-ADDR-LOAD, CLEAR,
;  up (active) is "1", down position is "0"     CONT, EXAM, and DEP *keys* :
;  ** also for the 6 pos. rotary knob ISS       up is "0", down (activate) is "1"
;
;                    +5
;                   |
;                  ---                          nc ----O 
;                  | | 2200R                         .---o----------------> input
;                  | |                          +5 ----O         |
;                  ---                                          ---
;     nc ----O      |                                           | | 2200R
;          .---o----------------> input                         | | ( 1000R would
;        +---O                                                  ---   be better!)
;        |                                                       |
;       === gnd                                                 === gnd
;
;
; Wiring from panel to I/O Board   (I like to keep all things together in this file)
;
;  =======================================||=========================================
;  |____________ OUT01 connector _________||_____________ OUT23 connector __________|
;  | pin# color code    pin# color code   || pin# color code     pin# color code    |
;  --------------------+------------------||-------------------+--------------------|
;  | 1   yellow        | 2   grey/pink    || 1   black         | 2   pink           |
;  | 3   blue          | 4   red          || 3   green         | 4   grey/pink      |
;  | 5   white         | 6   white/beige  || 5   purple        | 6   yellow         |
;  | 7   brown         | 8   black        || 7   red           | 8   blue           |
;  | 9   not connected | 10  green        || 9   brown         | 10  white          |
;  | 11  red/blue      | 12  brown/yellow || 11  not connected | 12  white/green    |
;  | 13  grey          | 14  purple       || 13  not connected | 14  grey           |
;  | 15  pink          | 16  brown/green  || 15  not connected | 16  white/beige    |
;  ==================================================================================
;
;
;  =======================================||=========================================
;  |_____________ IN01 connector _________||_____________ IN23 connector ___________|
;  | pin# color code    pin# color code   || pin# color code     pin# color code    |
;  --------------------+------------------||-------------------+--------------------|
;  | 1   brown/yellow  | 2   grey/pink    || 1   grey          | 2   brown/green    |
;  | 3   brown         | 4   white        || 3   pink          | 4   purple         |
;  | 5   purple        | 6   red          || 5   not connected | 6   brown/yellow   |
;  | 7   pink          | 8   green        || 7   not connected | 8   red/blue       |
;  | 9   not connected | 10  grey         || 9   blue          | 10  green          |
;  | 11  brown/green   | 12  yellow       || 11  grey/pink     | 12  white          |
;  | 13  white/beige   | 14  red/blue     || 13  not connected | 14  not connected  |
;  | 15  blue          | 16  white/green  || 15  brown         | 16  yellow         |
;  ==================================================================================
;
; Reset of the Core Board via the console keys.
; To maintain the original console panel, but still have a RESET button available for
; the 6809 processor, the CLEAR and the DEP key are wired to a 74LS00 that is glued
; to the rear side of the backplate. The output of the NAND gate is connected to the
; *correct* pin of the RESET header on the Core.
;
;
; Additional notes on the Real Console implementation
;
; - With a small performance penalty the lamps Fetch, Defer, and Eexec are added with
;   each simulated instruction.  When the 6809 code accesses the jump table to the
;   instruction, the Fetch lamp is turned on. In each instruction F is turned off,
;   and E is turned on; when an indirect memory access is done, D is also turned on.
;   When the next instruction is taken from the table, Fetch goes on, and Defer and
;   Exec are turned off.
;
;   Also the IR0-2 indicators are loaded during the fetch step.  PAUSE may also need
;   to be set during the F cycle of an IOT, but that must be checked.  That might be
;   the case when to wait for the E cycle.
;   MD should be the data moving to or from memory, so during Fetch it should be the
;   instruction, during Defer it should reflect the address being read (and possibly
;   modified), and during Exec it should be updated to reflect the contents of the
;   memory being referred to (except during IOTs and OPRs).
;
;
;
;  Real Console definitions
; --------------------------
;			                 - 7 -- 6 -- 5 -- 4 -- 3 -- 2 -- 1 -- 0 -
WR8e_0	=	0	; outport #0 :    A7   A6   A5   A4   A3   A2   A1   A0
WR8e_1	=	1	;         #1 :   EMA2 EMA1 EMA0   -   A11  A10  A9   A8
WR8e_2	=	2	;         #2 :    D7   D6   D5   D4   D3   D2   D1   D0
WR8e_3	=	3	;         #3 :     -    -    -   RUN  D11  D10  D9   D8

RD8e_0	=	0	; inport  #0 :    S4   S5   S6   S7   S8   S9   S10  S11
RD8e_1	=	1	;         #1 :   HALT SSTP  SW    -   S0   S1   S2   S3
RD8e_2	=	2	;         #2 :     -    -   STE  STS  AC   MD   MQ   BUS
RD8e_3	=	3	;         #3 :    CNT   -   AL   EAL  CLR  -    EXM  DEP
			;		note: CONT key *must* be bit-7 (performance)


  ; inport #RD8e_1
SW_hlt	=	0b10000000	; input bit 7 = HALT switch
SW_stp	=	0b01000000	;       bit 6 = SING STEP switch
SW_sw	=	0b00100000	;       bit 5 = SW switch
SW_swh	=	0b00001111	;       bit [3-0] : mask Switch Reg [0-3]

  ; inport #RD8e_2
ISSste	=	0b00100000	; input bit 5 = STATE position
ISSsts	=	0b00010000	;       bit 4 = STATUS position
ISSac	=	0b00001000	;       bit 3 = AC position
ISSmd	=	0b00000100	;       bit 2 = MD position
ISSmq	=	0b00000010	;       bit 1 = MQ position
ISSbus	=	0b00000001	;       bit 0 = BUS position

  ; inport #RD8e_3
SW_cnt	=	0b10000000	; input bit 7 = CONT key  (** must be bit 7 **)
SW_al	=	0b00100000	;       bit 5 = ADDR LOAD key
SW_eal	=	0b00010000	;       bit 4 = EXTD ADDR LOAD key
SW_clr	=	0b00001000	;       bit 3 = CLEAR key
SW_exm	=	0b00000010	;       bit 1 = EXAM key
SW_dep	=	0b00000001	;       bit 0 = DEP key

ALLKEYS	=	SW_al+SW_eal+SW_clr+SW_cnt+SW_exm+SW_dep

RUNled	=	0b00010000	; output bit where RUN LED is connected

  ; output #WR8e_3 and WR8e_2 (Indicator Selector Switch == "STATE")
LD_fet	=	0b00001000	; output bit 3 = Fetch
LD_def	=	0b00000100	;            2 = Defer
LD_exe	=	0b00000010	;            1 = Execute
LD_ir0	=	0b00000001	;            0 = Instruction msb
LD_ir1	=	0b10000000	;            7 = Instruction nsb
LD_ir2	=	0b01000000	;            6 = Instruction lsb
LD_mdd	=	0b00100000	;            5 = MD DIR
LD_dac	=	0b00010000	;            4 = DATA CONT
LD_swt	=	0b00001000	;            3 = SW
LD_pau	=	0b00000100	;            2 = PAUSE
LD_bkp	=	0b00000010	;            1 = BRK PROG
LD_brk	=	0b00000001	;            0 = BRK

  ; output #WR8e_3 and WR8e_2 (Indicator Selector Switch == "STATUS")
LD_lnk	=	0b00001000	; output bit 3 = Link
LD_gtf	=	0b00000100	;            2 = Greater Than flag
LD_irb	=	0b00000010	;            1 = Interrupt Bus
LD_nia	=	0b00000001	;            0 = No Interrupt Allowed
LD_ion	=	0b10000000	;            7 = Interrupt On
LD_umd	=	0b01000000	;            6 = User Mode
LD_if0	=	0b00100000	;            5 = Instruction Field 0
LD_if1	=	0b00010000	;            4 = Instruction Field 1
LD_if2	=	0b00001000	;            3 = Instruction Field 2
LD_df0	=	0b00000100	;            2 = Data Field 0
LD_df1	=	0b00000010	;            1 = Data Field 1
LD_df2	=	0b00000001	;            0 = Data Field 2


  ; Input ports and output ports states
  ; - NOTE: keep these location CONSECUTIVE
  ;        (processed as an array thru indexing with offsets)

  ; memory allocation starts at <RCRAM>

  ; * Interrupt routine used storage (for input switch/toggle processing )
  ; * Accessed through an index, base address == <RCRAM>.
  ; * Number between brackets [] indicates the number of bytes required

CONIHI	=	3		; highest input port number used

SWTDBC	=	0		; [CONIHI] debounce timers for each port
PREVIN	=	SWTDBC+CONIHI+1	; [CONIHI] intermediary storage locations
INPORT	=	PREVIN+CONIHI+1	; [CONIHI] input ports (for application)
LSTIRQ	=	INPORT+CONIHI	; last interrupt routine index accessed location
LSTIRX	=	LSTIRQ+1	; 1 byte safety margin

BOUNCE	=	8		; # ticks used as switch/toggle debounce delay
TFRDLY	=	BOUNCE+5	; response delay for toggle release/activate

;====================================================================================
;====================================================================================
;
;         IIIIIII     DDDDDDD      EEEEEEE
;            I         D     D     E                   d                 k
;            I         D      D    E                   d    i            k
;            I         D      D    E                   d                 k
;            I         D      D    EEEEE           ddddd    i     ssss   k  k
;            I         D      D    E              d    d    i    s       k k
;            I         D      D    E              d    d    i    ssss    kk
;            I         D     D     E              d    d    i        s   k k
;         IIIIIII     DDDDDDD      EEEEEEE         ddddd    i    ssss    k  k
;
;
;
; Interface of an IDE hard disk to the 6809 Core board with one I/O Board.
;
;
; pin #		name	IDE hard disk interface description
; -----------------------------------------------------------------------------------
;
;  1		RESET*	A low signal level on this pin will reset the device
;
;  2,19,22,24	GND	ground
;  26,30,40
;
;  3,5,7,9,11	D7..D0	low data bus, 3=D7 .. 17=D0. This part of the bus is used
;  13,15,17		for the command and parameter transfer. It is also used for
;			the low byte in 16-bits data transfers.
;
;  4,6,8,10	D8..D15	high data bus, 4=D8 .. 18=D15. This part of the bus is used
;  12,14,16,18		only for the 16-bits data transfer.
;
;  20		-	Pin #20 is missing. Key pin.
;
;  21		DMARQ	DMA request. Not used, pulled down to GND with resistor 10k.
;  23		WR*	Drive I/O Write, write strobe of the bus.
;  25		RD*	Drive I/O Read, read strobe of the bus.
;  27		IORDY*	I do not use or connect to this pin. It is used to slow down
;			a controller when it is going too fast for the bus.
;			The Core + I/O Board do not have that problem.
;  28		CSEL	Cable select. Not used.
;  29		DMACK	DMA acknowledge. Not used.
;  31		INTRQ	Drive interrupt, interrupt output from the IDE devices.
;			This pin can be connected to the 6809 to generate interrupts
;			when a command is finished. Not used.
;  32		IOCS16	In AT, signal to enable the upper data bus drivers. Not used.
;  33		A1	of the IDE device you want to communicate with.
;  34		PDIAG*	Master/slave interface on the IDE bus itself.
;			Not used, pulled up with resistor of 10k (10000 Ohms) to +5 V.
;  35		A0	Addresses of the IDE bus. With these you select which register
;  36		A2
;  37		CS0*	The two /CS signals of the IDE bus. Used in combination with
;  38		CS1*	the A0-A1-A2 lines to select the register on the IDE device to
;			communicate with.
;  39		ACT*	A low level on this pin indicates that the IDE device is busy.
;		DASP*	You could connect 2 LED's gated with the RD* and WR* signal to
;			get the WR and RD lights on an RK05 drive. The software reads
;			the "drive busy" status from the IDE status register.
;
;
;  * OUTPUT only signals (from I/O Board to IDE disk)
;	A0, A1, A2, CS0*, CS1*, WR*, RD* and RESET*
;  * BI-DIRECTIONAL signals
;	Databus [D0:15]
;
;  * INPUT only signals (from IDE disk to I/O Board)
;	IRQ and ACT* (also called DASP*)
;	Connect a 10 KOhm pull-down resistor from the IRQ pin to ground.
;	ACT* can drive a LED (with a 330 Ohms series resistor) if you like.
;
;
;  IDE disk to I/O Board connection
; ----------------------------------
;
; The access to the IDE hard disk would be simple if the Core Board had a second, not
; used, PIA. A separate circuit board that piggy-backed the PIA on the Core Board
; and added one extra PIA would be a possible solution, but with software overhead
; it is possible to connect the IDE disk to the I/O Board.
; The 8 output only signals are directly tied to an output port of the I/O Board.
; The lower 8 databus bits [D0:7] are connected to an output port *AND* an input
; port in parallel.  Since the output ports on the I/O Board are always enabled,
; a *patch* is required so that the output port can be set into 3-state mode so that
; reading data from the IDE disk is possible without shorting the databus against the
; output port.  The patch requires one bit from an extra output port; the OE* pin of
; the output latch used for the bi-directional databus is cut from its GND connection
; and it is wired to that single bit of the extra output port.
; IRQ, and ACT* are not tied to the I/O Board. The upper 8 databus bits [D8:15] is an
; option. Using only 8 bits means that half of the disk capacity will not be used,
; but who wants more than a few Mb's on a PDP8/e anyway ... Not using D8:15 saves one
; output port and one input port. The code has an assembly flag to select between
; 8/16 bit data transfers.

    ; OUTport #WRide_x

HDdirW	= 0x00		; bit 7 = "0" :: write to IDE disk 8/16 bit data
HDdirR	= 0x80		; bit 7 = "1" :: read from IDE disk lower 8/16 bit data


;
;			                 - 7 -- 6 -- 5 -- 4 -- 3 -- 2 -- 1 -- 0 -
RDide_dl	= 4	; INport  #4 :    D7   D6   D5   D4   D3   D2   D1   D0
WRide_dl	= 4	; outport #4 :    D7   D6   D5   D4   D3   D2   D1   D0
WRide_c		= 5	; outport #5 :    RST*  RD*  WR* CS1* CS0*  A2   A1   A0
WRide_x		= 6	; outport #6 :    DDIR  -    -    -    -    -    -    -
;
;			                 - 7 -- 6 -- 5 -- 4 -- 3 -- 2 -- 1 -- 0 -
RDide_du	= 5	; INport  #5 :    D15  D14  D13  D12  D11  D10  D9   D8
WRide_du	= 7	; outport #7 :    D15  D14  D13  D12  D11  D10  D9   D8



; The approach of using the available I/O Board for interfacing the IDE disk needs
; some special attention for the software implementation. I try explain this here.
; The I/O Board is a resource of the system, and as such, needs "scheduling".
; It must be prevented *at all time* that the console (FIRQ routine!) and some disk
; access routine are using the I/O channel(s) at the same time as they can change
; the data transport direction of the I/O control.
; In fact, all *access sequences* to the I/O Board are mutually exclusive and must
; be completed before any other process may use the I/O Board.
; For the pdp8/e simulator this is fairly easy because all accesses to the I/O Board
; are performed in the FIRQ routine. To support the IDE disk interface via the I/O
; board it is only necessary the disable the FIRQ bit in the 6809 CPU. As the state
; of the IRQ and the FIRQ flag can change, the IDE routines must save the current
; value and, when the disk access is finished, must restore that value.
; Thank goodness, this system has only one CPU, so P and V operations and process
; interlocks (OS fundamentals) are not needed here.
;
; Any disk task requested will need to change the I/O data transfer direction several
; times. For example: to *write* a command to the disk, the software must first check
; the disk's status by a *read* operation. After the command is written, again it is
; necessary to check the status (data request/available), etc.  To keep things simple
; all routines set the I/O data transfer direction explicit, but a flag indicates if
; the I/O data transfer direction change is required. the flag used is (piadir).
;
;
;
;   ==============                                ==========
;   #            #                                |        |--- 7  RST* ------+++
;   #            #                                |  OUT   |--- 6  RD* -------|||
;   #            #             control            |        |--- 5  WR* -------|||
;   #            #<<---------------------/--------| cntrl  |--- 4  CS1* ------||| C
;   #            #                       8        |        |--- 3  CS0* ------||| O
;   #            #                                |        |--- 2  A2 --------||| M
;   #            #                                |        |--- 1  A1 --------||| M
;   #    IDE     #                ==========      |WRide_c |--- 0  A0 --------||| O
;   #            #                |        |      ==========                  ||| N
;   #            #                |        |                                  |||
;   #   disk     #          (16)  |  OUT   |                                  |||
;   #   drive    #       +---/----|        |                                  ||| D
;   #            #       | +-/----|  data  |----------------------------------||| A
;   #            #       | | 8    |        |     +5                           ||| T
;   #            #       | |      |WRide_du|     |                            ||| A
;   #            #   (16)| |      |WRide_dl|    | |4700                       ||| B
;   #            #<<--/--+ |      ==========    |_|                           ||| U
;   #            #       | |        ^            |  =========                 ||| S
;   #            #<<--/----+        +--air-wire--+--|       |--- 7  HDdir ----|||
;   #            #    8  | |                        |  OUT  |                 |||
;   #            #       | |      ==========        |       |                 ||| t
;   #            #       | |      |        |        |       |                 ||| o
;   #            #       | |      |   IN   |        |       |                 |||
;   #            #       | |  8   |        |        |       |                 |||
;   #            #       | --/-->>|  data  |---+    |       |                 ||| P
;   #            #       +---/-->>|        |   |    |WRide_x|                 ||| I :
;   #            #          (16)  |        |   |    ---------                 ||| A :
;   #            #    :           |RDide_du|   |                              |||   :
;   #            #    :           |RDide_dl|   +------------------------------+++   :
;   ==============    :           ==========                                        :
;                     |                                                             |
;                     |__________________________________________I/O Board__________|
;
;
;
; Wiring from (my) IDE disk to I/O Board
;
;  ===========================================================================
;  |_____________________________ OUT56 connector ___________________________|
;  | pin# function     color code       | pin# function     color code       |
;  -------------------------------------|------------------------------------|
;  |  1    D4         white             |  2    D0         pink              |
;  |  3    D5         purple            |  4    D1         blue              |
;  |  5    D6         yellow            |  6    D2         brown             |
;  |  7    D7         grey              |  8    D3         green             |
;  |  9    CS1*       red               |  10   A0         white/beige       |
;  |  11   WR*        green/brown       |  12   A1         yellow/brown      |
;  |  13   RD*        green/white       |  14   A2         black             |
;  |  15   RST*       red/blue          |  16   CS0*       grey/pink         |
;  ===========================================================================
;
;
;  ===========================================================================
;  |_____________________________ OUT78 connector ___________________________|
;  | pin# function     color code       | pin# function     color code       |
;  -------------------------------------|------------------------------------|
;  |  1    -                            |  2    -                            |
;  |  3    -                            |  4    -                            |
;  |  5    -                            |  6    -                            |
;  |  7    DDIR       green             |  8    -                            |
;  |  9    D12        white             |  10   D8         pink              |
;  |  11   D13        purple            |  12   D9         blue              |
;  |  13   D14        yellow            |  14   D10        brown             |
;  |  15   D15        grey              |  16   D11        green             |
;  ===========================================================================
;
;
;  ===========================================================================
;  |______________________________ IN56 connector ___________________________|
;  | pin# function     color code       | pin# function     color code       |
;  -------------------------------------|------------------------------------|
;  |  1    D4         white             |  2    D0         pink              |
;  |  3    D5         purple            |  4    D1         blue              |
;  |  5    D6         yellow            |  6    D2         brown             |
;  |  7    D7         grey              |  8    D3         green             |
;  |  9    D12        white             |  10   D8         pink              |
;  |  11   D13        purple            |  12   D9         blue              |
;  |  13   D14        yellow            |  14   D10        brown             |
;  |  15   D15        grey              |  16   D11        green             |
;  ===========================================================================
;
;    NOTE 1:	INport #7 not used (for now).
;    NOTE 2:	OE* (pin #1) of OUTport #6 (IC7) is *not* in the IC socket
;		which means not connected to GND, but is connected literally
;		by an "air wire" to pin #19 of OUTport #5 (IC6).
;		To pull-up the signal voltage, a 4700 Ohms resistor is wired
;		from the OE* (pin #1) to +5 V. (pin #20).
;    NOTE 3:	If you decide to go for 16 bit data transfers, you must add
;		an extra output and input port. The extra output ports OE*
;		must *** ALSO *** be connected to the air wire !!
;
;  Ideas for the future.
;  INport #7 could be used to read two switches: LOAD/RUN and WRITE-PROT, as
;  found on the RK05 disk drive.
;  OUTport #5 bits could be used to drive a few LED's as found on the front
;  panel of the RK05. Attention: access to OUTport #5 *could* enable the OUT
;  port #6 (I/O data) at a very bad time ... The 7 remaining bits of OUTport #5
;  are enough, as the WR and RD light can be made with TTL logic from the WR*
;  and RD* signal, gated with the ACT* signal (pin #39) of the IDE disk drive.
;
;
;  IDE read/write and register description
; -----------------------------------------
; The IDE disk appears to the IDE bus as a set of registers. The register selection
; is done with the CS[0:1]* and A[0:2] lines. Reading/writing is done with the RD*
; and WR* signals.
;
;  * READ CYCLE
;    1. disable the data output port (set to 3-state).
;    2. set up an address, CS0* and CS1* signals on the control output port.
;    3. activate the RD* of the IDE disk.		\
;    4. read the data from the data input port.		 > repeat for each byte/word
;    5. deactivate the RD* signal.			/
;
;  * WRITE CYCLE
;    1. enable the data output port.
;    2. set up an address, CS0* and CS1* signals on the control output port.
;    3. write a data byte on the data output port.	\
;    4. activate the WR* of the IDE disk.		 > repeat for each byte/word
;    5. deactivate the WR* signal.			/
;
; ===================================================================================
; CS1* CS0*  A2  A1  A0
; -----------------------------------------------------------------------------------
;  1    0    0   0   0		RD/WR: data I/O register (16-bits)
;			    This is the only 16-bits wide register of the entire IDE
;			    interface. Used for all data block transfers from and to
;			    the IDE disk, but here only the lower 8 databits are used
;  1    0    0   0   1		RD: error information register
;				WR: write precompensation register
;			    Read this register when an error is indicated in the IDE
;			    status register (see below for the IDE status register).
;			    The write precompensation is not used.
;  1    0    0   1   0		Sector counter register
;			    This register can be used to do multi-sector transfers.
;			    You must write the number of sectors to transfer in this
;			    register.   I use one-sector only transfers, so in the
;			    code only 0x01 is written into this register.
;			    The register is also used to pass the parameter timeout
;			    for the idle mode command.
;  1    0    0   1   1		Start sector register
;			    Holds the start sector of the current track to start the
;			    read or write operation from.   For each transfer the
;			    start sector is written in this register.
;			    Note: the sector count starts at 1 and runs up to 256,
;				  so writing 0x00 results in sector number 256.
;			    If LBA is used, this register holds LBA bits [7:0].
;  1    0    1   0   0		Low byte of the cylinder number
;			    This register holds the low byte of the cylinder number
;			    for a disk data transfer.
;			    If LBA is used, this register holds LBA bits [15:8].
;  1    0    1   0   1		High two bits of the cylinder number
;			    The traditional IDE spec allows only cylinder numbers in
;			    the range 0..1023. This register gets the two upper bits.
;			    The cylinder number's upper 2 bits are written into this
;			    register before each transfer.
;			    If LBA is used, this register holds LBA bits [23:16].
;  1    0    1   1   0		Head and device select register
;			    Bits [3:0] of this register hold the head number (0..15)
;			    for a transfer. Bit [4] is to be written 0 for access to
;			    the IDE master device, 1 for access to the slave device.
;			    Bits [7:5] set the addressing mode to be used.
;				"101" = Cylinder/Head/Sector addressing mode
;				"111" = Logical Block Addressing mode is used.
;			    If LBA is used, this register holds LBA bits [27:24].
;  1    0    1   1   1		WR: Command register
;				RD: Status register
;			    When written the IDE disk regards the data written as a
;			    command. When read you get the status of the IDE disk.
;			    Reading this register also clears any interrupts from the
;			    IDE disk to the 6809 (not used / connected).
;  0    1    1   1   0		WR: interrupt and reset register
;				RD: 2nd status register
;			    When read this register gives you the same status byte as
;			    the primary (CS0*=0, CS1*=1, A[2:0]=111) status register.
;			    The only difference is that reading this register does
;			    not clear the interrupt from the IDE disk when read. When
;			    written you can enable/disable the interrupts the disk
;			    generates. Also used to give the disk a software reset.
;  0    1    1   1   1		active status of the IDE disk
;			    In this read-only register you can see if the IDE master
;			    or slave is currently active and find the currently
;			    selected head number. This register is not used.
; ___________________________________________________________________________________
;
;
;  Register bit specifications
; -----------------------------
;
;  * head and device register
;	A write register that sets the master/slave selection and the head number.
;		bits 3..0: head number [0..15]
;		bit  4   : master/slave select: 0=master, 1=slave
;		bits 7..5: fixed at '101' (512 bytes/sector, that's IDE)
;
;   * status register
;	Both the primary and secondary status register use the same bit coding.
;	The register is a read register.
;		bit 0    : error bit. If this bit is set then an error has occurred
;			   while executing the latest command. The error status
;			   itself is to be found in the error register.
;		bit 1    : index pulse. Each revolution of the disk this bit is
;			   pulsed to '1' once.
;		bit 2    : ECC bit. if this bit is set then an ECC correction on
;			   the data was executed. This bit is ignored here.
;		bit 3    : DRQ bit. If this bit is set then the disk either wants
;			   data (disk write) or has data for you (disk read).
;		bit 4    : SKC bit. Indicates that a seek has been executed with
;			   success. This bit is also ignored here.
;		bit 5    : WFT bit. indicates a write error has happened. Not used.
;		bit 6    : RDY bit. indicates that the disk has finished its power
;			   up. Wait for this bit to be active before doing any
;			   action (execpt reset) with the disk.
;		bit 7    : BSY bit. Bit is set when the disk is doing something
;			   for you. You must wait for this bit to clear before you
;			   can send any command to the disk.
;
;   * interrupt and reset register
;	This register has only two bits that do something. It is a write register.
;		bit 1    : IRQ enable. If this bit is '0' the disk will give an
;			   interrupt when it has finished executing a command. When
;			   it is '1' the disk will not generate interrupts.
;		bit 2    : RESET bit. If you pulse this bit to '1' the disk will
;			   execute a software reset. The bit is normally '0'. It is
;			   not used because there is full software control of the
;			   hardware RESET* line.
;
;   * active status register
;	This is a read register. Ignored because only one IDE disk is connected.
;		bit 0    : master active. If this bit is set then the master IDE
;			   disk is active.
;		bit 1    : slave active. If this bit is set then the slave IDE
;			   disk is active.
;		bits 5..2: complement of the currently active disk head.
;		bit 6    : write bit. This bit is set when the device is writing.
;		bit 7    : in a PC environment this bit indicates if a floppy is
;			   present in the floppy drive. Here it has no meaning.
;
;   * error register
;	The error register indicates what went wrong when a command execution
;	results in an error. The fact that an error has occurred is indicated in
;	the status register, the explanation is given in the error register.
;	This is a read register.
;		bit 0    : AMNF bit. Indicates that an address mark was not found.
;		bit 1    : TK0NF bit. When this bit is set the drive was not able
;			   to find track 0 of the device. I think you will have to
;			   throw away the disk if this happens.
;		bit 2    : ABRT bit. This bit is set when you have given a wrong
;			   command to the disk. Mostly wrong parameters (wrong head
;			   number etc..) cause the disk to respond with error flag
;			   in the status bit set and the ABRT bit set.
;		bit 3    : MCR bit. Indicates that a media change was requested.
;			   Ignored this bit; media change ...?
;		bit 4    : IDNF bit. Means that a sector ID was not found.
;			   It means that a non-existing sector is requested.
;		bit 5    : MC bit. Indicates that the media has been changed.
;			   This bit is ignored.
;		bit 6    : UNC bit. Indicates that an uncorrectable data error has
;			   occurred. Some read or write errors could cause this.
;		bit 7    : reserved.
;
;
;  Some historical points
; ------------------------
; At first (MSDOS versions till 3.3) the disk interface was not able to access more
; than 32MB on one volume. That was a imitation of the MSDOS file system rather than
; of the disk interface. The limit came from the fact that each disk sector
; (512 bytes) was registered in the FAT in a 16-bits word. The total partition size
; was limited by the fact that only 65536 sectors could be addressed. The partition
; size was thus limited to 65536 x 512 bytes = 32 MBytes.
;
; Later -as the disks became larger- the disk interface itself ran into its limits.
; The interface described here has room for 16 heads, 256 sectors per track and only
; 1024 cylinders. With the standard sector size of 512 bytes that leaves you a max
; disk size of 16 x 256 x 1024 x 512 = 2048 MBytes. That is a real limitation of the
; IDE interface decribed here. It can not access more than 2 Gb of disk space.
;
; This was overcome by introducing the so-called LBA modus. In LBA modus the sectors
; are simply numbered from 0 to -big-.  The lowest byte of the LBA sector number is
; written into the sector number register, the middle 16 bits of the LBA sector
; number are written in the cylinder number registers (low and high, all 16 bits are
; used). The highest 4 bits of the LBA sector number are written in the head and
; device register. That gives you 28 bits of LBA sector number.  The sector size was
; again fixed at 512 bytes, so in LBA modus you have access to: 2^28 x 512 = 1.37e11
; (some 137.4 Gb) of disk space.
; If you want to connect a disk larger than 2 GBytes to this IDE interface you too
; will have to use the LBA modus. How to do that: the bit 6 of the head and device
; register is set to indicate that LBA modus is used (the fixed pattern of '101' in
; the bits 7..5 of the head and device register is to be changed into '111').
; All other manipulation of the IDE interface is the same for Sector/Head/Cylinder
; mode and LBA mode.
;
;
;  IDE registers usage
; ---------------------
; 1) Before doing anything with a device you have to wait till it indicates that it
;    is ready (RDY bit in the status register)
; 2) Next load the parameters of a command into the appropriate registers.
;    For read/write commands, that comes down to writing the cylinder/head/sector
;    numbers into the registers.
; 3) Issue a read or write command.
; 4) Wait till the device signals that it is ready for data transfer
;    (DRQ in the status register).
; 5) Feed the device data (for write) or get the data from the device (for read).
;    In case of a write you could wait for the operation to complete and read the
;    status register to find out what has become of your data.
;
;  IDE commands
; --------------
; Here is the *used* command set described. See the ATA-3 spec for full details.
; Remember, when giving a command you first have to wait for device ready, next put
; the command parameters in the registers and only then can you give a command (by
; writing a command byte to the command register). The disk will start executing
; the command right after you've written the command into the command register.
;
;   IDE
; command	description
; -----------------------------------------------------------------------------------
;   0x1*	recalibrate the disk. Remark: the '*' means that the lower nibble of
;		the command byte is a don't care. All commands 0x10 ..0x1F will cause
;		a recalibrate disk command being executed.
;		This command has no parameters. You simply write the command code to
;		the command register and wait for ready status to turn active again.
;   0x20	Read sector with retry. Remark: 0x21 is read sector without retry.
;		For this command you must load the cylinder/head/sector first. When
;		the command completes (DRQ goes active) you can read 256 words
;		(16-bits) from the disk's data register.
;   0x30	Write sector (with retry; 0x31 is without retry).  Again, you must
;		load the cylinder/head/sector data. Then wait for DRQ to turn active.
;		Feed the disk 256 words of data in the data register. Next the disk
;		starts writing. When BSY goes not active you can read the status from
;		the status register.
;   0x7*	Seek. This normally does nothing on modern IDE drives. Modern drives
;		do not position the head if you do not command a read or write.
;   0xEC	Identify drive. This command prepares a buffer (256 words) with
;		information about the drive. To use it simply give the command, wait
;		for DRQ and read the 256 words from the drive.  Use the parameters
;		reported to autoconfig the interface to match the disk configuration.
;   0xE0	Spins down the drive at once. No parameters.
;   0xE1	Spins up the drive again.
;   0xE2, 0xE3	Auto-powerdown the disk. Write in the sector count register the time
;		(5 seconds units) of non-activity after which the disk spins down.
;		Write the command to the command register and the disk is set in an
;		auto-power-save modus. The disk will automatically spin-up again when
;		you issue read/write commands. 0xE2 will spin-down, 0xE3 will keep
;		the disk spinning after the command has been given.
;		Example: write 0x10 in the sector count register, give command 0xE2
;		and the disk will spin-down after 80 seconds of non-activity. BTW:
;   0xF2, 0xF3	The same as 0xE2 and 0xE3, only the unit in the sector count register
;		is interpreted as 0.1 sec units. Not tested.
;
;
    ; bits of the control byte on OUTport #WRide_c

DCNop	=	0b11111000	; nothing on IDE control bus
DCRes	=	0b10000000	; bit  [7]  = RST* bit
DCIor	=	0b01000000	; bit  [6]  = RD* bit
DCIow	=	0b00100000	; bit  [5]  = WR* bit
DCCs1	=	0b00010000	; bit  [4]  = CS1* bit
DCCs0	=	0b00001000	; bit  [3]  = CS0* bit
DCAd2	=	0b00000100	; bit  [2]  = A2 bit
DCAd1	=	0b00000010	; bit  [1]  = A1 bit
DCAd0	=	0b00000001	; bit  [0]  = A0 bit
DCAdr	=	0b00000111	; bit [2:0] = address mask bits

    ; IDE register addresses (RESET*,RD*, and WR* inactive)

IDECmd	=	0b11110111	; command (WR)
IDESts	=	0b11110111	; status register (RD)
IDEHd	=	0b11110110	; head number
IDEcylH	=	0b11110101	; cylinder number high
IDEcylL	=	0b11110100	; cylinder number low
IDEsec	=	0b11110011	; sector number
IDEnum	=	0b11110010	; number of sectors
IDEErr	=	0b11110001	; error register
IDEData	=	0b11110000	; data bus
IDEDvc	=	0b11101110	; drive control (WR)
IDEAlt	=	0b11101110	; alternate status (RD)

    ; head number (0..F) also has the mask for master/slave
    ; Hard-coded to 'master' as I will not use a second drive.

IDEHdA	=	0b00001111	; head number "AND" mask
; IDEHdO =	0b10100000	; head number "OR" mask: drive #0, use C/H/S
IDEHdO	=	0b11100000	; head number "OR" mask: drive #0, use LBA


    ; bits from Device control register - IDEDvc

IDERes	=	0b00000100	; Soft Reset bit
IDEIrq	=	0b00000010	; 0 :: IRQ active

    ; bits from Status Register - IDESts

StsBsy	=	0b10000000	; busy flag
StsRdy	=	0b01000000	; ready flag
StsWft	=	0b00100000	; Write error
StsSkc	=	0b00010000	; seek complete
StsDrq	=	0b00001000	; Data Request
StsCorr	=	0b00000100	; ECC executed
StsIdx	=	0b00000010	; Index found
StsErr	=	0b00000001	; error flag

    ; IDE disk Command opcodes

CmdRecal	=	0x10	; recalibrate disk
CmdRead		=	0x20	; read sector(s)
CmdWrite	=	0x30	; write sector(s)
CmdDiag		=	0x90	; execute diagnostics
CmdStop		=	0xE0	; stop disk
CmdStrt		=	0xE1	; start disk
CmdIdent	=	0xEC	; identify disk


    ; In the default state the control is inactive: RESET*, RD* and WR* signals
    ; are set to "1", CS1* and CS0* are set to "1", and A[2-1-0] = "000" (#DCNop).
    ; I use a memory command packet to pass parameters to the disk.
    ; Packet description:

SDA	=	0	; offset src/dest address in memory for the operation
LBA3	=	2	; offset LBA
LBA2	=	3	;   32-bits number for a disk of max 2.1E12 bytes !
LBA1	=	4	;
LBA0	=	5	;


    ; Lineair Block Access (LBA) is used ALL THE TIME.
    ; For the disk that does not support LBA by itself, a routine called SetLBA
    ; is available to convert the LBA to a CHS configuration.
    ; 2 parameters are used to describe the disk geometry:
    ;    - the number of blocks per cylinder; in CHS terms this is HxS
    ;    - the number of blocks per track; in CHS terms this is S
    ; From these two parameters, conversion of LBA to CHS is done.
    ; They are defined by the following two words.


    ; Micron MTCF016A 16 Mb CompactFlash card with specified CHS: 490/2/32.
    ; SanDisk 32 Mb CompactFlash card with specified CHS: 490/4/32.


    ; Quantum ProDrive LPS420 with specified CHS: 1010/16/51.

SPC	=	1010		; # Sectors per Cylinder
SPT	=	51		; # Sectors per track




;====================================================================================
;====================================================================================
;
;                           D                88888         //
;                           D               8     8       //
;                           D               8     8      //    EEEEE
;           PPPPPP     DDDDDD    PPPPPP      88888      //    E     E
;           P     P   D     D    P     P    8     8    //     EEEEEE
;           P     P   D     D    P     P    8     8   //      E
;           PPPPPP     DDDDDD    PPPPPP      88888   //        EEEEE
;           P                    P
;           P                    P
;           P                    P
;
;
;     This program emulates a pdp8/e on the 6809 Core board. It is based on
;     the PDP8.ASM program, with the addition of the KM8-E Memory Extension
;                            and the Time Share module.
;
;====================================================================================
;====================================================================================
;
;
;     PROBLEMS TO SOLVE  /  STUFF TO ADD
;    ------------------------------------
;
; 1. The BIN loader in this implementation reads a leader, the data part and
;    then a trailer. Papertapes exist that have multiple parts, so the trailer
;    is also the leader for the next part. The "real" bin loader used flow
;    control to stop the reader when it was halted.
;    Then you could read the checksum and press "CONTinue".
; 2. Implement a good solution for flow control. RTS/CTS is possible, but not
;    from the ACIA on the Core Board. The use of one in- and output pin of the
;    I/O Board's ports produces a workable situation ...(?)
; x.
;
;
;
; The PDP-8 is a 12-bit computer with 4096 words of memory. It has only eight
; instructions, a 12-bit program counter, the 12-bit accumulator AC, and a
; single-bit register, the link (L) bit, prefixed to AC as <L,AC>.
; All instructions have the following 12-bit format:
;
;  11  10  9   8   7   6   5   4   3   2   1   0
;  ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ 
; |   |   |   |   |   |   |   |   |   |   |   |   |  i - indirect bit [0=direct,
; |_ _|__ |__ |__ |__ |___|___|__ |__ |__ |__ |__ |                    1=indirect]
; :           :   :   :                           :  z - page bit [0=page zero, 
; :  opcode   : i : z :           address         :                1=current page]
;                                                    address - 12-bit word in page
;
; Note. Absolute memory locations 8 through 15 are incremented prior to use when
;       used as indirect addresses. These locations are called auto-index registers.
;
; The 8 opcodes (instructions) are the following:
;
;  000 - AND - logical AND of AC with memory
;  001 - TAD - 2-complement add of memory to <L,AC> (13-bit value)
;  010 - ISZ - increment memory and skip next instruction if result is 0
;  011 - DCA - deposit AC in memory and clear AC
;  100 - JMS - jump to subroutine. Current PC is stored in subroutine start address.
;  101 - JMP - jump unconditional
;  110 - IOT - input/output transfer
;  111 - OPR - microcoded operations, 3 groups
;
;
;  Examining memory
; ------------------
;  1. set the Indicator Selector Switch to the position "MD"
;  2. set the Switch Register to the address to be inspected
;  3. push down the "ADDR LOAD" key. The top row should match the Switch Register.
;  4. push down the "EXAM" key. The bottom row shows the contents of the address.
;     note: the top row now shows address+1. The address is one ahead of the contents
;           this is just the way how the pdp8/e works -- get used to it!
;  5. push "EXAM" again. The contents of address+1 is displayed on the bottom row,
;     and (as you guessed) the top row displays address+2.
;
;  Loading memory
; ----------------
;  1. set the Indicator Selector Switch to the position "MD"
;  2. set the Switch Register to the first address to be loaded
;  3. push the "ADDR LOAD" key
;  4. set the Switch Register to the first 12-bit data word to be stored
;  5. push the "DEP" key. Note that the address is now one ahead of the contents.
;  6. repeat steps 4-6 as long as you enter data in consecutive memory locations.
;
;  Running a program
; -------------------
;  1. set the Switch Register to the start address of the program loaded into memory
;  2. push the "ADDR LOAD" key
;  3. push the "CLEAR" key
;  4. make sure that "HALT" and "SING STEP" are in the 'up' position
;  5. push the "CONT" key.
;     When the program stops, the top row of lights shows the address+1 at which the
;     program terminated, and (if the Indicator Selector Switch is set to "MD") the
;     bottom row of lights shows the last executed instruction, e.g. Halt (7402).
;
;
;  Emulator data
; ---------------
;
; (1) Emulated program space is 2 (8-bit) bytes for each PDP-8 (12-bit) location.
; (2) Non-existent fields are read as 0.
; (3) Interrupt inhibit (removed by a JMP or JMS) is indicated by the first
;     bit in the interrupt flags being set.  Instruction delay (due to ION) is
;     handled by changing a pointer. If the pointer indicates that the interrupts
;     are already enabled it is not changed.  It is assumed that ION delay and
;     JMP/JMS inhibit are not counted in checking if the interrupt system is on.
;     See the JMP code for further explanation.
;
; Memory data format:
;
;   ooooCCCIZAAAAAAA
;
;	o = Must be 0 (only 12 bits used)
;	C = Opcode
;	I = Indirect address if 1
;	Z = Current page if 1
;	A = Page address
;
; Memory Extension and Time Share:
;
;    There are three register sets: Saved, Working and Buffer.
;    Each set has both Field numbers and pointers.  The User field is really
;    a flag for User/Executive mode.  If the bit is 0 it is in the Executive
;    (standard) mode, else it is in the User mode in which all LAS, HLT and IOT
;    instructions are bypassed and the Time Share interrupt set. The registers
;    include the memory pointers and the field number bytes, as follows:
;
;	  U= Time Share flag (0= Executive, 1= User).
;	III= Instruction field number.
;	DDD= Data field number.
;
;	__________________________________________________________
;	| Function  || RAM Registers                               |
;	|           ||_____________________________________________|
;	|           || I field | D field | Field bytes and bit map |
;	|           || pointer | pointer |_________________________|
;	|           ||         |         | U and I    | D          |
;	|           ||         |         | (U0000III) | (00000DDD) |
;	|===========||=========|=========|============|============|
;	|   Working ||  WrkIp  |  WrkDp  |     WorkUI |      WorkD |
;	|-----------||---------|---------|------------|------------|
;	| Interrupt ||  SavIp  |  SavDp  |     SaveUI |      SaveD |
;	|-----------||---------|---------|------------|------------|
;	|   J Delay ||  BufIp  |         |     BuffUI |            |
;	|___________||_________|_________|____________|____________|
;
;
; Modifications:
;
; (1) 4-8-97	F. Wilson: Fixed small errors in the Extended Memory and
;			 Time Share module.
;		Is not tested yet, and will need update of the test compiler.
; (2) 4-15-97	F. Wilson: Added checks for breakpoints and a monitor.
;			 Single stepping is the starting default.
; (3) 5-2-97	F. Wilson: Put Instruction Field into GTF.
; (4) 5-6-97	F. Wilson: Removed single step routine and incorporated into
;			 a primitive monitor. Checked emulator speed with
;			 the LED display on and it drops by over 40%.
; (5) 5-12-97	F. Wilson: Removed indirects from field pointer access in field
;			 change IOT's.
; (6) 6-11-97	F. Wilson: Changed vector lists to end in 0.
; (7) 9-3-97	F. Wilson: Put ^C exit into the binary loader.
;		Made location 200 in Field 0 the default program start.
; (8) 6-14-04	H. Gooijen: Adapted the 6809 code to be used with the Core Board,
;			 see for more info http://www.pdp-11.nl [homebrew PDP-11].
;		     The code can now be put in EPROM. The self-modifying code is
;		     removed and a 'decent' implementation has replaced it.
;		     Changed the source code to be compatible with the MACRO-11
;		     syntax, as the code is assembled with the ASxxxx assembler of
;		     A. Baldwin.  The original implementation used 2 bits of a PIA
;		     to create a serial interface. As the Core Board has a real
;		     ACIA, the PIA code is removed. With this the option to change
;		     the baudrate is gone too, though ...
; (9) 11-11-04	On HLT condition keep the console responsive instead of a busy
;		wait loop waiting for hardware reset (only *with* RealConsole).
; 10) dec-2004	Added IOT's for the DF32 Disk File and routines to interface
;		an IDE hard disk drive through the I/O Board.
; 11) jan-2005	Added the extra IOT's for the RF08. A conditional assembly flag
;		defines the simulated device on the IDE drive: DF32 *or* RF08.
; 12) feb-2005	Added a bootstrap for DMS and for OS/8.  DMS works!
;		OS/8 is a work in progress. Found errors (thanks Vince !!) in the
;		implementation of the GTF instruction, and now that the Df, Data
;		field 3 bits can be something else than "000", Pntreg prints out
;		the U mode flag wrong. Corrected. The GTF instruction restored the
;		fields from WorkUI and WorkD. That must be SaveUI and SaveD. Further
;		it is not yet clear how the Interrupt Inhibit bit (Time Share) works.
;		Added the restore of BuffIp and BuffUI in the Chkint routine.
;		The CIF instruction tests the Time Share bit, but used all bits. It
;		should only test bit [7], msb. Further, RDF loads WorkD into [B],
;		but rotates the bits into the correct position with 3 "lsla". That
;		must of course be 3 "lslb".
; 13) mar-2005	Running the MAINDEC-8E-D0AB shows that IAC does not set the LINK bit
;		correctly if AC overflows by the increment. Further IAC does the
;		following operation: Link.AC := (Link.AC) + 1   So, Link must be
;		included with AC *before* the increment is done.
;		JMPsub & JMSsub change the I field (caused by a CIF instruction just
;		before the JMP or JMS) *before* GetAdr is called. That sequence will
;		generate an effective address in the new I Field which is wrong.
;		If a field change must be done, it should be done at the end of the
;		JMPsub and JMSsub routines.
;		At the label ClrSys the upper byte of IOflgs was cleared. This also
;		clears the Jdelay flag, but that flag must be preserved. The fix
;		preserves Jdelay, and at the start of the program, before ClrSys is
;		called, the Jdelay bit in IOflgs is initially explicit set to zero.
; 14) summer	Spent a lot of time together with the help of Vince to debug the
;      2005	code to get OS/8 running. The MAINDEC for extended memory and Time
;		Share *and* the MAINDEC JMP-JMS test run succesfull, but when the
;		software uses I Field changes, we finally found an error in the
;		6809 code that processes the JMS instruction!  So, it turned out
;		that a succesfull running MAINDEC is not always a guarantee things
;		will indeed work as expected!
;		To locate the problems I added the emulation commands and the trace
;		facility. I keep that in the code, it might prove useful in the
;		future (if/when I add RK05 support through a 3.5" floppy disk...)!
;
;
;
;  Speed improvements
; --------------------
; 1) moved the GetAdr routine so that some often used pdp8 instructions can reach
;    that routine through "bsr" instead of "jsr".
; 2) to update the state flags, the contents of an accumulator must be saved (in
;    most instances). That was done through the stack (pshs/puls).  The psh/pul
;    instruction both need 5 cycles. When sta/lda is used accessing a direct page
;    address, both these instructions need only 4 cycles.  Thus this sequence is
;    20% faster. The memory location *used solely* for this purpose is "savacc".
; 3) The "ISZsub" routine uses at the end the sequence "std  ,u" followed by a
;    "cmpd  #0" and a "bne" instruction. Since the "std" instruction already sets
;    the Z condition code, the cmpd instruction can be removed.
; 5) The sequence clra/clrb uses 2 bytes, whereas ldd #0 gives the same result
;    but uses one byte more. The "ldd #0" is preferred as clra/clrb takes 4 cycles
;    and the ldd #0 takes 3 cycles (one less).
; 6) ... update the Core Board ...
;    change the 1 MHz parts (6809, 6821, 6850) to the 2 MHz version ("B")
;    and change the Xtal oscillator from 4 MHz to 8 MHz.
;    Check the access speed of the EPROM and the RAM chips!
;    This makes *everything* twice as fast!!
; x) ...
;
;
; To do:
; (1) Add "paper tape" load and punch.  Perhaps can have a buffer for main
;     program download to decrease the binary loader delay and then re-use the.
;     Use up and down loading from the PC, with time delay to indicate end of data.


;====================================================================================

	.module	pdp8e
	.radix	d
	.area	WORKPG	(PAG)
	.setdp	0,WORKPG


BINldr_txt	=	0	; 1 :: print BINloader text to identify received data
				; 0 :: do not print any text

RealCons	=	1	; 1 :: include the RealConsole!
				; 0 :: do not include the RealConsole!

  ;  DISK 'stuff'

ide2pdp		=	1	; 1 :: interface IDE disk to pdp8/e simulator
				; 0 :: no IDE disk attached
idetrc		=	0	; 1 :: output message for each IDE access (r/w)
				; 0 :: no trace output
RF08dev		=	1	; 0 :: DF32 disk  \   mutual exclusive : RF08 is sort
				; 1 :: RF08 disk  /   of superset of the DF32 IOT's

DEFER_IO	=	1	; 1 :: use "deferred I/O" for hardware effects
				; 0 :: not


    ; CORESYS09  I/O routines, memory & hardware address locations


inch	=	0xC009		; get keyboard entry in [A], no echo
ino	=	0xC00C		; input character (8 bits)
OUTHLO	=	0xC01B		; output one hex char, left half-byte
OUTHRO	=	0xC01E		; output one hex char, right half-byte
OUTB2O	=	0xC021		; print data in [A] as 2 hexadecimal digits
outch	=	0xC024		; print character in [A] - not destroyed
OUT2CH	=	0xC027		; output 2 chars [A] then [B]
OUTABT	=	0xC030		; print character in [A], [B] times repeated
OUT2HS	=	0xC036		; print [X] as 2 hex followed by one space
OUT4HS	=	0xC039		; print [X] as 4 hex followed by one space
OUTS	=	0xC03C		; print 1 space

comand	=	0xC6C7		; command list scan routine
skpdlm	=	0xC769		; skip leading delimiters
mnumber	=	0xC841		; scan for a number
out1by	=	0xC8E7		; print 1 byte [X] in <dbcode> base
out2by	=	0xC8EF		; print 2 bytes [X] in <dbcode> base
comlst	=	0xC9B4		; CORESYS-09 command tables
getlin	=	0xCAAC		; form input line from entered characters
outchr	=	0xCBFB		; print [A] - count char/line
docrlf	=	0xCC37		; output CR/LF to console (reset char-out count)

bufbeg	=	0x21A4		; input line start of buffer
comadr	=	0x21A8		; address of beginning of command lists
synptr	=	0x21AA		; input line character pointer for good syntax
linptr	=	0x21AC		; command line input pointer
bolflg	=	0x21AE		; "beginning of line flag"
ibcode	=	0x21B0		; input base (1=hex, 2=dec, 3=oct, 4=bin)
dbcode	=	0x21B1		; display base (1=hex, 2=dec, 3=oct, 4=bin)
nbrhi	=	0x21B3		; scanned number
cplcnt	=	0x21C6		; "characters per line" count

.firq	=	0x21F6		; FIRQ vector location in monitor RAM
.irq	=	0x21F8		; IRQ vector
.swi	=	0x21FA		; SWI vector
.nmi	=	0x21FC		; NMI vector


   ; parallel ports (MC6821, PIA)

PIADA	=	0x00E0		; PIA data A register
PIACA	=	PIADA+1		;     control A register
PIADB	=	PIADA+2		;     data B register
PIACB	=	PIADA+3		;     control B register

   ; PIA [port A] ConTroL register bitmask
   ; - during execution reconfigured to input/output port
   ; - CA2=output, IRQ on CA1 enabled/disabled

PACTLI	=	0b00111100	; CA2="1", buffer LS245 -> input
PACTLJ	=	0b00111101	; as PACTLI, but CA1 irq enabled
PACTLO	=	0b00110100	; CA2="0", buffer LS245 -> output
PACTLQ	=	0b00110101	; as PACTLO, but CA1 irq enabled
DIRBIT	=	0b00000100	; access data-in/out or data-dir
CA2BIT	=	0b00001000	; CA2 bit: 245-buffer ctrl in/out
				; 0 :: output, 1 :: input
IOADDR	=	0b00111111	; bit [5:0] - board & port select address bits
RDWR  	=	0b01000000	; bit [6] - RD/WR direction:  0::read   // 1::write
DIS3S 	=	0b10000000	; bit [7] - LS138's control:  0::enable // 1::disable


   ; PIA [port B] ConTroL register bitmask
   ; - always configured as output port
   ; - CB2=output set to "0" (buffer LS245 -> output)
   ; - IRQ on CB1 disabled/enabled

PBCTLO	=	0b00110100	; CB1 irq disabled
PBCTLQ	=	0b00110101	; CB1 irq enabled


   ; serial port (MC6850, ACIA)

ACIACS	=	0x00F0		; ACIA control & status
ACIADA	=	ACIACS+1	; ACIA data I/O
ACIANI	=	0b00010001	; /16, 8N2, receiver IRQ disabled
ACIAI	=	0b10010001	; /16, 8N2, receiver IRQ enabled
RTSBIT	=	0b01000000	; RTS off / on bit  


    ; ASCII constants

CR	=	0x0D		; Carriage return
LF	=	0x0A		; Line feed
CtrlC	=	0x03		; Control C
CtrlT	=	0x14		; Control T (trigger new trace spec request)
Space	=	0x20		; Space
BS	=	0x08		; Backspace
XON	=	0x11		; Data link "on"  prompt (DC1 or ^Q)
XOFF	=	0x13		; Data link "off" prompt (DC3 or ^S)


    ; pdp8/e simulator program equates

Opmask	=	0b00001110	; Upper byte opcode mask
Wmask	=	0b00001111	; Upper byte memory data mask
Wsize	=	0o7777		; Memory data mask
Fsize	=	0o10000*2	; 4096 words/memory location
Offset	=	0b01111111	; Page offset bits
LoPage	=	0b10000000	; Current page (low byte)
Lmask	=	0b00010000	; Isolate link
ACmask	=	0b00001111	; Drop link
IOTmask	=	0b01111110	; Isolate IOT address
LtoC	=	0b11110000	; Add to link (moves L bit to carry)
IOsetup	=	8		; IO device setup ("impossible" IOT)
IOreset	=	9		; Clear/CAF

Ubit	=	0b10000000	; Field number bit masks
Fbits	=	0b00000111
Dbits	=	0b00000111
Jdelay	=	0b10000000	; JMP/JMS interrupt inhibit
UserI	=	0b01000000	; User Interrupt bit
SetIO	=	0b0010000000000000	; Start bit for interrupt flags


    ; Simulated PDP-8 core memory.
    ; Maximum theoretical is 64 Kbytes, which is the full 6809 addressing space.
    ; Originate RAM at easily handled boundary.

RAMst	=	0x2300		; start of free RAM on Core Board ****
Field0	=	RAMst		; $2300-$42FF
Field1	=	Field0+Fsize	; $4300-$62FF
Field2	=	Field1+Fsize	; $6300-$82FF
Field3	=	Field2+Fsize	; $8300-$A2FF
;Field4	=	Field3+Fsize
;Field5	=	Field4+Fsize
;Field6	=	Field5+Fsize



    ; storage in WORKPG for the pdp8/e simulator

PC:	.rmb	2	; Program counter
AC:	.rmb	2	; Accumulator
MQ:	.rmb	2	; Multiplier quotient
Link:	.rmb	1
MD:	.rmb	2	; Present instruction
SR:	.rmb	2	; Front panel switch register
Temp2:	.rmb	2
Temp1:	.rmb	1
IOflgs:	.rmb	2	; Input output device interrupt requests
Uflag:	.rmb	2	; User interrupt bit
FLGset:	.rmb	2	; Temporary for device flag setup
Dcount:	.rmb	1	; Temporary for device #
IOctrl:	.rmb	1	; Temporary device IOT
Idelay:	.rmb	1	; ION delay count


    ; Following locations are used by the Memory Extension and Time Share.
    ; Keeping the User flag in the MSB allows a simple sign check.
    ; Have WorkD follow WorkUI and SaveD follow SaveUI so can transfer as a word.

WrkIp:	.rmb	2	; Working I field (also use X for pointer)
WrkDp:	.rmb	2
WorkUI:	.rmb	1
WorkD:	.rmb	1
SavIp:	.rmb	2	; saved during interrupt processing
SavDp:	.rmb	2
SaveUI:	.rmb	1
SaveD:	.rmb	1
BufIp:	.rmb	2	; loaded into working U and F by JMP or JMS
BuffUI:	.rmb	1


    ; flags for "display" and vectored jumps to emulation function
    ; - added to get rid of self-modifying code!

emufnc:	.rmb	2	; vector address is: Doinst, Chkdly, or Chkint


    ; flags from the input command line

verbos:	.rmb	1	; flag verbose identification (0=verbose off)
dsknum:	.rmb	1	; disk 'number' used during simulation (if IDE available)
core:	.rmb	1	; MOS or CORE "memory", MOS=0 (erase) // CORE=1 (not erased)
boottb:	.rmb	2	; selected bootstrap table
trcflg:	.rmb	1	; trace enabled flag


;  Explanation of the trace mechanism.
; -------------------------------------------------------------------------------
;  During initialisation the variables <trcflg> and <trcsts> are reset to $00.
;  VERY IMPORTANT: if no trace is needed, the simulator run at best performance.
;  There are two methods available to start the trace facility.
;  1. use the "-t" option on the command line when you start the pdp8 simulator
;  2. hit ^t (Control-t) when the simulator runs.
;  Both methods will sollicit for a trace specification, and then the simulator
;  starts, but at a lower performance, because after *every* simulated pdp8
;  instruction a check is done to see if the specified trace condition is met.
;  The first method allows trace control right from the start, the second method
;  is good when the simulator seems to "loop" and you want to investigate what
;  is going on. Of course, half the work is setting a good trace specification
;  that will generate the output that holds the clue ... :~)
;
;  Implementation details.
;  To implement the trace facility without affecting the original performance of
;  the simulator, the variable <conact> is (mis)used. The most significant bit of
;  conact indicates if the simulator must run or any other activity is wanted.
;  Using the most significant bit allows the fastest check for the "RUN" state;
;  all other states (any console interaction like Single Step) does not need the
;  best performance, as you are "interfering".
;  When both HALT and the S.STEP switch are OFF, and you press the CONT switch,
;  a test of the <trcflg> is done. If that flag is zero, the most significant bit
;  of conact [ACTrun] is set (run, best performance). If that flag is not zero,
;  [ACTrun] is not set, but the bit [ACTtrc].
;  When [ACTtrc] is set, the trace condition is tested after the simulation of
;  every pdp8 instruction.
;  When the "-t" option is specified on the command line <trcflg> is not 0, the
;  [TRCENA] bit in <trcsts> is set, indicating that the trace is enabled.
;  However, when "-t" was not specified at startup, but the trace facility is
;  started by "Control-t", no valid trace specification exists. The trick here
;  is to clear the [ACTrun] flag, and set <trcsts> to a special state, TRCNEW,
;  as response to the "Control-t" entry.
;  Because [ACTrun] is not set, the trace condition test routine is called, and
;  in this routine <trcsts> is tested for the value TRCNEW. TRCNEW is a special
;  value which has the most significant bit [TRCENA] clear.
;  When <trcsts> == TRCNEW, the simulator sollicits for a trace specification
;  and then immedately performs the first check on the specified trace.
;
;  <trcsts> - trace status
;  ___7_____6_____5_____4_____3_____2_____1_____0___
;  |     |     |     |     |     |     |     |     |
;  | ENA | ACT | R_M |   ACCESS  |   TRACE TYPE    |
;  -------------------------------------------------
;
TRCNEW	=	0x07	; [7]=0, request for new trace specification
TRCENA	=	0x80	; [7]=1, trace enabled
			;	 other bits have following meaning:

TRCACT	=	0x40	;	[6] = trace active
TRCREG	=	0x00	;	[5] = trace on register
TRCMEM	=	0x20	;	    = trace on memory
TRCACC	=	0x18	;     [4-3] = access type
TRCDAT	=	0x00	;	 00 = DATA
TRCRD	=	0x08	;	 01 = READ
TRCWR	=	0x10	;	 10 = WRITE
TRCCHG	=	0x18	;	 11 = CHANGED
TRCTYP	=	0x07	;   [2-1-0] = "trace" type
TRC_AL	=	0	;	000 : trace always
TRC_AF	=	1	;	001 : trace after
TRC_UN	=	2	;	010 : trace until
TRC_OY	=	3	;	011 : trace only
TRC_RT	=	4	;	100 : transition to REALTIME
TRC_RN	=	5	;	101 : RUN UNTIL
TRC_ST	=	6	;	110 : STEP
TRC_BF	=	7	;	111 : trace before
;
;
;  <trcext> - trace extension
;  ___7_____6_____5_____4_____3_____2_____1_____0___
;  |     |     |     |     |     |     |     |     |
;  | IDC | OCC |     |     |     |     |     |     |
;  -------------------------------------------------
;
TREIDC	=	0x80	; [7] Invert Datamatch Condition (normal=0 / invert=1)
TREOCC	=	0x40	; [6] OCCUR keyword specified, <trcocc> is valid
;
trcsts:	.rmb	1	; trace status byte
trcext:	.rmb	1	; trigger option extensions
trcmem:	.rmb	2	; memory address if trcsts[5] = 1 (2 bytes)
			; "register" specifier (x2) if trcsts[5] = 0 (in ms byte)
trctrg:	.rmb	2	; trace trigger data word (max 16 bits)
trcinf:	.rmb	2	; previous trigger specifier data (for "CHANGED" option)
trccnt:	.rmb	2	; step counter
trcdmp:	.rmb	1	; trace dump re-routing flag (==0 disabled , <>0 enabled)
trcbip:	.rmb	2	; trace dump buffer IN pointer
trcbop:	.rmb	2	;   "	 "	"   OUT pointer
trcocc:	.rmb	2	; trace event occurrence counter
trcirq:	.rmb	1	; save current irq flags during trace facility prompt
trctmp:	.rmb	2	; scratch
rf8inf:	.rmb	1	; flag: display RF08 "registers" on DMAR/DMAW (0::OFF / 1::ON)

    ; memory word to quick access the 'lamp values' for the different
    ; settings of the Indicator Selector Switch (ISS).  The words are
    ; kept "up-to-date" during the execution of the simulation at many
    ; locations to facilitate a reasonable correct RealConsole display.

savacc:	.rmb	1	; save location for updating the following data
rc_mad:	.rmb	2	; EMA + Memory Address
rc_ste:	.rmb	2	; ISS -> STATE
rc_bus:	.rmb	2	; ISS -> BUS


    ; local storage used in the subroutines

binptr:	.rmb	2	; [binary loader] absolute load location
bindat:	.rmb	2	; [binary loader] previous memory data
binsum:	.rmb	2	; [binary loader] checksum
binsv0:	.rmb	2	; [binary loader] last input data (low-byte)
binsv1:	.rmb	2	; [binary loader] present input data

Ibit:	.rmb	2	; [TTY] Interrupt flag bit
Icom:	.rmb	2	; [TTY] complement of flag bit
Istat:	.rmb	1	; [TTY] Interrupt enable (FALSE == not enabled)
			; [TTY] keep following two together:
TTIflg:	.rmb	1	; [TTY] character received flag (TRUE == received)
TTOflg:	.rmb	1	; [TTY] output character done flag (TRUE == done)
TTIdat:	.rmb	1	; [TTY] input character storage

issts2:	.rmb	2	; [ISSSTS] scratch word

_bin3:	.rmb	1	; [binwrd]
_bin4:	.rmb	1	; [binwrd]

_stp2:	.rmb	1	; formatted copy of Link
_stp3:	.rmb	1	; formatted User mode
_stp4:	.rmb	1	; true if interrupt system is on
_stp5:	.rmb	1	; true if interrupt requests blocked
_stp6:	.rmb	1	; true if have interrupt requests

_octp2:	.rmb	1	; [PrtOct] (print octal number) digit counter
_octp3:	.rmb	1	; [PrtOct] (print octal number) character counter
_octp4:	.rmb	1	; [PrtOct] (print octal number) data to print
_octp5:	.rmb	1	; [PrtOct] (print octal number)

read8:	.rmb	2	; entry of 4-digit octal number


	.if	RealCons
piadir:	.rmb	1	    ; indicator: PIA [port A] current data direction
FIRQsv:	.rmb	1	    ; save location, used *ONLY* local in FIRQ routine
rtssts:	.rmb	1	    ; RTS* pin status of ACIA
timcnt:	.rmb	1	    ; time-out interval setpoint value
timer:	.rmb	1	    ; running time-out interval value
RCRAM:	.rmb	LSTIRX	    ; assign index-referenced storage
procin:	.rmb	1	    ; current processing input port #

INBASE	=	RCRAM+INPORT  ; for easy reference

dlyexm:	.rmb	1	    ; EXAMINE toggle response control
dlydep:	.rmb	1	    ; DEPOSIT toggle response control
dlyclr:	.rmb	1	    ; CLEAR toggle response control
dlyal:	.rmb	1	    ; ADDRESS LOAD toggle response control
dlyeal:	.rmb	1	    ; EXTENDED ADDRESS LOAD toggle response control

booten:	.rmb	1	    ; bootstrap option possible flag
conact:	.rmb	1	    ; console action response
ACTrun	=	0b10000000  ; action = run
ACTstp	=	0b00000001  ;        = step instruction
ACTcyc	=	0b00000010  ;        = cycle instruction
ACTtrc	=	0b01000000  ;	     = enable pdp8 instruction execution trace

	.else

CCsave:	.rmb	1	    ; save interrupt mask
MONflg:	.rmb	1	    ; Monitor call flag

    ; Monitor flags

Console	=	0b10000000	; Call monitor (from keyboard ^S)
Break0	=	0b01000000	; Break point #0
Break1	=	0b00100000	; Break point #1
Break2	=	0b00010000	; Break point #2
Halt	=	0b00001000	; Halt instruction or HALT switch active
I_path	=	0b00000100	; Interrupt service
SnglStp	=	0b00000010	; Single stepping
Breaks	=	Break0+Break1+Break2

	.endif

    ; Binary loader interrupt routine data reception storage
    ; note:  the buffer is also used to store the TRACE BEFORE data

RDRin:	.rmb	2	    ; store into buffer pointer
RDRout:	.rmb	2	    ; retrieve from buffer pointer
RDRblo	=	0xBC00	    ; buffer start address *** hard coded ***
RDRbhi	=	0xC000	    ; buffer end (not inclusive)
TRCblo	=	RDRblo	    ; allocate same buffer space
TRCbhi	=	RDRbhi

	.if	BINldr_txt
RDRddd:	.rmb	1	    ; count printed "d" for data block
RDRltr:	.rmb	1	    ; count printed "." for leader/trailer block
RDRcnt:	.rmb	1	    ; count # printed characters on one line
	.endif


	.if	ide2pdp

accide:	.rmb	1	; access IDE flag: handshake between IDE and FIRQ

idectl:	.rmb	1	; current written IDE control byte
iodatd:	.rmb	1	; current I/O databus active data direction
idesv1:	.rmb	1	; byte scratch (-local- inside subroutine!)
idesv2:	.rmb	2	; word scratch (-local- inside subroutine!)

seccyl:	.rmb	2	; disk parameter: # sectors / cylinder
sectrk:	.rmb	2	; disk parameter: # sectors / track

CmdPkt:	.rmb	2	; source/destination data buffer address
	.rmb	4	; LBA[3:0]



;  * 3-cycle data break registers in pdp8/e memory space for DF32 and RF08

wc3cdb	=	Field0+0o7750*2	; 3-cycle data break Word Count register
ca3cdb	=	Field0+0o7751*2	; 3-cycle data break Current Address register


;  * RF08 / DF32 "registers"

df_dma:	.rmb	2		; disk memory address 12-bits, [11:0]
df_dmx:	.rmb	2		; disk address extension
				; DF32: 5-bits [8:4] , RF08: 8-bits [11:4]
				; bits [3:0] always zero (easy combined with df_dma)
	.if	RF08dev
rf_irq:	.rmb	2		; RF08 interrupt flags
rf_sts:	.rmb	2		; RF08 status flags
	.else
df_irq:	.rmb	1		; DF32 interrupt flags
df_sts:	.rmb	1		; DF32 status flags
	.endif
df_ema:	.rmb	1		; core memory address extension 3-bits [3:1]
DFirqF:	.rmb	2		; allocated DF32/RF08 interrupt flag bit in IOflgs
df_rts:	.rmb	1		; flag for DMAR/DMAW result:  0=ok / !0=failed
secnum:	.rmb	2		; RF08/DF32 'sector' number on IDE disk
secflg:	.rmb	1		; current sector status flags
secoff:	.rmb	4		; sector offset to disk # used in emulation


;  * RF08 / DF32 interrupt and status flags
;		 ("-" :: not implemented  //  "+" :: implemented)

	.if	RF08dev		; RF08 <rf_irq> interrupt flags
				; upper 4 bits of 12-bit word
rf_IER	=	0b00000001	; + Interrupt Enable Error

				; lower 8 bits of 12-bit word
rf_ICL	=	0b10000000	; - Interrupt Enable Clock
rf_ICO	=	0b01000000	; + Interrupt Enable Complete

	.else			; DF32 <df_irq> interrupt flags
df_TRC	=	0b10000000	; + transfer completed flag
df_NED	=	0b00000001	; + non-existent disk flag
	.endif


	.if	RF08dev		; RF08 <rf_sts> status flags
				; upper 4 bits of 12-bit word *plus* bit15:
rf_TRC	=	0b10000000	; + Transfer Complete flag (internal housekeeping)
				;   bit[15]. The 12 bits, bits[11:0], used by pdp8
rf_PCA	=	0b00001000	; - PhotoCell Sync Mark
rf_DRE	=	0b00000100	; - Data Request Enable
rf_WLS	=	0b00000010	; - Write Lock Status
rf_EIE	=	0b00000001	; + Error Interrupt Enable

				; lower 8 bits of 12-bit word
rf_PIE	=	0b10000000	; - PhotoCell Interrupt Enable
rf_CIE	=	0b01000000	; + Completion Interrupt Enable
rf_FLD	=	0b00111000	; + Field bits (core address extension)
rf_DRL	=	0b00000100	; - Data Request Late
rf_NXD	=	0b00000010	; - NoneXistent Disk
rf_PER	=	0b00000001	; - Parity Error

	.else			; DF32 <df_sts> status flags
df_DRL	=	0b00000100	; - data request late
df_EWL	=	0b00000010	; - write lock / non-existent disk
df_PER	=	0b00000001	; - parity errors
	.endif


;  IDE disk <secflg> sector flags

dfNIBF	=	0b10000000	; flag: "RF08/DF32 'sector' not in IDE buffer"
dfDIRT	=	0b01000000	; flag: buffer changed, must be written to IDE disk


;  * RF08 / DF32 maximum number of sectors with the maximum of 4 disks "installed"

	.if	RF08dev
Sec4dsk	=	0x1000		; 4 RS08 disks --> 1024 kW / 256 bytes/sector = 2**12
	.else
Sec4dsk	=	0x0200		; 4 DS32 disks --> 128 kW / 256 bytes/sector = 2**9
	.endif




;  The Data Break facility of the pdp8 is actual what we call nowadays DMA !
;  That means that a real pdp8 can execute instructions *while* the data transfer
;  from/to disk is in progress. In fact, instruction execution proceeds, whenever
;  the device is unable to present data in the next cycle.
;
;  OS/8 makes use of this, and executes instructions that will be overwritten when
;  the disk transfer (in this case READ) is completed.
;  The implementation (as it was) does the READ operation when the instruction is
;  decoded (DMAR in this case), and the next instruction is executed when the data
;  transfer is completed. This scheme overwrites the instructions that should have
;  been executed before the transfer finishes.
;  OS/8 issues the DMAR, does some extra instructions, and then checks the device
;  status, because it knows that the device cannot complete the I/O in zero time.
;
;  To make sure that "some" pdp8 instructions are executed after the DMAR starts
;  in the simulator, there are two possible implementations.
;  1) "interleaved I/O"
;     after the I/O command is executed, some mechanism "interleaves" the execution
;     of the following pdp8 instructions. After every 8 (for example) instructions
;     a small part of the I/O transfer is done.
;     This approach mimics the Data Break (DMA) quite accurate, but is difficult
;     to implement.
;  2) "deferred I/O"
;     When the I/O transfer IOT is executed, it only sets a global flag ("defer"),
;     does the I/O initialisation stuff (so that test IOTs operate correct), and
;     sets a counter. After every executed (following) pdp8 instruction the count
;     is decremented. When the count reaches zero, the actual I/O transfer code
;     is executed. When not any device has an active count, the global "defer" flag
;     is cleared.
;     In the main pdp8 instruction decode loop a test for "defer" is added.
;     Remark: in the [not "deferred I/O"] code, the actual read/write is part of
;	      the IOT (DMAR or DMAW). In the ["deferred I/O"] code that actual
;	      read/write code becomes a subroutine, and returns a OK/NOK status.


;  Data Break (DMA) facility implementation ["deferred I/O"]

	.if	DEFER_IO

defer:	.rmb	1		; global "deferred I/O" active flag

MAXDEV	=	1		; only one DMA device for now (RF08)
				; RF08 == device index 0

deftab:	.rmb	MAXDEV*4	; "deferred I/O" administration table
;
;   [0]  - deferred_active flag (0 = not active)
;   [1]  - deferred_count
;  [2:3] - I/O transfer routine address

DEFCNT	=	10		; # instructions deferred actual I/O transfer

	.endif



	.area	DATAPG

   ; IDE disk data buffer

IDEbuf:	.rmb	512		; room for 256 words

	.endif

;====================================================================================
;====================================================================================


	.area	PDP8E	(ABS,OVR)

	; Start the pdp8/e emulator

X_PDP8:	ldd	#WORKPG	    ; set all var's in 6809 'zero-page'
	tfr	a,dp
	jsr	strtup	    ; process command line input flags
	jsr	IDinfo	    ; issue identification message(s)

	tst	*core	    ; MOS or CORE memory ?
	bne	10$	    ; CORE -> do not erase
	bsr	RAMclr	    ; MOS  -> clear (existing) "pdp8/e memory"
10$:
	.if	RealCons
	jsr	initRC	    ; initialize RealCons hardware
	.else
	lda	#Console    ; call Monitor on start
	sta	*MONflg
	.endif

	jsr	PDPld	    ; load at least one binary file

	.if	RealCons
	ldd	#0o200	    ; set a common default start PC
	.else
getpc1:	ldx	#askpc
	jsr	pdatao	    ; ask for start PC
	ldb	#4	    ; get 4 octal digits max
	jsr	getoct
	bvs	getpc1	    ; entry aborted - try again
	bcc	getpc2	    ; good entry, new PC accepted
	ldd	#0o200	    ; <CR> - set default start PC (0200)
getpc2:
	.endif

	std	*PC	    ; Console LOAD ADDR can change this default

	clr	*IOflgs	    ; make sure Jdelay is cleared in IO flags
			    ; ClrSys will clear the other flag bits
	lda	#IOsetup    ; set up IO devices (KL8-E turns on IRQ)
	bsr	ClrSys
	clra		    ; IF0-1-2 and DF0-1-2: set to Field zero
	sta	*WorkUI	    ; save
	sta	*WorkD
	sta	*BuffUI	    ; same as EXTendeD ADDRess LOAD on panel
	lsla		    ; get field pointer
	ldu	#GetFld
	ldx	a,u
	beq	Flderr	    ; no memory allocated
	stx	*WrkIp
	stx	*WrkDp
	stx	*BufIp
	clra		    ; set the Save Field registers to field 0
	sta	*SaveUI	    ; NOTE: Real PDP-8 may have random value
	sta	*SaveD
	ldd	0,u	    ; (keep X set to the Instruction field)
	std	*SavIp
	std	*SavDp

    ; initialize display
    ; the routine "initRC" enabled the FIRQ that processes the switches/keys.
    ; wait a little to obtain stable switch/key states in RCRAM memory array.
    ; Switch Register usually at start address, as standard startup sequence
    ; is Address Load, Clear and Continue.

	.if	RealCons
	ldb	#3	; delay 300 mS
inidly:	ldy	#14285	; delay for 100 mSec (at 1 MHz CPU clock)
1$:	leay	-1,y
	bne	1$
	decb
	bne	inidly
	.else
	ldd	*PC	; set Switch Register to start address
	std	*SR
	.endif

	jmp	Emulat	; ----->>>  start  pdp8/e  simulation



;  * RAMclr - clear existing RAM fields

;    The memory locations are filled with HLT (7402) or zeroes (0000).
;    The available memory is indicated by the GetFld table.
;    registers destroyed: [A,B,X,Y,U]

RAMclr:	ldx	#clrcor	; print progress (if verbose mode)
	jsr	vbprnt
	lda	#8	; max is 8 blocks of 8kbyte RAM (4k word)
	pshs	a	; push block counter on stack
; #->	ldy	#0o7402	; fill with 7402 (HLT
	ldy	#0	; fill with 0000 (zeroes, AND)
	ldu	#GetFld	; base of available memory table
ramnxF:	ldx	,u++	; get base address of 8kbyte block
	beq	ramnoF	; non-existent -
	ldd	#Fsize	; set up size (words) of one block
ramclF:	sty	,x++	; store HLT or zero in entire block
	subd	#2
	bne	ramclF	; until block filled/erased
ramnoF:	dec	0,s	; all blocks available checked?
	bne	ramnxF	; n -
	ldx	#okmsg	; y - print progress (if verbose mode)
	jsr	vbprnt
	puls	a,pc


    ; Field location array.
    ; If location == 0 the field does not exist. Field0 must always exist.

GetFld:	.fdb	Field0	; -> must exist
	.fdb	Field1	; -> minimum RAM requirement for OS8
	.fdb	Field2
	.fdb	0	; Field3
	.fdb	0	; Field 4 (system RAM)
	.fdb	0	; Field 5 (system RAM)
	.fdb	0	; EPROM - monitor space
	.fdb	0	; EPROM - monitor space


    ; Memory not allocated for the specified field error print out

Flderr:	jsr	Errpnt
	.fdb	NoFLD



    ; Panel "start" or CAF. Clears the AC, MQ and link. Does not affect the data and
    ; instruction field settings.  Sets the Time Share to executive mode (U= 0).
    ; Peripheral flags cleared, but interrupt requests (usually) enabled.
    ; Interrupt system is off and ION and JMP/JMS inhibits cleared.
    ; On exit D == 0.

Clear:	lda	#IOreset	; reset flags. Enable interrupt requests
ClrSys:	ldb	*IOflgs
	andb	#Jdelay		; clear flags, keep Jdelay
	stb	*IOflgs
	clr	*IOflgs+1
	jsr	IOset		; enter with [A] = IO setup flag
	bne	clrerr
	ldd	#Doinst
	std	*emufnc		; set vector to processing routine
	clra
	clrb
	std	*AC		; clear "registers"
	std	*MQ
	std	*MD
	std	*rc_mad
	std	*rc_ste
	std	*rc_bus
	sta	*Link
	rts
clrerr:	jsr	Errpnt
	.fdb	Clrmsg







; The ISZ and other skip instructions conditionally skip the next
; instruction in sequence.  The ISZ is commonly used to increment a loop
; counter and skip if done, and it is also used as an general increment
; instruction, either followed by a no-op or in contexts where it is known
; that the result will never be zero.

    ; ISZ increments the data at the address and skips the next instruction
    ; if it becomes 0. AC and Link not affected.
    ; Routine moved here so the code can fall through into the PC skip.

ISZsub:	jsr	GetAdr		; clears FETCH flag in STATE
	beq	NextI		; data field does not exist
				; NOTE: if field does not exist, memory read == 0000?
				;       if so:  "NextI" should be "PCskip"
	lslb
	rola			; translate to 16-bit "real memory" address
	leau	d,u
	ldd	*rc_ste
	ora	#LD_exe		; set EXECUTE flag in STATE
	orb	#LD_dac+LD_mdd	; set DATA-CONT and MD-DIR flag in STATE
	std	*rc_ste
	ldd	,u		; get data from address
	addd	#1		; +1
	anda	#Wmask		; prevent 12-bit overflow
	std	,u
; ---	cmpd	#0		; data result == 0000 ?  [ccZ set by previous std]
	bne	NextI		; n -

;   skip next instruction

PCskip:	ldd	*PC
_skip:	addd	#2
	anda	#Wmask		; prevent 12-bit overflow
	std	*PC		; code duplicated from _savpc for performance
	jmp	[emufnc]	; jump target can change

;   next instruction routine

NextI:
;	lda	*PC
;	ora	*AC		; ***  TEMPORARY CHECK ROUTINE  ***
;	ora	*MQ		; Check data at last step to see
;	ora	*MD		; if have any extra bits set
;	bita	#0b11110000	; all four high bits must be 0
;	bne	b4err
;	lda	*Link
;	bita	#Lmask^0xFF
;	beq	GetI
;b4err: jsr	Errpnt
;	.fdb	Bitmsg


GetI:	ldd	*PC		; get the PC
				; ***  END CHECK  ***

_nxtni:	addd	#1		; update to next
_savpc:	anda	#Wmask		; prevent 12-bit overflow
	std	*PC


    ; Vectored jump to one of the following routines, depending on interrupts
    ; are enabled (Chkint) or interrupts disabled (Doinst), set-up in <emufnc>.
    ; Chkdly is an intermediary state to delay the transition into Chkint.

Emulat:	jmp	[emufnc]	; jump target can change

;  ________________________________________________________________________________
;  ################################################################################
;  ================================================================================



;  ##===========================================##
;  ##						##
;  ##    Emulation "target" routine:  Chkdly    ##
;  ##						##
;  ##===========================================##

    ; Process ION delay count. When the counter Idelay is 0, the vectored jump
    ; is set to Chkint to interrupt check and then fall through into that check.

Chkdly:	dec	*Idelay		; ION delay finished?
	bne	Doinst
	ldd	#Chkint		; change Emulat object
	std	*emufnc		; set vector to processing routine



;  ##===========================================##
;  ##						##
;  ##    Emulation "target" routine:  Chkint    ##
;  ##						##
;  ##===========================================##

    ; Process interrupts.
    ; The interrupts have to be enabled and the ION delay (1 instruction)
    ; and JMP/JMS delay (due to I field load) finished.

Chkint:	ldd	*IOflgs		; get device flags
	ble	Doinst		; check for request (exit if none or J delay)
	ldd	#Doinst		; turn off interrupts
	std	*emufnc		; set vector to processing routine
	ldd	*WorkUI		; save U, I and D
	std	*SaveUI
	ldd	*WrkDp
	std	*SavDp
	stx	*SavIp
	ldx	GetFld		; set F and D to field 0, and User flag to exec
	stx	*WrkIp
	stx	*BufIp
	stx	*WrkDp
	ldd	*PC		; save return address at location 0
	std	,x
	ldd	#0
	std	*WorkUI		; set WorkUI and WorkD to 0
	stb	*BuffUI		; set BuffUI also to 0
	incb			; move PC to next instruction
	std	*PC
	lslb
	bra	_getI



;  ##===========================================##
;  ##						##
;  ##    Emulation "target" routine:  Doinst    ##
;  ##						##
;  ##===========================================##

    ; go to next instruction

Doinst:	ldd	*PC	; restore PC data
	lslb		; multiply by 2 for relative location
			; (16-bit word :: one PDP8 instruction)
	rola
_getI:	ldd	d,x	; get instruction
	std	*MD

    ; set STATE for new instruction
    ;
    ;   ==> copy IR0-IR1-IR2 from MD
    ;   ==> set the FETCH flag bit
    ;   ==> clear DEFER, EXECUTE, MD-DIR, DATA-CONT
    ;   ==> SW flag bit reflects position of the SW console switch
    ;   ==> clear PAUSE, BRK-PROG, and BRK

				; [A] = ....012i             (from MD)
	lsra			; [A] = .....012  [Cc] = ind
	lsra			;       ......01  [Cc] = IR2
	rora			;       2......0  [Cc] = IR1
	rora			;       12......  [Cc] = IR0

	ldb	#LD_fet		; FETCH always set, DEFER & EXECUTE clear
	bcc	_ste1		; test IR0 in [Cc], clear -
	orb	#LD_ir0		; IR0 bit is set, set -
_ste1:	stb	*rc_ste

	.if	RealCons
	ldb	*INBASE+RD8e_1	; get HALT, SSTEP, and SW
	bitb	#SW_sw		; SW switch active ?
	beq	_ste2		; - clear
	ora	#LD_swt		; - set: add to IR1-IR2 bits
	.endif
_ste2:	sta	*rc_ste+1	; DATA-CONT, MD-DIR always clear
				; PAUSE, BRK-PROG, BRK always clear


    ; next instruction ready for decode.
    ; State checks; implement flags to execute a check or bypass the check
    ; !!! NOTE : preserve [X] (= instruction field location) !!!


;     _______________________________________________________
;  ###							     ###
;  ###  process the switches and the toggles of the console  ###
;  ###_______________________________________________________###


	.if	RealCons

   ; [conact] flags   -- mutual exclusive --
   ;  ___ ___ ___ ___ ___ ___ ___ ___
   ; | 7 | 6 | 5 | 4 | 3 | 2 | 1 | 0 |
   ; |___|___|___|___|___|___|___|___|
   ;
   ;   * running                   * cycle step request
   ;       * pdp8 exec trace   * instruction step request
   ;
   ; If all bits are clear none of the operations is executed.
   ; All bits causes the code to loop checking the console, which will
   ; set one of the flag bits [7-6-1-0] eventually.
   ; Bit 6 only indicates that the registers -may- be printed every pdp8
   ; instruction. The flag is set by the -t switch at startup.
   ; For efficiency reasons, the ACTrun flag {running} is bit 7.
   ; Bit 7 set allows the simulator to run "at full speed", that is, with
   ; minimal overhead instructions. Tracing is enabled when bit 6 is set and
   ; then, if the trace condition is met, the pdp8 instruction is executed
   ; with a register dump on the console.
   ; Implementation: whenever a change from Step/Cycle/Nop action
   ; is requested to {running} action, then [trcflg] is tested.
   ;    if [trcflg] == 0 --> CONACT := 10000000
   ;       [trcflg] != 0 --> CONACT := 01000000
   ;
   ; No flags are set at startup. This causes the simulator to loop, waiting
   ; for initiation from the console. That can be one of the following:
   ; * HALT + CONT      --> execute a single instruction
   ; * HALT + SING STEP --> execute a single instruction plus register dump
   ; * CONT             --> run OR trace if -t specified at startup)
   ;
   ; "trace" can be stopped by HALT or SING STEP. Pressing CONT will give the
   ; expected behaviour, execution of a single pdp8 instruction. When CONT is
   ; pressed while HALT nor SING STEP is active, "trace" is active again.
   ; To get "run" mode active you must start the sim *without* the -t option.

	lda	*conact		; console action response needed ?
; --	bita	#ACTrun		; "run" ?
; --	bne	Stpxit
	bmi	Stpxit		; y - (max efficiency: ACTrun at bit 7)

;    "full speed" path check complete.
;    Check Cycle / Step request before testing the trace flag. This order is
;    needed otherwise they would be ignored when the trace facility is active!

	bita	#ACTcyc+ACTstp	; n - check for Cycle or Step request from console
	beq	Stptrc		; no Cycle or Step entry, check if trace is enabled
	clr	*conact		; clear console Step or Cycle *request*, drop Trace
	bita	#ACTcyc		; test for "Cycle" or "Step" instruction
	beq	Stpxit		; -> not "Cycle" => must be "Step"
	bra	Stpreg		; -> "Cycle" == "Step" plus dump of registers

Stptrc:	bita	#ACTtrc		; "trace" ?
	beq	Emulat		; n - no active state :: await console action
	jsr	trcchk		; y - check trace condition
	beq	Stpxit		;     condition not met -> continue execution

Stpreg:	jsr	Pntreg		; print / dump registers and info

Stpxit:
	.if	DEFER_IO
	lda	*defer		; check for "deferred I/O" active
	beq	100$		; n -

	pshs	x		; y - check for "deferred I/O" time-outs
	clra			;     and process when timed out
10$:	ldx	#deftab
	tfr	a,b		;     for ( i=0 to MAXDEV )
	lslb			;     {
	lslb
	abx
	tst	,x		;	if (deftab[i].active == TRUE)
	beq	20$		;	{
	dec	1,x		;	  deftab[i].count--
	bne	20$		;	  if (deftab[i].count == 0)
				;	  {
	clr	,x		;	    deftab[i].active := FALSE
	pshs	d,x
	jsr	[2,x]		;	    call ( deftab[i].IO_tfr )
	puls	d,x
	lbne	Flderr		;	    return status == NOK
				;	  }
				;	}
20$:	inca
	cmpa	#MAXDEV
	bne	10$		;     }
	clr	*Temp1		;     any_active := FALSE  // local flag
	clra			;     
30$:	ldx	#deftab
	tfr	a,b		;     for ( i=0 to MAXDEV )
	lslb			;     {
	lslb
	tst	b,x		;	if (deftab[i].active == TRUE)
	beq	40$
	inc	*Temp1		;	   any_active := TRUE
40$:	inca
	cmpa	#MAXDEV
	bne	30$		;     }
	lda	*Temp1		;     defer := any_active
	sta	*defer
	puls	x
	.endif

100$:	lda	*MD		; get instruction opcode bits
	anda	#Opmask		; isolate them
	ldu	#Ops		; already "opcode*2" for word-sized table index
	jmp	[a,u]		; execute pdp8 instruction --->



	.else

	tst	*MONflg		; Monitor called if MONflg != 0
	bne	PDPmon
Stpxit:	lda	*MD		; get instruction opcode bits
	anda	#Opmask		; isolate them
	ldu	#Ops		; already "opcode*2" for word-sized table index
	jmp	[a,u]		; execute instruction --->


    ; Enter monitor if:
	; - HLT, keyboard break (^S)
	; - or single stepping
	; - breakpoint(s) became true (not implemented, future use?)

    ; NOTES:
	; (1) If HLT: Was a direct jump and special processing is needed.
	; (2) On entry, may want to drop single stepping.

    ; Process any breakpoints and then give a menu.
    ; Include all the usual stuff, such as:
	; Print memory
	; Set various breaks:
	;  - Memory location / Memory data / Instruction cycles
	; Load switch register (for OSR) / Single Step
	; Load address / Load extended address / Clear / Continue

    ; Start Monitor:
	; Presently only monitor startup (from a keyboard ^S call) and
	; single step flags are serviced.
	; In order to allow TTY output during stepping the interrupts
	; are off in this mode.

    ; Monitor flags:
    ;  ___ ___ ___ ___ ___ ___ ___ ___
    ; | 7 | 6 | 5 | 4 | 3 | 2 | 1 | 0 |
    ; |___|___|___|___|___|___|___|___|            label:
    ;                                             ---------------------
    ;                               * not used
    ;                           * single step      SnglStp
    ;                       * interrupt service    I_path
    ;                   * halt instruction         Halt
    ;               * breakpoint #2                Break2 (future use?)
    ;           * breakpoint #1                    Break1 (future use?)
    ;       * breakpoint #0                        Break0 (future use?)
    ;   * call monitor (from keyboard Ctrl-S)      Console


PDPmon:	lda	*MONflg	        ; check for keyboard call
	bita	#Console
	beq	Single	        ; branch if not (default)
	anda	#Console^0xFF   ; remove keyboard call
	ora	#SnglStp	; assume want single stepping
	sta	*MONflg
	tfr	CC,a		; save condition codes (IRQ flags)
	sta	*CCsave
	orcc	#0x10		; disable IRQ interrupt
	ldx	#SSmsg
	jsr	pdatao		; print options & fall through to "Single"


    ; Print system data and then wait for
    ;   - Space to continue
    ;   - K to enter keyboard character
    ;   - CR to exit stepping (start run)
    ;   - 1..6 for register display in ASCII artwork

Single:	jsr	Pntreg		; show system state
_sschk:	jsr	inch		; await keyboard command entry
	cmpa	#CR
	bne	Chkspc
	lda	*MONflg		; drop single stepping
	anda	#SnglStp^0xFF
	sta	*MONflg
	jsr	docrlf		; clear up screen a bit
	lda	*CCsave		; restore interrupts
	tfr	a,CC
	bra	_ssxit

Chkspc:	cmpa	#Space
	beq	_ssxit		; next step
	cmpa	#'?
	bne	_sskbd
	ldx	#SSmsg
	jsr	pdatao		; print command options
	bra	Single		; and show register data again

_sskbd:	cmpa	#'1
	blo	_sskbk		; not numeric
	cmpa	#'6
	bhi	_sskbk		; or out of numeric range (1-6)
	suba	#'0+1		; correct base for table
	ldx	#ISScmd
	lda	a,x		; get selection hardware switch value
	jsr	ShwReg		; display selected register in ACSII art
	bra	_sschk

_sskbk:	anda	#0b01011111	; drop lower case bit
	cmpa	#'K		; keyboard input?
	bne	_sschk		; n - ignore all other entries
	ldx	#Kbdmsg
	jsr	pdatao		; ask for keyboard input
	jsr	inch		; get it
	jsr	outch		; echo entry
	jsr	ttisub		; install it
	jsr	docrlf		; new line and continue

_ssxit:				; --> exit the monitor
	lda	*MONflg		; remove remaining breaks
	anda	#Break0|Break1|Break2|I_path^0xFF
	sta	*MONflg

	ldx	*WrkIp		; restore CPU registers
	jmp	Stpxit		; continue emulation


   ; translation table: command number entry to Indicator Selector Switch value
   ;	"<1>:STATE  <2>:STATUS  <3>:AC  <4>:MD  <5>:MQ  <6>:BUS"

ISScmd:	.fcb	ISSste,ISSsts,ISSac,ISSmd,ISSmq,ISSbus

	.endif


   ; pdp8/e opcodes routine start address table

Ops:	.fdb	ANDsub
	.fdb	TADsub
	.fdb	ISZsub
	.fdb	DCAsub
	.fdb	JMSsub
	.fdb	JMPsub
	.fdb	IOTsub
	.fdb	OPRsub


    ; Illegal instruction trap

Illins:	jsr	Errpnt
	.fdb	Insterr




    ; Access all IO devices. Enter with the IOP code in register [A].
    ; Only codes >= 8 used.
    ; If return with EQUAL status all existing devices were able to comply.
    ; On an error the routine takes an immediate exit and returns the
    ; exit code in [A].

IOset:	sta	*IOctrl
	ldu	#Device
	ldd	#SetIO		; set up initial IRQ flag bit available
	std	*FLGset
	lda	#64		; init device count
	sta	*Dcount
ioloop:	ldd	,u++		; does the device exist ?
	cmpd	#Nodev
	beq	iolxit
	lda	*IOctrl		; y - get IOT
	pshs	u		;     save [U] because device changes [U]
	jsr	[-2,u]		;     execute the setup routine
	puls	u		;     restore [U]
	bne	seterr		;     return immediate if an error occured
iolxit:	dec	*Dcount		; n - all devices processed ?
	bne	ioloop		;     n - go to next device
seterr:	rts			; return == 0 status if no error



    ; Allocate device interrupt flag bit in [D].
    ; Used by IO devices during initial setup.
    ; NOTE: FLGset must be initialised with the SetIO value.
    ; There are 14 flags available. Bit 15 (msb) is reserved for interrupt processing
    ; inhibit till JMP/JMS. The next bit (bit 14) is for the Time Share interrupt.
    ; Bits 13 thru 0 are available to be allocated to devices, like TTY and DF32.
    ;
    ; return : [D] = interrupt flag allocated for this device.
    ;	       if [D] == 0 there were no more interrupt flags available,
    ;			   and the calling device must return with NOT_EQUAL status.

GetFlg:	ldd	*FLGset		; get current allocated interrupt flags
	pshs	d
	lsra			; allocate and update the allocated interrupt flags
	rorb
	std	*FLGset
	puls	d,pc



    ; AND the AC with the data at the address.
    ; Link not affected.
    ; Indirect will use the data field. Non existent data field read as 0.

ANDsub:	bsr	GetAdr		; clears FETCH flag in STATE
	beq	dcaxit		; check for indirect and no field
_doand:	lslb
	rola
	sta	*savacc
	lda	*rc_ste
	ora	#LD_exe		; set EXECUTE flag in STATE
	sta	*rc_ste
	lda	*savacc
	ldd	d,u
	anda	*AC
	andb	*AC+1
	std	*AC

    ; active code from "NextI" *duplicated* here to save 6809-JMP instruction
    ; for speed improvement each time this pdp8/e instruction is executed.

	ldd	*PC		; update PC
	addd	#1
	anda	#Wmask		; prevent 12-bit overflow
	std	*PC
	jmp	[emufnc]



    ; TAD similar to AND, except that the Link is complemented by any overflow

TADsub:	bsr	GetAdr		; clears FETCH flag in STATE
	beq	TADnxI
	lslb
	rola
	sta	*savacc
	lda	*rc_ste
	ora	#LD_exe		; set EXECUTE flag in STATE
	sta	*rc_ste
	lda	*savacc
	ldd	d,u
	ora	*Link
	addd	*AC
	sta	*savacc		; temporary save: preserve Link bit
	anda	#ACmask
	std	*AC
	lda	*savacc		; get Link bit back
	anda	#Lmask
	sta	*Link

    ; active code from "NextI" *duplicated* here to save 6809-JMP instruction
    ; for speed improvement each time this pdp8/e instruction is executed.

TADnxI:	ldd	*PC		; update PC
	addd	#1
	anda	#Wmask		; prevent 12-bit overflow
	std	*PC
	jmp	[emufnc]



    ; DCA similar to AND. AC stored at the address and then cleared.
    ; Link is not affected.

DCAsub:	bsr	GetAdr		; clears FETCH flag in STATE
	beq	dcaxit
	lslb
	rola
	leau	d,u
	ldd	*rc_ste
	ora	#LD_exe		; set EXECUTE flag in STATE
	orb	#LD_mdd		; set MD-DIR flag in STATE
	std	*rc_ste
	ldd	*AC
	std	,u
dcaxit:	ldd	#0
	std	*AC

    ; active code from "NextI" *duplicated* here to save 6809-JMP instruction
    ; for speed improvement each time this pdp8/e instruction is executed.

	ldd	*PC		; update PC
	addd	#1
	anda	#Wmask		; prevent 12-bit overflow
	std	*PC
	jmp	[emufnc]




   ; Get the address referenced by the instruction.
   ;
   ;   call   : MD holds current instruction
   ;		 <rc_ste> STATE bits [FETCH] = set and [DEFER] = clear
   ;   return : [D] = effective address
   ;            If AND, TAD, ISZ or DCA an indirect will be from the
   ;            data field. Set U (6809) register for use by the above instructions.
   ;            indirect: U:= [WrkDp]  //  direct: U:=[WrkIp]
   ;		 STATE bit [FETCH] = clear and [DEFER] = set if indirect access
   ;   algorithm
   ;   If the I and Z bits indicate indirect page 0 and the location is (octal)
   ;   010 through 017, the location is a pre-increment autoindex access.
   ;   In this condition, the DATA-CONT light is also turned on.
   ;
   ;  11  10  9   8   7   6   5   4   3   2   1   0
   ;  ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ 
   ; |   |   |   |   |   |   |   |   |   |   |   |   |  i - indirect bit [0=direct,
   ; |_ _|__ |__ |__ |__ |___|___|__ |__ |__ |__ |__ |                    1=indirect]
   ; :           :   :   :                           :  z - page bit [0=page zero, 
   ; :  opcode   : i : z :           address         :                1=current page]
   ;                                                    address - 12-bit word in page

	.if	0

GetAdr:	ldd	*MD		; get instruction
	tstb			; test for page 0
	bmi	CurPg
	lsra			; test for indirect
	bcs	Pg0Ind
	lda	*rc_ste		; page 0, direct
	anda	#LD_fet^0xFF	; => clear FETCH bit in STATE (DEFER is clear)
	sta	*rc_ste
	clra			; set PC for page 0
	ldu	*WrkIp
	rts
Pg0Ind:	lda	*rc_ste		; page 0, indirect
	ora	#LD_def		; => set DEFER bit in STATE
	anda	#LD_fet^0xFF	;    clr FETCH bit
	sta	*rc_ste
	cmpb	#0o10		; check for autoindex range
	blo	Get0I
	cmpb	#0o17
	bhi	Get0I
	lda	*rc_ste+1
	ora	#LD_dac		; set DATA-CONT flag in STATE
	sta	*rc_ste+1
	clra
	lslb			; translate to 16-bit address
	leau	d,x		; get effective address
	ldd	,u
	addd	#1		; pre-increment
	anda	#Wmask		; prevent 12-bit overflow
	std	,u
	ldu	*WrkDp
	rts
Get0I:	clra
	lslb			; translate to 16-bit address
	ldd	d,x
	ldu	*WrkDp
	rts
CurPg:	andb	#Offset		; remove page bit (z)
	stb	*Temp1		; save current page offset
	lsra			; check for indirect
	bcs	CurInd
	lda	*rc_ste		; current page, direct
	anda	#LD_fet^0xFF	; => clear FETCH bit in STATE (DEFER is clear)
	sta	*rc_ste
	ldd	*PC		; get current page
	andb	#LoPage
	orb	*Temp1		; insert offset
	ldu	*WrkIp
	rts
CurInd:	lda	*rc_ste		; current page, indirect
	ora	#LD_def		; => set DEFER bit in STATE
	anda	#LD_fet^0xFF	;    clr FETCH bit
	sta	*rc_ste
	ldd	*PC
	andb	#LoPage
	orb	*Temp1
	lslb			; translate to 16-bit address
	rola
	ldd	d,x		; get indirect address
	ldu	*WrkDp
	rts

	.else
; ---------------------------------------------------------------------------------
GetAdr:	ldd	*MD		; get instruction
	lsra			; test for indirection
	bcs	InDir		; - indeed

    ; direct mode, current page or page 0

	lda	*rc_ste
	anda	#LD_fet^0xFF	; => clear FETCH bit in STATE (DEFER is clear)
	sta	*rc_ste
	tstb			; current page / page 0 ?
	bmi	DirCur
	clra			; page 0: [A]=00, [B:bit7]=clear --> [D] = eff.addr
	ldu	*WrkIp
	rts
DirCur:	andb	#Offset		; remove page bit (z)
	lda	*PC+1		; check bit 7 of PC
	bpl	10$		; -> [bit 7] :: clear
	addb	#0b10000000	; [bit 7] :: set, add to complete effective address
10$:	lda	*PC
	ldu	*WrkIp
	rts

    ; indirect mode, current page or page 0
    ; note: with the current page bit *set* (thus current page is accessed instead of
    ;       page 0), if any auto-index location [010-017] is referenced in page 0,
    ;       it should still auto-index!

InDir:	lda	*rc_ste		; indirection, so:
	ora	#LD_def		; => set DEFER bit in STATE
	anda	#LD_fet^0xFF	;    clr FETCH bit
	sta	*rc_ste
	tstb			; current page / page 0 ?
	bpl	InDir0
InDirC:	andb	#Offset		; remove page bit (z)
	lda	*PC+1		; check bit 7 of PC
	bpl	20$		; -> [bit 7] :: clear
	addb	#0b10000000	; [bit 7] :: set, add to complete effective address
20$:	lda	*PC		; check for current page == page 0 !
	bne	30$		; n -
	cmpb	#0o17		; might - PC[11:8] = 0, check for auto-index range
	bhi	30$
	cmpb	#0o10
	bhs	InDr0		; it is auto-index range [010-017] !
30$:	lslb			; translate to 16-bit address
	rola
	ldd	d,x		; get indirect address
	ldu	*WrkDp
	rts

InDir0:	cmpb	#0o17		; indirect/page 0 : check for autoindex range
	bhi	Get0I
	cmpb	#0o10
	blo	Get0I
InDr0:	lda	*rc_ste+1
	ora	#LD_dac		; set DATA-CONT flag in STATE
	sta	*rc_ste+1
	clra			; page 0, thus PC[11:8] = '0000'
	lslb			; translate to 16-bit address
	leau	d,x		; get effective address
	ldd	,u
	addd	#1		; pre-increment
	anda	#Wmask		; prevent 12-bit overflow
	std	,u
	ldu	*WrkDp
	rts
Get0I:	clra
	lslb			; translate to 16-bit address
	ldd	d,x
	ldu	*WrkDp
	rts
; ---------------------------------------------------------------------------------
	.endif




    ; Go to an address.
    ; Checks for the interrupt inhibit, and if set removes it.
    ; The I and U fields must be current with the Buffer *after* this instruction.
    ; An instruction field change always implies an interrupt inhibit. The effective
    ; address calculation is done with the current Ifield, and *after* that a change
    ; in I field may be activated; NOT before the effective address calculation!
    ; The Time Share instruction CUF may not set the inhibit,
    ; as it stops the interrupt.

JMPsub:	bsr	GetAdr		; calc eff. address (clears FETCH flag in STATE)
	sta	*savacc
	lda	*rc_ste
	ora	#LD_exe		; set EXECUTE flag in STATE
	sta	*rc_ste
	lda	*IOflgs		; now check for interrupt inhibit
	bpl	_dojmp		; no I field change ->
	anda	#Jdelay^0xFF	; remove inhibit
	sta	*IOflgs
	ldx	*BufIp		; update working I field pointers
	stx	*WrkIp
	lda	*BuffUI		; update User-Mode bit (also does I field number)
	sta	*WorkUI
_dojmp:	lda	*savacc

    ; active code from "_savpc" *duplicated* here to save 6809-JMP instruction
    ; for speed improvement each time this pdp8/e instruction is executed.

	anda	#Wmask		; prevent 12-bit overflow
	std	*PC
	jmp	[emufnc]




; The JMS instruction stores the return address in relative word zero of
; the subroutine, and the execution starts at relative word one.
; Subroutine return is done with an indirect JMP through the return address.
; Subroutines commonly increment their return addresses to index through
; inline parameter lists or to perform conditional skips over instructions
; following the call.

    ; Jump to a subroutine.
    ; Stores the return address at the subroutine entry point and continues at
    ; the next instruction.
    ; Interrupt inhibit and U, F and D handling as in the above JMP instruction.

JMSsub:	ldd	*rc_ste		; do LEDs first otherwise MD-DIR is ON very short
	ora	#LD_exe		; set EXECUTE flag in STATE
	orb	#LD_mdd		; set MD-DIR flag in STATE
	std	*rc_ste
	jsr	GetAdr		; calc eff. address (clears FETCH flag in STATE)
	std	*Temp2
	lda	*IOflgs		; check for interrupt inhibit
	bpl	_dojms		; no I field change ->
	anda	#Jdelay^0xFF	; remove inhibit
	sta	*IOflgs
	ldx	*BufIp		; update working I field pointers
	stx	*WrkIp
	lda	*BuffUI		; update User-Mode bit (also does I field number)
	sta	*WorkUI
_dojms:	ldd	*Temp2
	lslb
	rola
	leau	d,x
	ldd	*PC
	addd	#1
	anda	#Wmask
	std	,u		; store return address in field
				; *after* CIF I field change takes effect
	ldd	*Temp2

    ; active code from "_nxtni" *duplicated* here to save 6809-JMP instruction
    ; for speed improvement each time this pdp8/e instruction is executed.

	addd	#1
	anda	#Wmask		; prevent 12-bit overflow
	std	*PC
	jmp	[emufnc]





; The IOT instruction has the following form.
;         _ _ _ _ _ _ _ _ _ _ _ _
;        |1|1|0|_|_|_|_|_|_|_|_|_|
;        |     |           |     |
;        |     |   device  | op  |
;
; The IOT instruction specifies one of up to 8 operations on one of 64
; devices.  Typically (but not universally), each bit of the op field
; evokes an operation, and these can be microcoded in right to left
; order.  Prior to the PDP-8/E, there were severe restrictions on the
; interpretation of the op field that resulted from the fact that the
; operation was delivered as a sequence of IOP pulses, each on a separate
; line of the I/O bus.  Each line was typically used to evoke a different
; device function, so essentially, the operation 000 was always a no-op
; because it evoked no functions, and the code 111 evoked all three
; functions in series.
;
; As an example of the use of IOT instructions, consider the console
; terminal interface.  On early PDP-8 systems, this was always assumed to
; be an ASR 33 teletype, complete with low-speed paper tape reader and
; punch.  It was addressed as devices 03 (the keyboard/reader) and 04
; (the teleprinter/punch):
;         _ _ _ _ _ _ _ _ _ _ _ _
;        |1|1|0|_|_|_|_|_|_|_|_|_|
;              |0 0 0 0 1 1|0 0 1  - KSF - keyboard skip if flag
;              |0 0 0 0 1 1|0 1 0  - KCC - keyboard clear flag
;              |0 0 0 0 1 1|1 0 0  - KRS - keyboard read static
;
; The keyboard flag is set by the arrival of a character.  The KCC
; instruction clears both the flag and the accumulator. KRS ors the 8 bit
; input data with the low order 8 bits of AC.  The commonly used KRB
; instruction is the or of KCC and KRS.  To await one byte of input, use
; KSF to poll the flag, then read the byte with KRB.
;         _ _ _ _ _ _ _ _ _ _ _ _
;        |1|1|0|_|_|_|_|_|_|_|_|_|
;              |0 0 0 1 0 0|0 0 1  - TSF - teleprinter skip if flag
;              |0 0 0 1 0 0|0 1 0  - TCF - teleprinter clear flag
;              |0 0 0 1 0 0|1 0 0  - TPC - teleprinter print static
;
; The teleprinter flag is set by the completion of the TPC operation (as
; a result, on startup, many applications output a null in order to get
; things going).  TCF clears the flag, and TPC outputs the low order 8
; bits of the accumulator.  The commonly used TLS instruction is the or
; of TCF and TPC.  To output a character, first use TSF to poll the flag,
; then write the character with TLS.
;
; IOT instructions may be used to initiate data break transfers from block
; devices such as disk or tape.  The term "data break" was, for years,
; DEC's preferred term for cycle-stealing direct-memory-access data
; transfers.
;
; Some CPU functions are accessed only by IOT instructions.  For example,
; interrupt enable and disable are IOT instructions:
;         _ _ _ _ _ _ _ _ _ _ _ _
;        |1|1|0|_|_|_|_|_|_|_|_|_|
;              |0 0 0 0 0 0|0 0 1  - ION - interrupts turn on
;              |0 0 0 0 0 0|0 1 0  - IOF - interrupts turn off
;
; An interrupt is requested when any device raised its flag.  The console
; master clear switch resets all flags and disables interrupts.  In
; effect, an interrupt is a JMS instruction to location zero, with the
; side effect of disabling interrupts.  The interrupt service routine
; is expected to test the device flags and perform the operations needed
; to reset them, and then return using ION immediately before the indirect
; return JMP.  The effect of ION is delayed so that interrupts are not
; enabled until after the JMP.
;
; The instructions controlling the optional memory management unit are
; also IOT instructions.  This unit allows the program to address up to
; 32K of main memory by adding a 3 bit extension to the memory address.
; Two extensions are available, one for instruction fetch and direct
; addressing, the other for indirect addressing.


    ; Use the middle 6 bits of the instruction as the device address and
    ; the last 3 bits to indicate the action to be taken by the device.
    ; Access the device handler with the IOT code in [A].
    ; Codes above 7 are "impossible" and used for special access. In those
    ; conditions the handler will exit with a status code in [A].

IOTsub:	tst	*WorkUI		; in User Mode?
	lbmi	U_Mode		; y - User interrupt, do No-OPeration
	ldu	#Device
	ldd	*MD
	lsra
	rorb
	rorb
	andb	#IOTmask	; [B] = device (*2 for table indexing)
	lda	*MD+1
	anda	#0b111		; [A] = IOT code
	jmp	[b,u]



; OPR routine.
; One of the most complex due to the microcoded execution.
; A wide variety of operations are available through the OPR
; microcoded instructions:
;          _ _ _ _ _ _ _ _ _ _ _ _
; Group 1 |1|1|1|0|_|_|_|_|_|_|_|_|
;                  1                - CLA - clear AC
;                    1              - CLL - clear the L bit
;                      1            - CMA - ones complement AC
;                        1          - CML - complement L bit
;                                1  - IAC - increment <AC>
;                          1 0 0    - RAR - rotate <L,AC> right
;                          0 1 0    - RAL - rotate <L,AC> left
;                          1 0 1    - RTR - rotate <L,AC> right twice
;                          0 1 1    - RTL - rotate <L,AC> left twice
;                          0 0 1    - 6-bit "byte" swap AC (on pdp8/e)
;
; In general, the above operations can be combined by oring the bit
; patterns for the desired operations into a single instruction.
; If none of the bits are set, the result is the NOP instruction.
; When these operations are combined, they operate top to bottom in the
; order shownabove. The exception to this is that IAC cannot be combined
; with the rotate operations on some models, and attempts to combine
; rotate operations have different effects from one model to another
; (for example, on the PDP-8/E, the rotate code 001 means swap 6 bit
; bytes in the accumulator (BSW), while previous models took this to
; mean something like "shift neither left nor right 2 bits").
;
;          _ _ _ _ _ _ _ _ _ _ _ _
; Group 2 |1|1|1|1|_|_|_|_|_|_|_|0|
;                    1     0        - SMA - skip on AC < 0      \
;                      1   0        - SZA - skip on AC = 0       > OR group
;                        1 0        - SNL - skip on L != 0      /
;                    0 0 0 1        - SKP - skip unconditional
;                    1     1        - SPA - skip on AC >= 0     \
;                      1   1        - SNA - skip on AC != 0      > AND group
;                        1 1        - SZL - skip on L = 0       /
;                  1                - CLA - clear AC
;                            1      - OSR - or switches with AC
;                              1    - HLT - halt
;
; The above operations may be combined by oring them together, except
; that there are two distinct incompatible groups of skip instructions.
; When combined, SMA, SZA and SNL, skip if one or the other of the
; indicated conditions are true (logical or), while SPA, SNA and SZL skip
; if all of the indicated conditions are true (logical and).  When combined,
; these operate top to bottom in the order shown; thus, the accumulator may
; be tested and then cleared. Setting the halt bit in a skip instruction is
; a crude but useful way to set a breakpoint for front-panel debugging.
; If none of the bits are set, the result is an alternative form of no-op.
;
;          _ _ _ _ _ _ _ _ _ _ _ _
; Group 3 |1|1|1|1|_|_|_|_|_|_|_|1|
; 
; A third group of operate microinstructions (with a 1 in the least
; significant bit) deals with the optional Extended Arithmetic Element to
; allow such things as hardware multiply and divide, 24 bit shift
; operations, and normalize.  These operations involve an additional data
; register, MQ or multiplier quotient, and a small step count register.
; On the PDP-8/E and successors, MQ and the instructions for loading and
; storing it were always present, even when the EAE was absent, and the
; EAE was extended to provide a useful variety of 24 bit arithmetic
; operations.

OPRsub:	lda	*rc_ste
	ora	#LD_exe		; set EXECUTE bit in STATE
	sta	*rc_ste
	ldd	*MD
	rora			; test for group 1
	bcs	chkpg2

  ; OPR group 1

	bitb	#0b10000000
	beq	_cll
	clra			; CLA
	sta	*AC
	sta	*AC+1
_cll:	bitb	#0b01000000
	beq	_cma
	clr	*Link		; CLL
_cma:	bitb	#0b00100000
	beq	_cml
	ldd	*AC		; CMA
	coma
	anda	#ACmask
	comb
	std	*AC
	ldb	*MD+1
_cml:	bitb	#0b00010000
	beq	_iac
	lda	*Link		; CML
	coma
	anda	#Lmask
	sta	*Link
_iac:	bitb	#0b00000001
	beq	rotate
	ldd	*AC		; IAC
	ora	*Link		; include Link before the increment action
	addd	#1
	sta	*savacc		; save for Link bit update
	anda	#ACmask
	std	*AC
	lda	*savacc
	anda	#Lmask
	sta	*Link		; store possible 12-bit overflow in Link
	ldb	*MD+1
rotate:	andb	#0b00001110 	; isolate rotate and swap bits
	ldu	#Group1
	jmp	[b,u]

  ; OPR group 2

chkpg2:	bitb	#0b00000001
	bne	chkpg3
	bitb	#0b01000000
	beq	_skp1
	lda	*AC		; SMA
	bita	#0b00001000
	bne	SkpTst
_skp1:	bitb	#0b00100000
	beq	_skp2
	ldd	*AC		; SZA
	beq	SkpTst
	ldb	*MD+1
_skp2:	bitb	#0b00010000
	beq	_skp3
	lda	*Link		; SNL
	bne	SkpTst
_skp3:	bitb	#0b00001000
	beq	_skp4
Skip:	ldd	*PC
	addd	#1
	anda	#ACmask
	std	*PC
	bra	_skp4
SkpTst:	ldb	*MD+1
	bitb	#0b00001000
	beq	Skip
_skp4:	ldb	*MD+1
	bpl	_skp5
	ldd	#0		; CLA
	std	*AC
	ldb	*MD+1


    ; test for OSR and HLT. Start with a fast check.

_skp5:	bitb	#0b00000110 	; fast exit check: OSR|HLT?
	lbeq	NextI
	tst	*WorkUI		; check for OSR/HLT block
	lbmi	U_Mode
_skp6:	bitb	#0b00000100
	beq	_skp7
	ldd	*AC		; OSR
	ora	*SR
	orb	*SR+1
	std	*AC
	ldb	*MD+1
_skp7:	bitb	#0b00000010
	lbeq	NextI
	.if	RealCons
	clr	*conact		; HALT --> clear all "action" flags
	.else
	ldb	*MONflg		; HLT
	orb	#Halt		; -> causes RUN lamp to go OFF
	stb	*MONflg
	.endif
	ldd	*PC		; get the PC
	addd	#1		; bump to instruction past the HLT
	anda	#Wmask		; prevent 12-bit overflow
	std	*PC
	jsr	Errpnt
	.fdb	Hltmsg


  ; OPR group 3

chkpg3:	bitb	#0b10000000
	beq	chkmqa
	clra
	sta	*AC		; CLA
	sta	*AC+1
chkmqa:	andb	#0b01010000 	; isolate MQA and MQL bits
	lbeq	NextI
	cmpb	#0b01000000
	bne	chkmql
	ldd	*AC		; MQA
	ora	*MQ
	orb	*MQ+1
	std	*AC
	jmp	NextI
chkmql:	cmpb	#0b00010000
	bne	_swp
	ldd	*AC		; MQL
	std	*MQ
	ldd	#0
	std	*AC
	jmp	NextI
_swp:	ldy	*AC		; SWP
	ldd	*MQ
	std	*AC
	sty	*MQ
	jmp	NextI



BSW:	ldd	*AC
	rolb			; upper and lower bytes in A and B
	rola
	rolb
	rola
	exg	a,b		; swap bytes
	lsra			; align A
	lsra
	rolb			; align B
	rolb
	lsra			; align AC
	rorb
	lsra
	rorb
	std	*AC
	jmp	NextI

ROL:	lda	*Link
	adda	#LtoC		; move Link to carry
	ldd	*AC		; does not change carry
_left:	rolb			; rotate AC and Link (carry) left
	rola
	sta	*Link
	anda	#ACmask		; save AC
	std	*AC
	lda	*Link		; save Link
	anda	#Lmask
	sta	*Link
	jmp	NextI

RTL:	lda	*Link
	adda	#LtoC
	ldd	*AC
	rolb
	rola
	adda	#LtoC		; just affects the Link bit
	bra	_left

ROR:	ldd	*AC
	ora	*Link
_right:	lsra
	rorb
	std	*AC
	rora			; carry to Link "the easy way"
	lsra
	lsra
	lsra
	sta	*Link
	jmp	NextI

RTR:	ldd	*AC
	ora	*Link
	lsra
	rorb
	bcc	_right
	ora	#Lmask
	bra	_right

	.if	0
RORROL: ldd	*AC		; start with ROR
	ora	*Link
	lsra
	rorb
	std	*Temp2		; first shift to MD
	lda	*Link		; now ROL
	adda	#LtoC		; move Link to carry
	ldd	*AC		; does not change carry
	rolb			; rotate AC and Link (carry) left
	rola
	sta	*Link
	anda	#ACmask		; MD and AC to AC
	anda	*Temp2
	andb	*Temp2+1
	std	*AC
	lda	*Link		; save Link
	anda	#Lmask
	sta	*Link
	jmp	NextI
	.endif

ROLROR:	lda	*Link		; start with ROL
	adda	#LtoC		; move Link to carry
	ldd	*AC		; does not change carry
	rolb			; rotate AC and Link (carry) left
	rola
	anda	#ACmask		; save AC
	std	*Temp2
	ldd	*AC		; now do ROR
	ora	*Link
	lsra
	rorb
	anda	*Temp2		; MD and AC to AC
	andb	*Temp2+1
	std	*AC
	rora			; carry to Link "the easy way"
	lsra
	lsra
	lsra
	sta	*Link
	jmp	NextI


    ; default "no device" IOT
    ; if the device does not exist, there is no action

Nodev	=	NextI

    ; NOTE: the following was the previous inplementation for "no device"

;Nodev:	jsr	Errpnt
;	.fdb	NoIOT
;NoIOT:	.asciz	"Access to non-existent I/O device"




    ; Rotate and swap routines for group 1 OPR.
    ; NOTE: The "undefined" instructions may have rotated the AC and
    ;       Link into the MD register and then in the opposite direction into itself.
    ;       Next the AC and MD registers may have been anded and put into the AC.
    ;       State of the link would depend on which direction was done first.

Group1:	.fdb	NextI		; No operation
	.fdb	BSW		; Byte swap AC
	.fdb	ROL		; Rotate AC and Link left
	.fdb	RTL
	.fdb	ROR		; Rotate AC and Link right
	.fdb	RTR
	.fdb	ROLROR		; ROR-ROL (undefined on pdp8/e)
	.fdb	Illins		; RTR-RTL (undefined on pdp8/e)


    ; Device routine address array. 64 possible devices.
    ; NOTE: The KM8-E is done but not tested.

Device:	.fdb	IOT00	; Device 00 Interrupts (basic system)
	.fdb	IOT01	; Device 01 PC02 High Speed Papertape Reader
	.fdb	IOT02	; Device 02 PC03 High Speed Papertape Punch
	.fdb	IOT03	; Device 03 KL8-E Teletype keyboard/reader
	.fdb	IOT04	; Device 04 KL8-E Teletype teleprinter/punch
	.fdb	Nodev	; Device 05.
	.fdb	Nodev	; Device 06.
	.fdb	Nodev	; Device 07.
	.fdb	Nodev	; Device 10 KP8-E Power Fail Detect
	.fdb	Nodev	; Device 11.
	.fdb	Nodev	; Device 12.
	.fdb	Nodev	; Device 13 DK8-EA/EC/EP Real Time Clocks
	.fdb	Nodev	; Device 14.
	.fdb	Nodev	; Device 15.
	.fdb	Nodev	; Device 16.
	.fdb	Nodev	; Device 17.
	.fdb	IOT20	; Device 20 KM8-E Memory ext. & Time Share
	.fdb	IOT21	; Device 21   "      "    "   "   "    "
	.fdb	IOT22	; Device 22   "      "    "   "   "    "
	.fdb	IOT23	; Device 23   "      "    "   "   "    "
	.fdb	IOT24	; Device 24   "      "    "   "   "    "
	.fdb	IOT25	; Device 25   "      "    "   "   "    "
	.fdb	IOT26	; Device 26   "      "    "   "   "    "
	.fdb	IOT27	; Device 27   "      "    "   "   "    "
	.fdb	Nodev	; Device 30.
	.fdb	Nodev	; Device 31.
	.fdb	Nodev	; Device 32.
	.fdb	Nodev	; Device 33.
	.fdb	Nodev	; Device 34.
	.fdb	Nodev	; Device 35.
	.fdb	Nodev	; Device 36.
	.fdb	Nodev	; Device 37.
	.fdb	Nodev	; Device 40.
	.fdb	Nodev	; Device 41.
	.fdb	Nodev	; Device 42.
	.fdb	Nodev	; Device 43.
	.fdb	Nodev	; Device 44.
	.fdb	Nodev	; Device 45.
	.fdb	Nodev	; Device 46.
	.fdb	Nodev	; Device 47.
	.fdb	Nodev	; Device 50.
	.fdb	Nodev	; Device 51.
	.fdb	Nodev	; Device 52.
	.fdb	Nodev	; Device 53.
	.fdb	Nodev	; Device 54.
	.fdb	Nodev	; Device 55.
	.fdb	Nodev	; Device 56.
	.fdb	Nodev	; Device 57.
	.fdb	IOT60	; Device 60 DF32/RF08 DECdisk File
	.fdb	IOT61	; Device 61     "       "      "
	.fdb	IOT62	; Device 62     "       "      "
	.fdb	IOT63	; Device 63     "       "      "
	.fdb	IOT64	; Device 64     "       "      "
	.fdb	Nodev	; Device 65.
	.fdb	Nodev	; Device 66.
	.fdb	Nodev	; Device 67.
	.fdb	Nodev	; Device 70.
	.fdb	Nodev	; Device 71.
	.fdb	Nodev	; Device 72.
	.fdb	IOT73	; Device 73 RK8-P Disk System
	.fdb	IOT74	; Device 74   "    "     "
	.fdb	IOT75	; Device 75   "    "     "
	.fdb	Nodev	; Device 76.
	.fdb	Nodev	; Device 77.



    ; Interrupt IOT module.  Comes with the basic system.
    ;
    ; Standard IOPs' are 0 through 7.  Higher ones are "impossible",
    ; and used by the emulator for system initiation (8) and reset (9).

IOT00:	ldu	#IOP00		; use an opcode array
	lsla			; get the IOP code
	jmp	[a,u]


IOP00:	.fdb	SKON		; 0
	.fdb	ION		; 1
	.fdb	IOF		; 2
	.fdb	SRQ		; 3
	.fdb	GTF		; 4
	.fdb	RTF		; 5
	.fdb	IOT0_6		; 6
	.fdb	CAF		; 7
	.fdb	No_op		; 8 = initialise
	.fdb	No_op		; 9 = reset


    ; Skip if interrupt on and turn off.
    ; Assume "on" not influenced by ION and JMP/JMS inhibits.

SKON:	ldd	#Doinst
	cmpd	*emufnc		; is the processing routine?
	lbeq	NextI		; y -
	std	*emufnc		; n - make it so! (set vector to processing routine)
	jmp	PCskip


    ; Enable interrupts. Actual response is delayed till after the next pdp8
    ; instruction is simulated (unless interrupts are already enabled).
    ; NOTE: This is independent of JMP/JMS inhibit.

ION:	ldd	*emufnc		; get current selected processing routine
	cmpd	#Doinst		; interrupts already enabled?
	lbne	NextI		; y - one instruction delay not needed
	lda	#2		; n - delay one simulated pdp8 instruction
	sta	*Idelay		;     before processing any interrupt
	ldd	#Chkdly
setjmp:	std	*emufnc		; set vector to processing routine
	jmp	NextI


    ; Turn interrupts off (unconditional).
    ; NOTE: Does not change JMP/JMS inhibit.

IOF:	ldd	#Doinst
	bra	setjmp


    ; Skip if interrupt request

SRQ:	lda	*IOflgs
	anda	#Jdelay^0xFF	; drop JMP/JMS inhibit
	ora	*IOflgs+1
	lbne	PCskip
	jmp	NextI


    ; Get flags
    ; NOTE: GT flag of the Extended Arithmetic Element not processed!

    ; DEBUG test - not sure how Interrupt Inhibit bit (Time Share) works
    ; e-mail from Vince:
	; According to the web page, RTF is wrong in the same way as GTF -- both are
	; are supposed to operate on the saved fields, not the work/buff fields.
	; However, the DEC manual for the 8/a seems todisagree, and so does SIMH.
	; The RTF instruction as implemented matches the description on page A-6 of
	; the 8/a handbook.
	; In any case, that doesn't seem to be the cause of the failure.  The value
	; of AC suggests that the value read by the GTF was 5600 instead of 5200.
	; That says that interrupt inhibit was set by the RTF.
	; In SIMH, RTF loads Link, GT, DF, and buffered versions of IF, then pretends
	; a CIF is pending (and turns on interrupts whether the IE bit was set or
	; not).  That actually seems equivalent to what this simulator is doing!
	; There, they get away with it, because SIMH never returns a value with 0400
	; (Interrupt inhibit) set for a GTF instruction.
	; So, after all that, it looks like the GTF instruction should just omit the
	; stuff at getBlk that sets 0b00000001 in A. (I'd have to check the drawings
	; (ugh) to see if that's what a real 8e would do, too.)

dbg_II	=	0		; for now we can remove it with 6 NOP's


GTF:	ldb	*SaveUI		; get I field and user state
	bpl	getInt		; test for User mode
	orb	#0b00001000	; set User bit
getInt:	andb	#0b00001111	; keep User and I field bits (defensive code)
	ldu	*emufnc		; check interrupt status
	cmpu	#Doinst
	beq	_getD
	orb	#0b00010000	; set Ion bit
_getD:	lslb			; move to correct bit position
	lslb
	lslb
	orb	*SaveD		; and add D field bits
	lda	*IOflgs		; check for pending interrupt requests
	anda	#Jdelay^0xFF	; drop JMP/JMS delay
	ora	*IOflgs+1

	.if	dbg_II		; ###  DEBUG test - Interrupt Inhibit bit  ###
	beq	getBlk		; [A] = 00
	lda	#0b00000010	; set Interrupt Request flag
getBlk:	tst	*IOflgs		; check for Interrupt Enable/Disable
	bpl	getLnk
	ora	#0b00000001	; set JMP/JMS delay (Interrupt Disable)
	.else
	beq	getLnk		; [A] = 00
	lda	#0b00000010	; set Interrupt Request flag

	.endif

getLnk:	tst	*Link
	beq	savFlg
	ora	#0b00001000	; set the Link bit
savFlg:	std	*AC
	jmp	NextI


    ; Restore flags

RTF:	ldd	*AC		; restore Link
	rola
	anda	#Lmask
	sta	*Link
	andb	#Dbits		; restore D
	stb	*WorkD
	lslb
	ldu	#GetFld
	ldd	b,u
	std	*WrkDp
	lda	*AC+1		; restore U and I (to Buffer)
	lsra
	lsra
	lsra
	anda	#0b00001111
	adda	#0b01111000	; shift U to MSB
	anda	#Ubit|Fbits
	sta	*BuffUI
	lsla			; drops U
	ldd	a,u
	std	*BufIp
	ldd	#Chkint		; Interrupts on.
				; The JMP/JMS delay takes care of the 1 cycle delay
	std	*emufnc		; set vector to processing routine
	bra	_setJ		; set JMP/JMS delay


    ; Clear all Flags

CAF:	jsr	Clear
	jmp	NextI


IOT0_6:	jsr	Errpnt
	.fdb	NoIerr


    ; Use for bypassing IOT device setup or initialisation.

No_op:	clra			; no error
	rts			; do not do setup or initialise



    ; KM8-E Memory Extension and Time-Share Option.
    ;
    ; Memory extension allows up to 8 fields (4096 words/field).
    ; Time share has Executive and User modes, controlled by the U bit.
    ; If it is 1 the computer is in time share mode and all OSR, HLT and
    ; IOT instructions are ignored and a time share interrupt is generated.
    ; The interrupt sets the Instruction and Data fields and the time share bit to 0.

IOT20:	ldb	#0*2		; field number times 2
	bra	km8iot
IOT21:	ldb	#1*2
	bra	km8iot
IOT22:	ldb	#2*2
	bra	km8iot
IOT23:	ldb	#3*2
	bra	km8iot
IOT24:	ldb	#4*2
	bra	km8iot
IOT25:	ldb	#5*2
	bra	km8iot
IOT26:	ldb	#6*2
	bra	km8iot
IOT27:	ldb	#7*2
km8iot:	lsla			; IOT times 2
	ldu	#mm8op0
	jmp	[a,u]

mm8op0:	.fdb	IOT2N		; 0
	.fdb	CDF		; 1
	.fdb	CIF		; 2
	.fdb	CDFCIF		; 3
	.fdb	mm8_N4		; 4
	.fdb	IOT2N		; 5
	.fdb	IOT2N		; 6
	.fdb	IOT2N		; 7
	.fdb	No_op		; 8
	.fdb	No_op		; 9


mm8_N4:	ldu	#mm8op1
	jmp	[b,u]

mm8op1:	.fdb	CINT
	.fdb	RDF
	.fdb	RIF
	.fdb	RIB
	.fdb	RMF
	.fdb	SINT
	.fdb	CUF
	.fdb	SUF

    ; Change Data field

CDF:	ldu	#GetFld
	ldu	b,u
	stu	*WrkDp
	lsrb
	stb	*WorkD
	jmp	NextI


    ; Change Instruction field

CIF:	ldu	#GetFld		; must have I field
	ldu	b,u
	lbeq	Flderr
	lsrb
_cif0:	stu	*BufIp
	tst	*BuffUI		; keep Time Share Mode
	bpl	_cif1
	orb	#Ubit
_cif1:	stb	*BuffUI
_setJ:	lda	#Jdelay		; set JMP/JMS delay
	ora	*IOflgs
	sta	*IOflgs
	jmp	NextI


    ; Change Data and Instruction field

CDFCIF:	ldu	#GetFld
	ldu	b,u
	lbeq	Flderr		; must have I field
	stu	*WrkDp
	lsrb
	stb	*WorkD
	bra	_cif0


    ; Read Data field value

RDF:	ldb	*WorkD
_rdf0:	lslb
	lslb
	lslb
_rdf1:	orb	*AC+1
	stb	*AC+1
	jmp	NextI


    ; Read Instruction field value

RIF:	ldb	*WorkUI
	bra	_rdf0


    ; Read Interrupt Buffer

RIB:	ldb	*SaveUI		; [A]=SaveUI  //  [B]=SaveD
	lslb
	lslb
	lslb
	orb	*SaveD
	tst	*SaveUI
	bpl	_rdf1
	orb	#0b01000000	; set U mode
	bra	_rdf1


;    Restore Memory field

RMF:	ldd	*SaveUI
	sta	*BuffUI
	stb	*WorkD
	ldd	*SavDp
	std	*WrkDp
	ldd	*SavIp
	std	*BufIp
	bra	_setJ


CINT:	lda	*IOflgs
	anda	#UserI^0xFF
	sta	*IOflgs
	jmp	NextI

SINT:	lda	*IOflgs
	anda	#UserI
	lbne	PCskip
	jmp	NextI


    ; Set Time Share mode to Executive.
    ; NOTE: Not clear if this sets the JMP/JMS delay,
    ;       as it stops Time Share interrupts.

CUF:	lda	*BuffUI
	anda	#Ubit^0xFF
	sta	*BuffUI
	jmp	NextI

SUF:	lda	*BuffUI
	ora	#Ubit
_suf0:	sta	*BuffUI
	bra	_setJ


    ; Time share interrupt.
    ; User mode and LAS, HLT or IOT instruction.
    ; Instruction is treated as a no-op and an interrupt is generated.

U_Mode:	lda	*IOflgs		; user interrupt bit is in upper byte
	ora	#UserI		; set User flag
	sta	*IOflgs
	jmp	NextI

    ; KM8-E decode error: illegal / not recognised IOT

IOT2N:	jsr	Errpnt
	.fdb	KM8Eerr



    ; PC02 High Speed Papertape Reader
    ;
    ; kept very simple: re-use part of the Teletype IOT #3 reader-in code.

IOT01:	ldu	#IOP01		; use an opcode array
	bra	iotdec		; get the IOP code

IOP01:	.fdb	RPE		; 0 RPE - set reader/punch interrupt enable
	.fdb	KSF		; 1 RSF - skip if reader flag = 1 (as KSF)
	.fdb	RRB		; 2 RRB - read reader buffer, clear buffer
	.fdb	IOT013		; 3
	.fdb	KCC		; 4 RFC - clear flag & buffer, fetch char (as KCC)
	.fdb	IOT015		; 5
	.fdb	RRC		; 6 RRC - read buffer & fetch char (RRB+RFC)
	.fdb	IOT017		; 7
	.fdb	No_op		; 8
	.fdb	No_op		; 9

   ; PC02 error message(s)

IOT013:
IOT015:
IOT017:	jsr	Errpnt
	.fdb	PC02_IOP




    ; PC03 High Speed Papertape Punch
    ;
    ; kept very simple: re-use the Teletype IOT #4 reader-out code.

IOT02:	ldu	#IOP02		; use an opcode array
	bra	iotdec		; get the IOP code

IOP02:	.fdb	PCE		; 0 PCE - clear reader/punch interrupt enable
	.fdb	TSF		; 1 PSF - skip on punch flag = 1 (as TSF)
	.fdb	TCF		; 2 PCF - clear punch flag and buffer (as TCF)
	.fdb	IOT023		; 3
	.fdb	PPC		; 4 PPC - load punch buffer and punch
	.fdb	IOT025		; 5
	.fdb	PLS		; 6 PLS - clear flag, load buffer and punch
	.fdb	IOT027		; 7
	.fdb	No_op		; 8
	.fdb	No_op		; 9

   ; PC03 error message(s)

IOT023:
IOT025:
IOT027:	jsr	Errpnt
	.fdb	PC03_IOP





    ; KL8-E Console Teleprinter Control Module
    ;
    ; Includes both IOT 03 and IOT 04. Interrupt request enable is common to both,
    ; and enabled at reset. Also controlled by AC[11] on a KIE instruction.
    ; If the AC[11] bit is 0, interrupt requests are disabled.

IOT03:	ldu	#IOP03		; use an opcode array
	bra	iotdec


IOT04:	ldu	#IOP04
iotdec:	lsla			; get the IOP code
	jmp	[a,u]

IOP03:	.fdb	KCF		; 0 - clear keyboard/reader flag, device not operated
	.fdb	KSF		; 1 - skip next instruction if kb/rdr flag = 1
	.fdb	KCC		; 2 - clear AC, clear keyboard/reader flag
	.fdb	IOT033		; 3
	.fdb	KRS		; 4 - read char from kb/rdr buffer, flag set on compl
	.fdb	KIE		; 5 - enable/disable kb/rdr interrupt if AC[11] = 1/0
	.fdb	KRB		; 6 - clear AC and kb/rdr flag, read char from buffer
	.fdb	IOT037		; 7
	.fdb	No_op		; 8
	.fdb	No_op		; 9

IOP04:	.fdb	TFL		; 0
	.fdb	TSF		; 1
	.fdb	TCF		; 2
	.fdb	IOT043		; 3
	.fdb	TPC		; 4
	.fdb	TSK		; 5
	.fdb	TLS		; 6
	.fdb	IOT047		; 7
	.fdb	TTinit		; 8 initialise
	.fdb	TTrst		; 9 reset


    ; RS_ion: enable RS232 input interrupt

RS_ion:	lda	ACIADA		; clear possible pending ACIA interrupt
	lda	#ACIAI		; enable ACIA-IRQ, clock/16, 8N2
	sta	ACIACS		; RTS=0 (LED off)
	.if	RealCons
	orcc	#0x40		; disable FIRQ: mutual excl. access to rtssts
	sta	*rtssts
	andcc	#0xAF		; enable FIRQ _and_ IRQ (RS232)
	.else
	andcc	#0xEF		; enable IRQ
	.endif
	rts



    ; Initialise the TTY handler

TTinit:	jsr	GetFlg		; get the interrupt bit
	std	*Ibit
	coma			; Interrupt bit complement
	comb
	std	*Icom
	ldd	#TTIchr		; set up the input interrupt vector
	std	.irq
	bsr	RS_ion		; turn the interrupt on
TTrst:	ldd	#0		; clear the device flags
	std	*TTIflg
	inca			; set up the interrupt status
	sta	*Istat
	ldd	*IOflgs
	anda	*Icom
	andb	*Icom+1
	std	*IOflgs
	andcc	#0xEF		; turn IRQ on
	orcc	#4		; ccZ=1 :: successful setup
	rts



    ; Character input handler, driven by the IRQ interrupt.
    ; check for ^T to start trace facility (if RealConsole is connected).
    ; Check for ^S to call the monitor (if RealConsole is not connected).

TTIchr:	lda	ACIADA		; get serial input data / clear IRQ
	.if	RealCons
	tst	*trcflg		; is trace facility option selected at start-up?
	beq	10$		; n - accept all data (could be binary incl. ^T)
	cmpa	#CtrlT		; y - "turn on trace facility" character ^T ?
	beq	tttrac
10$:	bsr	ttisub		;     n - put character in system
	rti
tttrac:	lda	#TRCNEW		;     y - enable emulation commands / tracing
	sta	*trcsts
	sta	*trcflg
	clr	*trcdmp		;	  disable output TRACE BEFORE re-routing
	lda	*conact
	anda	#255-ACTrun	;	  drop out of RUN state
	ora	#ACTtrc		;	  enable trace facility
	sta	*conact
	rti
	.else
	cmpa	#XOFF		; turn on single step (^S)?
	beq	ttstep
	bsr	ttisub		; n - put character in system
	rti
ttstep:	lda	*MONflg		; y - call monitor
	ora	#Console
	sta	*MONflg
	rti
	.endif


ttisub:	sta	*TTIdat
	lda	#1		; set keyboard/reader flag
	sta	*TTIflg
	tst	*Istat		; update interrupt request
	beq	charxt
	ldd	*IOflgs		; interrupt was enabled
	ora	*Ibit
	orb	*Ibit+1
	std	*IOflgs		; request interrupt
charxt:	rts


    ; Update the interrupt request.
    ;  - if interrupt enabled (and a flag set) request an interrupt.
    ;  - if interrupt disabled clear the request.

Isetup:	tst	*Istat		; interrupt enabled?
	beq	Iclr
	ldd	*TTIflg		; any requests?
	beq	Iclr
Iset:	ldd	*IOflgs		; y - request interrupt
	ora	*Ibit
	orb	*Ibit+1
	bra	isetxt
Iclr:	ldd	*IOflgs		; n - clear interrupt request
	anda	*Icom
	andb	*Icom+1
isetxt:	std	*IOflgs
	jmp	NextI


    ;-----------------------------------------------------------------------'
    ; TTI uses the 6809-IRQ interrupt to get a character from the keyboard. '
    ; TTO does not turn the interrupt off while printing characters.	    '
    ;-----------------------------------------------------------------------'


    ; ***							 ***
    ; ***  TTI section (Teletype keyboard or paper tape reader)  ***
    ; ***							 ***


KSF:	tst	*TTIflg		; skip on keyboard flag
ksfskp:	lbne	PCskip
	jmp	NextI


    ; Clear the keyboard flag.
    ; Enable the paper tape reader to get another character.
    ; NOTE: DEC modifies the standard ASR-33 Teletype to input
    ;       a single character from the reader on each request.
    ;       It is assumed that a keyboard character always sets the flag.

KCC:	ldd	#0
kccset:	std	*AC
; --	jsr	RDRrun		;	remark: routine does not (yet/ever?) exist

				; Clear keyboard flag (-> does not enable the paper
KCF:	clr	*TTIflg		;	       tape reader to get another character)
	bra	Isetup


KRS:	ldd	*AC		; read keyboard buffer static
	orb	*TTIdat
	orb	#0x80		; add in for TTY compatibility
krsset:	std	*AC
	jmp	NextI


KIE:	ldd	*AC		; set/clear interrupt enable
	andb	#1
kieset:	stb	*Istat
	bra	Isetup


KRB:	clra			; read keyboard buffer dynamic
	ldb	*TTIdat
	orb	#0x80		; add in for TTY compatibility
	bra	kccset


   ; IOT's for the high-speed papertape reader
   ;  - RRB is identical to KRS, but does *not* set bit 7 (msb).

RPE:	ldb	#1		; set reader/punch interrupt enable
	bra	kieset


RRB:	ldd	*AC		; RRB - read reader buffer, clear buffer
	orb	*TTIdat
	clr	*TTIflg
	bra	krsset


RRC:	ldd	*AC		; RRC - read buffer & fetch char (RRB+RFC)
	orb	*TTIdat
	bra	kccset


   ; TTI error message(s)

IOT033:
IOT037:	jsr	Errpnt
	.fdb	TTI_3_7


    ; ***						       ***
    ; ***  TTO section (Teletype printer or paper tape punch)  ***
    ; ***						       ***

TFL:	lda	#1		; set the teleprinter flag
	sta	*TTOflg
	bra	Isetup


TCF:	clr	*TTOflg		; Clear the teleprinter flag
	bra	Isetup

				; Load teleprinter and print.
TPC:	lda	*AC+1		;  - does not clear the Teleprinter flag
	anda	#0x7F		; drop high bit
tpcset:	jsr	outch
	lda	#1
	sta	*TTOflg
	bra	Isetup


TSF:	tst	*TTOflg		; Skip on teleprinter flag
	bra	ksfskp


TSK:	lda	*TTIflg		; Skip on printer or keyboard flag
	ora	*TTOflg		;   if the interrupt is enabled
	anda	*Istat
	lbeq	NextI
	jmp	PCskip


    ; Load teleprinter sequence.
    ; Same as TPC, except that the Teleprinter flag is cleared first.
    ; Since the character output is not interrupt driven the effect is
    ; the same as TPC.

TLS:	bra	TPC		; Load punch buffer sequence - no difference


   ; IOT's for the high-speed papertape punch
   ;  - PPC and PLS are identical to TPC and TLS, but do *not* clear bit 7 (msb).

PCE:	clrb			; Clear reader/punch interrupt enable
	bra	kieset

PLS:				; Load punch buffer and punch character
PPC:	lda	*AC+1		; Load punch buffer sequence - no difference
	bra	tpcset


    ; TTO error message(s)

IOT043:
IOT047:	jsr	Errpnt
	.fdb	TTO_3_4




; ===================================================================================
; ##				  DF32  DECdisk File				   ##
; ##_______________________________________________________________________________##
;
;
;				   ***	REMARK  ***
;
;			The DF32 DECdisk File and the RF08 Disk File
;			share the same IOT's.  What follows is first
;			a description of the DF32, then the RF08.
;			The generated code depends on the conditional
;			assembly flag "RF08dev". If the flag is zero,
;			code is generated for the DF32.  If the flag
;			is set to one code is generated for the RF08.
;
;
;   DF32 introductory information
;  -------------------------------
;
; The Type DF32 Disk File and Control is a fast *random* or *sequential* access
; bulk storage device. One DF32 provides a capacity of 32768 13-bit words, or
; 16 tracks of 2048 13-bit words. If the specified data transfer encompasses more
; than one track, the data transfer operation continues automatically to the next
; track. Similarly, operation continues to the next unit when the last track and
; track address have been accessed, provided that a next unit exists ... Up to 3
; DS32 Extender Disks can be attached to the DF32, extending the capacity to four
; times 32768 words = 131072 13-bit words.
; Addressing scheme: random or sequential addressing from 0 to 32768 words, with
; variable block size from one word to 4096 words.
; The Disk File is a program controlled device and operates through the 3-cycle
; data break facility with the following addresses:
; WC register (address 7750) - specifies the number of word transfers
; CA register (address 7751) - current address of core memory to receive/send a word
;
; The DF32 has a memory buffer (DMB) to buffer the data between the core memory and
; the disk, and a disk memory address (DMA) to specify the track address, the EMA
; (Extended Memory Address) to specify the track and unit number.
;
;
;  DF32 I/O Transfer instructions (IOT's)
; ----------------------------------------
;
;    DCMA [6601] - Clear the disk memory address, parity error, completion flags and
;		   interrupt flags.
;    DMAR [6603] - Load the disk memory address with information (initial address)
;		   in the accumulator AC. Then clear the AC. Begin to read information
;		   from the disk into the specified core location. Clear parity and
;		   completion flags. Clear interrupt flags, AC[0:11] , DMA[0:11].
;    DMAW [6605] - Load the disk memory address with information (initial address)
;		   in the accumulator AC. Then clear the AC. Begin to write information
;		   onto the disk from the specified core loaction. Clear parity and
;		   completion flags. Clear interrupt flags, AC[0:11] , DMA[0:11].
;    DCEA [6611] - Clear the disk extended address and memory address extension
;		   register (for extended memory in the computer).
;    DSAC [6612] - Skip next instruction if the address confirmed flag is a 1.
;		   Flag is set for 16 sec. AC is cleared.
;		   Used primarily in diagnostic in diagnostic programming.
;    DEAL [6615] - Load the disk extended address and memory address extension
;		   registers with the track address data held in the AC.
;
;	  ___0_____1_____2_____3_____4_____5_____6_____7_____8_____9____10____11___
;	  |#####| EMA4| EMA3| EMA2| EMA1| EMA0| EM2 | EM1 | EM0 |#####|#####|#####|  AC
;	  |_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|
;
;		    EMA4 \   disk memory	       EM2 \
;		     ::   >  extended address	       EM1  > Core Field extension bits
;		    EMA0 /	 bits		       EM0 /
;
;    DEAC [6616] - Load the contents of the disk extended address register into the AC
;		   to allow program evaluation. Skip next instruction if the Address
;		   Confirmed Flag is a 1.
;
;	  ___0_____1_____2_____3_____4_____5_____6_____7_____8_____9____10____11___
;	  | PHC | EMA4| EMA3| EMA2| EMA1| EMA0| EM2 | EM1 | EM0 | DRL | NED | PER |  AC
;	  |_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|
;
;		   PHC - PhotoCell Sync Mark (200 s)
;		   EMA4 \   disk memory		       EM2 \
;		    ::   >  extended address	       EM1  > Core Field extension bits
;		   EMA0 /	 bits		       EM0 /
;		   DRL - Data Request Late	       NED \  [ NonExisting Disk ]
;		   PER - Parity Error		       EWL /  [ Error Write Lock ]
;
;    DFSE [6621] - Skip next instruction if the parity error, data request late and
;		   write lock switch flag are all 0 (no errors).
;    DFSC [6622] - Skip next instruction if the completion flag is a 1 (data transfer
;		   is complete).
;    DMAC [6626] - Load the contents of the disk memory address register into the AC to
;		   allow program evaluation.  DMA[0:11] --> AC[0:11]
;		   During read the final address is the last address transferred +1.
;		   During write the final address is the last address transferred.
;
;
;  DF32 interrupt flags
; ----------------------
; There are 2 flags that generate an interrupt:
;   1) NED (nonexistent disk)
;   2) TRC (transfer complete)
; The NED flag is set when the program selects (via EMA) a unit which does not exist,
; or the data transfer operation increments the EMA to next unit and it does not
; exist. A non-existent unit means that the program has selected a unit number and no
; disk file is set to that unit number.
; The TRC flag signifies the end of the data transfer. It is set at the completion of
; the last word transfer by the disk, following a word count overflow of the WC
; register. The completion flag can be sensed by the DFSC instruction.
;
;
;  Hardware delay effects
; ------------------------
; It is logical that a disk access to read or write data needs time: the correct
; sector must pass the head. This is called latency delay.  The data transfer
; completion is an event that occurs *asynchronous* to the code: after the I/O
; request is issued, the processor *can* continue with the next instruction!
; The actual data transfer ocuurs in "Data Break" cycles. This is called nowadays
; DMA. When the data transfer is completed (several/many pdp8 instructions later)
; the device generates an interrupt to signal that the I/O transfer is completed.
; A counter is added in the simulator to get this Data Break behaviour.
; When a data transfer IOT is executed (for DF32/RF08 these are DMAR and DMAW),
; the IOT itself *only* sets up the status flags of the device and marks the
; device "active" in the <deftab> indicating that "deferred I/O" is in progress.
; In <deftab> is also a counter initialized. In the main instruction decode loop
; this counter is decremented after a pdp8 instruction is simulated.
; When the counter reaches zero, the device "active" deferred I/O transfer flag
; is cleared, and the actual data transfer occurs in "one cycle". When the data
; transfer is completed a routine is called that sets the status bits and issues
; the interrupt request flag.
;
;
;  DF32 error flags
; ------------------
; The error flags can be sensed by the IOT 6621 instruction. The program skips when
; no error exists. The error flags are:
;    DRL - Data Request Late
;		* a data transfer operation occurred between the disk DMB and the
;		  disk before the previous transfer was handled by the data break
;		  facility,
;		* address accepted was not received back from the computer, leaving
;		  the break request flag up.
;    PER - Parity error
;    WIA - a write operation was attempted on the lower 16K memory addresses of the
;	   DF32 when they were locked out by the write lockout switches.
;    WIB - same as WIA except on upper 16K of disk memory.
;    EWL - WIA or WIB error occurred on a selected disk expander. When sensed by the
;	   DEAC instruction, WIA, WIB, and EWL are contained in the same list.
;
;
; The DF32 Disk File can transfer between 1 and 4096 words in *one* transfer,
; and the disk can be either sequential *or* random accessed !
; This functionality has the following consequences:
;  1) I/O request from/to IDE disk less than or equal to 256 words
;   	READ
;	means reading the whole sector, but pass on only the requested number
;	of words, starting at the correct word in the pdp8/e simulator.
;	WRITE
;	means reading the whole sector in a buffer, writing the requested number
;	of words into the buffer, starting at the correct word, and the writing
;	the whole buffer to the IDE disk sector.
;  2) I/O request from/to IDE disk more than 256 words
;   	READ
;	means reading more than one sector. Possible conditions are that the
;	first sector is passed partial, which is also possible for the last
;	sector. As an example, execute the request of 500 words, start at word 200:
;	   sector 0: partial   (words 200 ... 255)
;	   sector 1: complete  (words 256 ... 511)
;	   sector 2: partial   (words 512 ... 701)
;	WRITE
;	means writing more than one sector. Same considerations apply as for READ.
;
;
;
;
; ===================================================================================
; ##                              RF08  Disk File                                  ##
; ##_______________________________________________________________________________##
;
;   RF08 introductory information
;  -------------------------------
;
; The RF08 Disk Control and the RS08 Disk provide high-speed bulk storage.
; Each RS08 Disk has a storage capacity of 2048 (decimal) addresses on each of 128
; tracks, giving a total of 262144 (decimal) addresses. Each address contains a
; 12-bit word plus a 13th bit for parity checking (not in this simulation).
; The RF08 Disk Control controls up to four RS08 Disks giving the disk file a
; maximum capacity of 1048576 (decimal) words.
; The Disk File uses the 3-cycle data break facility of the computer.
; The length of a data block is variable, and is specified by loading the length
; as a negative number into the word count register. The length of a single block
; is only limited by the maximum of words (4096) that can be specified in 12 bits.
; The start address can be randomly selected.
;
;  RF08 interrupt flags
; ----------------------
; There are 2 flags that generate an interrupt:
;   1) NED (nonexistent disk)
;   2) TRC (transfer complete)
;
;
;  RF08 I/O Transfer instructions (IOT's)
; ----------------------------------------
;
; The RF08 IOT's are identical to the IOT's of the DF32. Thus, without sysgen to
; change the device address, either DF32 or RF08 can be installed in a single system.
; The following description of the RF08 IOT's lists the differences if they exist.
; If an asterisk leads the IOT name, it indicates this is a new IOT (not in DF32).
;
;    DCMA [6601] - Clear the disk memory address, parity error and completion flags.
;		   Does not clear interrupt enable or extended address registers.
;    DMAR [6603] - Load the disk memory address with information (initial address) in
;		   the accumulator AC. Then clear the AC. Begin to read information
;		   from the disk into the specified core location. Clear parity and
;		   completion flags. Clear interrupt flags, AC[0:11] , DMA[0:11].
;    DMAW [6605] - Load the disk memory address with information (initial address) in
;		   the accumulator AC. Then clear the AC. Begin to write information
;		   onto the disk from the specified core loaction. Clear parity and
;		   completion flags. Clear interrupt flags, AC[0:11] , DMA[0:11].
;    DCIM [6611] - Clear the disk interrupt enable [DIE], and the core (Field)
;		   extended address register (for extended memory in the computer).
;    DSAC [6612] - Skip next instruction if the address confirmed flag is a 1.
;		   Flag is set for 16 microseconds. AC is cleared. Used primarily
;		   in diagnostic programming.
;    DIML [6615] - Load interrupt enable bits and the core memory address extension
;		   register with the data held in the AC. Then clear AC.
;		   NOTE: DEAL [6615] is similar, but functions transmitted from AC
;			 are different. EMA information is not transmitted. See DXAL.
;
;	  ___0_____1_____2_____3_____4_____5_____6_____7_____8_____9____10____11___
;	  |#####|#####|#####| IER | ICL | ICO | EM2 | EM1 | EM0 |#####|#####|#####|
;	  |_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|
;
;		   IER - Interrupt Enable Error	    EM2 \
;		   ICL - Interrupt Enable Clock	    EM1  >  Core Field Extension bits
;		   ICO - Interrupt Enable Complete  EM0 /
;
;    DIMA [6616] - Load the contents of the status register into AC to allow program
;		   evaluation. Similar to DSAC; however, MB9 is a '1', inhibit skip
;		   on address confirmed flag.
;		   NOTE: DEAC [6616] is similar, but functions transmitted to AC
;			 are different. See DXAC.
;
;	  ___0_____1_____2_____3_____4_____5_____6_____7_____8_____9____10____11___
;	  | PCA | DRE | WLS | EIE | PIE | CIE | EM2 | EM1 | EM0 | DRL | NXD | PER |
;	  |_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|
;
;		   PCA - Photo Cell Sync mark	    DRE - Data Request Enable (maint)
;		   WLS - Write Lock Status	    EIE - Error Interrupt Enable
;		   PIE - Photocell Interrupt Enable CIE - Completion Interrupt Enable
;		   EM2 \			    DRL - Data Request Late
;		   EM1  > Core Field extension bits NXD - Non Existing Disk
;		   EM0 /			    PER - Parity Error
;
;    DFSE [6621] - Skip next instruction if the parity error, data request late,
;		   write lock switch or non-existent disk flag is a 1 (error occurred).
;		   NOTE: DFSE on DF32 is the opposite: skip on -no- error.
;    DFSC [6622] - Skip next instruction if the completion flag is a 1 (data transfer
;		   is complete).
;  * DISK [6623] - Skip next instruction if either the error or completion flag is
;		   set, or both. In fact a combination of DFSC and DFSE.
;    DMAC [6626] - Load the contents of the disk memory address register into the AC
;		   to allow program evaluation.   DMA[0:11] --> AC[0:11].
;		   DMA contents does not change.
;  * DCXA [6641] - Clear the 8 high-order disk extended address bits.
;  * DXAL [6643] - Load the 8 high-order disk extended address bits with bits 4-11 of
;		   the AC. Data in bits 0-3 of AC are ignored. Then the AC is cleared.
;  * DXAC [6645] - Load the 8 high-order disk extended address bits into bits 4-11 of
;		   the AC.  Data in bits 0-3 of AC are zero.
;		   The disk extended address bits do not change.
;  * DMMT [6646] - For Maintenance purposes only. The AC is cleared.



IOT60:	ldu	#IOP60		; use an opcode array
	bra	iot6x
IOT61:	ldu	#IOP61
	bra	iot6x
IOT62:	ldu	#IOP62
	bra	iot6x
IOT63:	ldu	#IOP63
	bra	iot6x
IOT64:	ldu	#IOP64
iot6x:	lsla			; get the IOP code
	jmp	[a,u]


	.if	RF08dev

IOP60:	.fdb	NORF08		; 0
	.fdb	DCMA		; 1
	.fdb	NORF08		; 2
	.fdb	DMAR
	.fdb	NORF08		; 4
	.fdb	DMAW		; 5 (combination, write)
	.fdb	NORF08		; 6
	.fdb	NORF08		; 7
	.fdb	No_op		; 8
	.fdb	No_op		; 9

IOP61:	.fdb	NORF08		; 0
	.fdb	DCIM		; 1
	.fdb	DSAC		; 2
	.fdb	NORF08		; 3
	.fdb	NORF08		; 4
	.fdb	DIML		; 5 (combination, clear and load disk EMA register)
	.fdb	DIMA		; 6 (combination, 0 -- AC, load AC with disk EMA)
	.fdb	NORF08		; 7
	.fdb	No_op		; 8
	.fdb	No_op		; 9

IOP62:	.fdb	NORF08		; 0
	.fdb	DFSE		; 1 (opposite!)
	.fdb	DFSC		; 2
	.fdb	DISK		; 3 new
	.fdb	NORF08		; 4
	.fdb	NORF08		; 5
	.fdb	DMAC		; 6 (combination, 0 -- AC, load AC with DMAR)
	.fdb	NORF08		; 7
	.fdb	No_op		; 8
	.fdb	No_op		; 9

IOP63:	.fdb	NORF08		; 0
	.fdb	NORF08		; 1
	.fdb	NORF08		; 2
	.fdb	NORF08		; 3
	.fdb	NORF08		; 4
	.fdb	NORF08		; 5
	.fdb	NORF08		; 6
	.fdb	NORF08		; 7
	.fdb	No_op		; 8
	.fdb	No_op		; 9

IOP64:	.fdb	NORF08		; 0
	.fdb	DCXA		; 1 new
	.fdb	NORF08		; 2
	.fdb	DXAL		; 3 new
	.fdb	NORF08		; 4
	.fdb	DXAC		; 5 new
	.fdb	DMMT		; 6 new
	.fdb	NORF08		; 7
	.fdb	df_ini		; 8 = initialise
	.fdb	df_rst		; 9 = reset

	.else

IOP60:	.fdb	NODF32		; 0
	.fdb	DCMA		; 1
	.fdb	NODF32		; 2
	.fdb	DMAR
	.fdb	NODF32		; 4
	.fdb	DMAW		; 5 (combination, write)
	.fdb	NODF32		; 6
	.fdb	NODF32		; 7
	.fdb	No_op		; 8
	.fdb	No_op		; 9

IOP61:	.fdb	NODF32		; 0
	.fdb	DCEA		; 1
	.fdb	DSAC		; 2
	.fdb	NODF32		; 3
	.fdb	NODF32		; 4
	.fdb	DEAL		; 5 (combination, clear and load disk EMA register)
	.fdb	DEAC		; 6 (combination, 0 -- AC, load AC with disk EMA)
	.fdb	NODF32		; 7
	.fdb	No_op		; 8
	.fdb	No_op		; 9

IOP62:	.fdb	NODF32		; 0
	.fdb	DFSE		; 1
	.fdb	DFSC		; 2
	.fdb	NODF32		; 3
	.fdb	NODF32		; 4
	.fdb	NODF32		; 5
	.fdb	DMAC		; 6 (combination, 0 -- AC, load AC with DMAR)
	.fdb	NODF32		; 7
	.fdb	No_op		; 8
	.fdb	No_op		; 9

IOP63:	.fdb	NODF32		; 0
	.fdb	NODF32		; 1 = TAS (maintenance instruction, TTA simulator)
	.fdb	NODF32		; 2 = TBS (maintenance instruction, TTB simulator
	.fdb	NODF32		; 3
	.fdb	NODF32		; 4 = DBRS (maint. instruction, data brk request FF)
	.fdb	NODF32		; 5
	.fdb	NODF32		; 6
	.fdb	NODF32		; 7
	.fdb	No_op		; 8
	.fdb	No_op		; 9

IOP64:	.fdb	NODF32		; 0
	.fdb	NODF32		; 1
	.fdb	NODF32		; 2
	.fdb	NODF32		; 3
	.fdb	NODF32		; 4
	.fdb	NODF32		; 5
	.fdb	NODF32		; 6
	.fdb	NODF32		; 7
	.fdb	df_ini		; 8 = initialise
	.fdb	df_rst		; 9 = reset

	.endif



;    ____________________________________________________________________________
;   ##										 ##
;   ##    implementation of routines to support the disk access functionality    ##
;   ##___________________________________________________________________________##


	.if	ide2pdp

;  dskhss - disk access handshake start between IDE and FIRQ routines
;
;  call   : -
;  return : 1. FIRQ routine has turned on the BRK LED
;	    2. the 6809 FIRQ flag is set (FIRQ disabled)

dskhss:	pshs	a
	lda	#0b10000000	; flag IDE access to FIRQ routine
	sta	*accide
10$:	lda	*accide
	bita	#0b00000001	; wait for ACK from FIRQ routine
	beq	10$
	puls	a
	orcc	#0b01000000	; set FIRQ flag
	rts



;  dskhsc - disk handshake complete between IDE and FIRQ routines
;
;  call   : -
;  return : 1. access IDE flag is cleared
;	    2. the 6809 FIRQ flag is cleared (FIRQ enabled)

dskhsc:	clr	*accide		; finish access IDE cycle
	andcc	#0b10111111	; clear FIRQ flag
	rts




;	   _____________________________
;	***				***
;	***  RF08 / DF32  IOT routines  ***
;	***_____________________________***


df_ini:	jsr	GetFlg		; allocate interrupt flag for RF08/DF32 device
	std	*DFirqF		; remember allocated bit mask
df_rst:	lda	#dfNIBF
	sta	*secflg		; nothing in IDE disk buffer
	jsr	df_cli		; RF08/DF32 RESET - clear interrupt request flag

	.if	RF08dev
	clra
	clrb
	std	*rf_irq		; RF08: clear Error & Completion IRQ enable flags
	std	*df_dma		; - disk memory address [11:0]
	std	*df_dmx		; - disk address extension [11:4]
	sta	*df_ema		; - core memory address extension [3:1]
	std	*rf_sts		; - clear all status bits
	rts			;   (return with ccZ=0 --> init OK)

	.else
	lda	#df_TRC
	sta	df_irq		; DF32 - TRansfer Completed flag is set
	clra
	clrb
	std	*df_dma		; - disk memory address [11:0]
	std	*df_dmx		; - disk address extension [11:4]
	sta	*df_ema		; - core memory address extension [3:1]
	sta	*df_sts		; - status is OK
	rts			;   (return with ccZ=0 --> init OK)

	.endif



   ; DCMA [6601] - clear disk memory address register, parity error/completion flags

DCMA:	clra
	clrb
	std	*df_dma		; clear disk memory address

	.if	RF08dev
	ldd	*rf_sts
	anda	#255-rf_TRC		     ; clear "housekeeping" Complete flag
	andb	#255-rf_PER-rf_NXD-rf_DRL    ; clear status bits only
	std	*rf_sts			     ; not core & disk extension bits

	.else
	sta	*df_sts		; DF32: clear PER and EWL error flags
	sta	*df_irq		; 	clear TRC and NED interrupt flags
	.endif

	jsr	df_cli		; clear simulator interrupt request flag
	jmp	NextI



   ; DMAR [6603] - load the disk memory address register from the accumulator.
   ;		   Clear AC. Clear parity error and interrupt completion flags.
   ;		   Start read operation from disk into the specified core location.

DMAR:	pshs	x		; save simulator environment [X]
	.if	idetrc
	ldx	#p_dmar
	jsr	cyc3br		; print the 3-cycle data break information	
	.endif
	ldd	*AC
	anda	#Wmask		; defensive code: make sure [15:12] is "0000"
	std	*df_dma
	clra
	clrb
	std	*AC

	.if	RF08dev
	ldd	*rf_sts
	anda	#255-rf_TRC		     ; clear "housekeeping" Complete flag
	andb	#255-rf_PER-rf_NXD-rf_DRL    ; clear status bits only
	std	*rf_sts			     ; not core & disk extension bits

	.else
	sta	*df_sts		; DF32: clear PER and EWL error flags
	sta	*df_irq		; 	clear TRC and NED interrupt flags
	.endif

	jsr	df_cli		; clear simulator interrupt request flag


;  The address of the word to be read from disk is a combination of the disk memory
;  extension (DF32: 5 bits, RF08: 8 bits) and the disk memory address (12 bits).
;  These 17 / 20 (DF32/RF08) bits allow for the maximum capacity of 4 DS32 Disk File
;  or 4 RS08 Disk units which is 128K / 1024K words (DF32/RF08).
;  Since the sector of an IDE disk drive holds 256 words, the lower 8 bits of the
;  17/20 bit word address is simply the word address in the 256 word IDE disk buffer.
;  The upper 9/12 bits is the sector number of the IDE disk, stored in <secnum>.
;  <secflg> bit [7] indicates whether that sector is in the IDE buffer or not (set).

	ldd	*df_dmx		; get disk extension bits [11:4]
	addb	*df_dma		; lower 4 bits sector number == upper 4 bits df_dma
	std	*secnum		; set IDE sector number to start reading from
	lda	#dfNIBF		; set "rf08/df32 'sector' not in IDE buffer"
	sta	*secflg
;
	bsr	dskhss		; disk access handshake start IDE / FIRQ routines
;
	.if	DEFER_IO
	ldx	#deftab
	lda	#1
	sta	,x		; deftab[RF08].active := TRUE
	ldb	#DEFCNT
	stb	1,x		; deftab[RF08].count := DEFCNT
	ldd	#dmard
	std	2,x		; deftab[RF08].IO_tfr := &dmard
	lda	#1
	sta	*defer		; deferred_IO := TRUE
	puls	x
	jmp	NextI

	.else
	.endif

dmard:	lda	*rf8inf		; display RF08 "registers" ON?
	beq	dmar00		; n - 
	lda	#'R		; y - display RF08 parameters
	jsr	rf8reg
dmar00:				; READ loop
	lda	*df_dmx		; test if disk address <= max RF08/DF32 capacity
	.if	RF08dev
	anda	#0b11110000	; bits [15:12] must be '0', else overflow
	.else
	anda	#0b11111110	; bits [15:9] must be '0', else overflow
	.endif
; --	bne	dmarw1
	ldd	wc3cdb
	addd	#1		; increment Word Count (in 2-complement!)
	anda	#Wmask		; 12-bit max
	std	wc3cdb
	ldd	ca3cdb
	addd	#1		; initial Current Address is 1 "too low"
	anda	#Wmask		; 12-bit wrap around in current Field
	std	ca3cdb
;
	ldb	*df_ema		; get core memory extension bits [3:1]
	ldx	#GetFld
	ldd	b,x		; get start address of current Field
	beq	dmarw3		; current Field = NXM
	addd	ca3cdb		; -> current pdp8/e memory space address
	addd	ca3cdb		; *2 -> current 6809 memory space address
	tfr	d,y
;
	ldd	*df_dmx		; get disk extension bits [11:4]
	addb	*df_dma		; add lower 4 bits of sector number (upper 4 df_dma)
	tfr	d,x		; set IDE sector number to read from
	ldb	*df_dma+1	; requested word in sector
	jsr	rd32wd		; read one word from DF32 Disk File into memory
	bcs	dmarw1		; - IDE request failed
;
	ldd	*df_dma		; move to next disk word to read
	addd	#1
	std	*df_dma
	cmpa	#0b00010000	; 12-bit overflow occurred ?
	bne	dmar20		; n -
	clr	*df_dma		; y - reset to zero (lower 8 bits are already 0)
	ldd	*df_dmx		;     increment disk extension: to next track/device
	addd	#0x10		;     lowest bit is bit[4]
	std	*df_dmx
dmar20:	ldd	wc3cdb		; Word Count == 0000 ?
	bne	dmar00		; n - more to read


  ; set I/O transfer completion: succesfull or failed

dmarw2:	clra			; - OK
	bra	dmarwx
dmarw1:	lda	#1		; - failed

  ; exit DMAR/DMAW IOT

dmarwx:	sta	*df_rts		; set internal "housekeeping" return status

	.if	DEFER_IO
	bsr	df_rdy		; set device I/O completion flag

	.else			; no latency delay consideration
				; return I/O transfer result status immediately
	.if	RF08dev
	beq	20$		; test df_rts status
	ldd	*rf_sts
	ora	#rf_TRC		; set "housekeeping" Complete flag
	orb	#rf_NXD		; preserve core extension & IRQ enable flags
	std	*rf_sts
	bita	#rf_EIE		; Error Interrupt Enable set?  (bit[8])
	beq	30$		; n -
	bsr	df_sei		; y - set RF08/DF32 allocated interrupt flag
	bra	30$

20$:	ldd	*rf_sts
	ora	#rf_TRC		; set "housekeeping" Complete flag
	sta	*rf_sts
	bitb	#rf_CIE		; Completion Interrupt Enable set?
	beq	30$		; n -
	bsr	df_sei		; y - set RF08/DF32 allocated interrupt flag
30$:

	.else
	beq	20$		; test df_rts status
	lda	#df_EWL
	sta	*df_sts		; failed - set in DF32 status register
	lda	#df_NED+df_TRC	;	 - set non-existent disk
	bra	30$		;	   and completion interrupt bit

20$:	lda	#df_TRC		; OK - set transfer completed interrupt flag
30$:	sta	*df_irq
	bsr	df_sei		; set RF08/DF32 allocated interrupt flag
	.endif			; RF08/DF32 result status handling
	.endif			; DEFER_IO

	jsr	dskhsc		; disk access handshake complete IDE / FIRQ routines
	.if	idetrc
	jsr	rw_sts		; print return status
	.endif

	.if	DEFER_IO
	clra			; return status = OK
	rts
	.else
	puls	x		; restore simulator environment [X]
; --	jmp	NextI		; just move to the next pdp simulated instruction
	jmp	ION		; to next pdp8 instruction if interrups are disabled
				; or create 1 instruction delay if IRQ's are enabled
	.endif



   ; df_cli - clear simulator interrupt request flag for DF32/RF08

df_cli:	ldd	*DFirqF		; get the allocated interrupt request flag
	coma
	comb
	anda	*IOflgs		; clear RF08/DF32 allocated interrupt request flag
	andb	*IOflgs+1
	std	*IOflgs
	rts


   ; df_sei - set simulator interrupt request flag for DF32/RF08

df_sei:	ldd	*DFirqF		; get the allocated interrupt request flag
	ora	*IOflgs
	orb	*IOflgs+1
	std	*IOflgs		; set RF08/DF32 allocated interrupt flag
	rts


	.if	DEFER_IO
	.if	RF08dev

   ; df_rdy - set I/O completion status for RF08

df_rdy:	lda	*df_rts		; get result of DMAR/DMAW :  0=ok  / <>0=fail
	bne	10$
	ldd	*rf_sts
	ora	#rf_TRC		; set "housekeeping" Complete flag
	sta	*rf_sts
	bitb	#rf_CIE		; Completion Interrupt Enable set?
	beq	30$		; n -
	bra	df_sei		; y - set RF08/DF32 allocated interrupt flag and RTS

10$:	ldd	*rf_sts
	ora	#rf_TRC
	orb	#rf_NXD		; preserve core extension & IRQ enable flags
	std	*rf_sts
	bita	#rf_EIE		; Error Interrupt Enable set?  (bit[8])
	beq	30$		; n -
	bra	df_sei		; y - set RF08/DF32 allocated interrupt flag and RTS
30$:	rts

	.else

   ; df_rdy - set I/O completion status for DF32

df_rdy:	lda	*df_rts		; get result of DMAR/DMAW :  0=ok  / <>0=fail
	bne	10$
	lda	#df_TRC		; ok	 - set transfer completed interrupt flag
	bra	20$
10$:	lda	#df_EWL		; failed - set flag in DF32 status register
	sta	*df_sts
	lda	#df_NED+df_TRC	;	   set non-existent disk and completion
20$:	sta	*df_irq		;	   interrupt bits
	bra	df_sei		; set DF32 allocated interrupt flag and RTS
	.endif			; RF08 / DF32
	.endif			; DEFER_IO



    ; access attempted to non-existing core: hard error exit

dmarw3:	jsr	dskhsc		; disk access handshake complete IDE / FIRQ routines
	.if	idetrc
	jsr	rw_sts		; print return status
	.endif

	.if	DEFER_IO

	lda	#1		; return status = NOK
	rts

	.else

	puls	x
	jmp	Flderr

	.endif




   ; DMAW [6605] - load the disk memory address register from the accumulator.
   ;		   Clear AC. Clear parity error and interrupt completion flags.
   ;		   Start write operation to disk from the specified core location.

DMAW:	pshs	x		; save simulator environment [X]
	.if	idetrc
	ldx	#p_dmaw
	jsr	cyc3br		; print the 3-cycle data break information
	.endif
	ldd	*AC
	anda	#Wmask		; defensive code: make sure [15:12] is "0000"
	std	*df_dma
	clra
	clrb
	std	*AC

	.if	RF08dev
	ldd	*rf_sts
	anda	#255-rf_TRC		     ; clear "housekeeping" Complete flag
	andb	#255-rf_PER-rf_NXD-rf_DRL    ; clear status bits only
	std	*rf_sts			     ; not core & disk extension bits

	.else
	sta	*df_sts		; DF32: clear PER and EWL error flags
	sta	*df_irq		; 	clear TRC and NED interrupt flags
	.endif

	bsr	df_cli		; clear simulator interrupt request flag

;  The address of the word to be written to disk is a combination of the disk memory
;  extension and the disk memory address (12 bits).
;  The disk memory extension is 5 bits for the DF32, and it is 8 bits for the RF08.
;  These 17 (DF32) or 20 (RF08) bits allow for the maximum capacity of 4 RS32/DS32
;  Disk File units which is 128K words for DF32 and 1024K words for the RF08.
;  Since the sector of an IDE disk drive holds 256 words, the lower 8 bits of the 17
;  or 20 bit word address is simply the word address in the 256 word IDE disk buffer.
;  The upper 9 bits (DF32) or 12 bits (RF08) form the sector number of the IDE disk
;  and is stored in <secnum>.
;  <secflg> bit [7] indicates whether that sector is in the IDE buffer or not (set),
;  and bit [6] indicates whether the contents of the IDE buffer is changed.

	ldd	*df_dmx		; get disk extension bits [11:4]
	addb	*df_dma		; lower 4 bits sector number == upper 4 bits df_dma
	std	*secnum		; set IDE sector number to start writing to
	lda	#dfNIBF		; set "df32 'sector' not in IDE buffer"
	sta	*secflg
;
	jsr	dskhss		; disk access handshake start IDE / FIRQ routines

	.if	DEFER_IO
	ldx	#deftab
	lda	#1
	sta	,x		; deftab[RF08].active := TRUE
	ldb	#DEFCNT
	stb	1,x		; deftab[RF08].count := DEFCNT
	ldd	#dmawd
	std	2,x		; deftab[RF08].IO_tfr := &dmawd
	lda	#1
	sta	*defer		; deferred_IO := TRUE
	puls	x
	jmp	NextI

	.else
	.endif

dmawd:	lda	*rf8inf		; display RF08 "registers" ON?
	beq	dmaw00		; n - 
	lda	#'W		; y - display RF08 parameters
	jsr	rf8reg
dmaw00:				; WRITE loop
	lda	*df_dmx		; test if disk address <= max RF08/DF32 capacity
	.if	RF08dev
	anda	#0b11110000	; bits [15:12] must be '0', else overflow
	.else
	anda	#0b11111110	; bits [15:9] must be '0', else overflow
	.endif
; --	bne	dmarw1		; -> else overflow
	ldd	wc3cdb
	addd	#1		; increment Word Count (in 2-complement!)
	anda	#Wmask		; 12-bit max
	std	wc3cdb
	ldd	ca3cdb
	addd	#1		; initial Current Address is 1 "too low"
	anda	#Wmask		; 12-bit wrap around in current Field
	std	ca3cdb
;
	ldb	*df_ema		; get core memory extension bits [3:1]
	ldx	#GetFld
	ldd	b,x		; get start address of current Field
	beq	dmarw3		; current Field = NXM
	addd	ca3cdb		; -> current pdp8/e memory space address
	addd	ca3cdb		; *2 -> current 6809 memory space address
	tfr	d,y
;
	ldd	*df_dmx		; get disk extension bits [11:4]
	addb	*df_dma		; add lower 4 bits of sector number (upper 4 df_dma)
	tfr	d,x		; set IDE sector number to read from
	ldb	*df_dma+1	; requested word in sector
	jsr	wr32wd		; write one word from core memory to DF32 Disk File
	lbcs	dmarw1		; - IDE request failed
;
	ldd	*df_dma		; move to next disk word to write
	addd	#1
	std	*df_dma
	cmpa	#0b00010000	; 12-bit overflow occurred ?
	bne	dmaw20		; n -
	clr	*df_dma		; y - reset to zero (lower 8 bits are already 0)
	ldd	*df_dmx		;     increment disk extension: to next track/device
	addd	#0x10		;     lowest bit is bit[4]
	std	*df_dmx
dmaw20:	ldd	wc3cdb		; Word Count == 0000 ?
	bne	dmaw00		; n - more to write
	ldy	#0		; y - flush dirty buffer to IDE disk
	jsr	wr32wd
	lbcs	dmarw1		; -> return with error status
	jmp	dmarw2		; -> return with status OK



   ; DCIM [6611] - RF08: clear memory address extension reg, clear interrupt enable
   ; DCEA [6611] - DF32: clear memory address extension and disk extended address reg

	.if	RF08dev
DCIM:	clr	*df_ema		; core memory extension address register (3 bits)
	ldd	*rf_sts
	anda	#255-rf_IER-rf_TRC   ; clear Error Interrupt Enable flag & Complete
	andb	#255-rf_ICL-rf_ICO   ; clear Interrupt Enable Clock and Completion
	std	*rf_sts
	jsr	df_cli		; clear simulator RF08/DF32 interrupt request flag
	bra	jmpNxI

	.else
DCEA:	clra
	clrb
	std	*df_dmx		; disk extended address register DF32: 5 bits [8:4]
	sta	*df_ema		; core memory extension address register (3 bits)
	bra	jmpNxI
	.endif



   ; DSAC [6612] - skip next instruction if the address confirmed flag is a 1.
   ;		   Flag is set for 16 microseconds. AC is cleared.
   ;		   Address confirmed flag is primarily used in diagnostic programming.
   ;		   For this implementation the flag is always clear, no skip is done.

DSAC:	clra
	clrb
	bra	svACnx



   ; DIML [6615] - RF08: load memory address extension register, then clear AC.
   ;			 Set interrupt enable. Disk extended address is *not* set,
   ;			 see DXAL. AC bit assignment is different from DF32 DEAL.
   ;
   ;	___0_____1_____2_____3_____4_____5_____6_____7_____8_____9____10____11___
   ;	|#####|#####|#####| IER | ICL | ICO | EM2 | EM1 | EM0 |#####|#####|#####|
   ;	|_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|
   ;
   ;		IER - Interrupt Enable Error	EM2 \
   ;		ICL - Interrupt Enable Clock	EM1  >  Core Field Extension bits
   ;		ICO - Interrupt Enable Comp	EM0 /
   ;
   ;
   ; DEAL [6615] - DF32: load memory address extension and disk extended address
   ;			 registers with the track address data held in the AC.
   ;
   ;	___0_____1_____2_____3_____4_____5_____6_____7_____8_____9____10____11___
   ;	|#####| EMA4| EMA3| EMA2| EMA1| EMA0| EM2 | EM1 | EM0 |#####|#####|#####|
   ;	|_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|
   ;
   ;		EMA4 \   disk memory		EM2 \
   ;		 ::   >  extended address	EM1  > Core Field extension bits
   ;		EMA0 /	 bits			EM0 /

	.if	RF08dev
DIML:	ldb	*AC+1
	lsrb			; move EM2-EM1-EM0 to bits [3:1]
	lsrb			; (multiplied by 2 for 6809 -> pdp8 memory space)
	andb	#0b00001110
	stb	*df_ema
	ldd	*rf_sts		      ; clear the Interrupt Enable flags in status
	andb	#255-rf_ICL-rf_ICO    ; - Completion and Clock (not used)
	anda	#255-rf_IER-rf_TRC    ; - Error, Complete ("internal housekeeping")
	std	*rf_sts
	ldd	*AC		; isolate the Interrupt Enable flags
	andb	#rf_ICL+rf_ICO	; - Completion and Clock (not used)
	anda	#rf_IER		; - Error
	orb	*rf_sts+1
	ora	*rf_sts
	std	*rf_sts		; set Interrupt Enable flag in status
	bra	DSAC		; clear AC, goto Next Instruction

	.else

DEAL:	ldb	*AC+1
	lsrb			; move EM2-EM1-EM0 to bits [3:1]
	lsrb			; (multiplied by 2 for 6809 -> pdp8 memory space)
	andb	#0b00001110
	stb	*df_ema
	ldd	*AC
	lsra			; move EMA5:EMA1 to bits [8:4]
	rorb
	lsra
	rorb
	anda	#0b00000001
	andb	#0b11110000
	std	*df_dmx
	bra	jmpNxI
	.endif



   ; DIMA [6616] - RF08: Store status register in AC for program evaluation. Similar
   ;			 to DSAC, however, MB9 is a '1' , inhibit skip on Address
   ;			 Confirmed flag.   Functions transmitted to AC are different
   ;			 compared with DEAC; see DXAC.
   ;
   ;	___0_____1_____2_____3_____4_____5_____6_____7_____8_____9____10____11___
   ;	| PCA | DRE | WLS | EIE | PIE | CIE | EM2 | EM1 | EM0 | DRL | NXD | PER |
   ;	|_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|
   ;
   ;		PCA - Photo Cell Sync mark	   DRE - Data Request Enable (maint)
   ;		WLS - Write Lock Status		   EIE - Error Interrupt Enable
   ;		PIE - Photocell Interrupt Enable   CIE - Completion Interrupt Enable
   ;		EM2 \				   DRL - Data Request Late
   ;		EM1  > Core Field extension bits   NXD - Non Existing Disk
   ;		EM0 /				   PER - Parity Error
   ;
   ;
   ; DEAC [6616] - DF32: load contents of disk extended address register into AC.
   ;			 Skip next instruction if address confirmed flag is a 1.
   ;			 Address confirmed flag is primarily used in diagnostic
   ;			 programming. In this implementation the flag is always
   ;			 clear, no skip is done.
   ;
   ;	___0_____1_____2_____3_____4_____5_____6_____7_____8_____9____10____11___
   ;	| PHC | EMA4| EMA3| EMA2| EMA1| EMA0| EM2 | EM1 | EM0 | DRL | NED | PER |  AC
   ;	|_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|_____|
   ;
   ;		PHC - Photo Cell Sync Mark (200 s)
   ;		EMA4 \   disk memory		     EM2 \
   ;		 ::   >  extended address	     EM1  > Core Field extension bits
   ;		EMA0 /	 bits			     EM0 /
   ;		DRL - Data Request Late		     NED \  [ NonExisting Disk ]
   ;		PER - Parity Error		     EWL /  [ Error Write Lock ]

	.if	RF08dev
DIMA:	ldd	*rf_sts
	anda	#255-rf_TRC	; remove "housekeeping" Complete flag
	bra	svACnx

	.else

DEAC:	ldd	*df_dmx		; get disk address extension bits [8:4]
	lslb			; and put EMA5:EMA1 in place, bits [10:6]
	rola
	lslb
	rola
	anda	#0b00000111	; PhotoCell SYNC & high nibble always zero, bit [15:12]
	andb	#0b11000000
	stb	*Temp1
	ldb	*df_ema		; get memory address extension bits
	lslb			; and put EM2:EM0 in place, bits [5:3]
	lslb
	andb	#0b00111000
	addb	*Temp1
	stb	*Temp1
	ldb	*df_sts
	andb	#df_PER+df_EWL+df_DRL	; get the error bits [2:0]
	addb	*Temp1
svACnx:	std	*AC
jmpNxI:	jmp	NextI
	.endif



   ; DFSE [6621] - skip next instruction if -
   ;		   RF08: *any* error bit is set
   ;		   DF32: *all* error bits are clear
   ;			(error bits are:
   ;				PER - parity error
   ;				DRL - data request late
   ;				EWL - write lock switch

DFSE:
	.if	RF08dev
	ldb	*rf_sts+1		; get RF08 status
	andb	#rf_PER+rf_NXD+rf_DRL	; any error bit set ?
	bne	jPCskp			; RF08: skip on any error
	bra	jmpNxI

	.else
	lda	*df_sts			; get DF32 status
	anda	#df_PER+df_EWL+df_DRL	; any error bit set ?
	bne	jmpNxI
jPCskp:	jmp	PCskip			; DF32: skip on no errors
	.endif



   ; DFSC [6622] - skip next instruction if the completion flag is a 1
   ;		   (transfer complete).

DFSC:
	.if	RF08dev
	lda	*rf_sts		; test "internal housekeeping" Complete flag (bit 15)
	bmi	jPCskp
	bra	jmpNxI

	.else
	lda	*df_irq		; test TRC Transfer Completed flag (msb, bit 7)
	bmi	jPCskp
	bra	jmpNxI
	.endif



   ; DMAC [6626] - load the contents of the disk memory address register into the AC.
   ;		   During read, final address will be the last address transferred +1
   ;		   during write, final address will be the last address transferred.
   ;		   DMAC is a combination of 6624 . 6622 (DFSC)

DMAC:	ldd	*df_dma
	std	*AC
	bra	DFSC



	.if	RF08dev		; new RF08 only IOT's

   ; DISK [6623] - skip next instruction on error or data completion flag or both
   ;		   (a combination of DFSE and DFSC).

DISK:	lda	*rf_sts		; test "internal housekeeping" Complete flag (bit 15)
	bmi	jPCskp
	ldb	*rf_sts+1		; get RF08 status
	andb	#rf_PER+rf_NXD+rf_DRL	; any error bit set ?
	beq	jmpNxI			; n -
jPCskp:	jmp	PCskip



   ; DCXA [6641] - Clears the 8 extended disk address bits.

DCXA:	clra			; clear 4 upper bits [11:8]
	clrb			; clear 4 lower bits [7:4]
	std	*df_dmx
	bra	jmpNxI



   ; DXAL [6643] - Load the 8 extended disk address bits with bits [4-11] of the AC.
   ;		   Data in bits 0 through 3 of AC are ignored. Then AC is cleared.
   ; DMMT [6646] - For Maintenance purposes only. AC is cleared.

DXAL:	ldd	*AC
	lslb			; move AC bits [4-11] 4 bit positions left
	rola			; to store them in df_dmx [11:4]
	lslb
	rola
	lslb
	rola
	lslb
	rola
	anda	#0b00001111	; defensive code: make sure upper 4 bits are zero
	std	*df_dmx
DMMT:	clra
	clrb
svACnx:	std	*AC
jmpNxI:	jmp	NextI



   ; DXAC [6645] - Load the AC bits [4-11] with the 8 extended disk address bits.

DXAC:	ldd	*df_dmx
	lsra			; move df_dmx [11:4] back 4 bit positions right
	rorb			; to store then in AC [4-11]
	lsra
	rorb
	lsra
	rorb
	lsra
	rorb
	bra	svACnx		; save AC and goto NextI



NORF08:	jsr	Errpnt
	.fdb	niRF08

	.else

NODF32:	jsr	Errpnt
	.fdb	niDF32

	.endif




;  * rd32wd - read one word from RF08/DF32 Disk File
;
;  call   : [Y]      = core memory address to store word into
;	    [X]      = requested sector number
;	    [B]      = requested word offset in sector
;	    <secnum> = current sector in IDE buffer
;	    <secflg> = "not in buffer" flag [dfNIBF]
;  return : cC = 1 :: request failed
;	    cC = 0 :: request succesfull completed
;	    - if requested word/sector is not in the IDE buffer
;		the sector is read, <secnum> is updated, "not in buffer" flag cleared
;	    - word read from buffer stored into core memory
;	    <Temp1> and <Temp2> destroyed
;
;  pseudo code:
;	if (requested_sector != current_sector)
;	   current_sector := requested_sector
;	   sector_flags := [dfNIBF]
;	if (sector_flags[dfNIBF] == 1)
;	   sector_flags := 0
;	   read_IDE_sector(current_sector)
;	read word from (current_sector + requested_word_offset)
;	store word in (core_memory_address)

rd32wd:	stb	*Temp1		; save word offset
	stx	*Temp2
	ldd	*secnum		; get current sector number
	cmpd	*Temp2		; requested sector# == current sector# ?
	beq	10$		; y -
	ldd	*Temp2		; n - current sector# := requested sector#
	std	*secnum
	lda	#dfNIBF		;     set "not in buffer" flag
	sta	*secflg
10$:	lda	*secflg		; get sector flags
	bita	#dfNIBF		; is sector in the IDE buffer ?
	beq	20$		; y -
	clr	*secflg		; n - clear "not in buffer" flag
				;     read the sector [D] from IDE disk
	bsr	SetPkt		;     build the IDE command parameter packet
	jsr	rdDsec		;     read sector
	bcc	20$		;     check IDE status: OK ->
	jsr	DskErr		;		error occurred -> report IDE error
	orcc	#0b00000001	;
	rts
;
20$:	ldx	#IDEbuf		; get base of IDE buffer
	ldb	*Temp1		; get word offset in buffer
	abx
	abx
	ldd	,x		; get word from "DF32 Disk File"
	std	,y		; store in pdp8/e core memory space
	andcc	#0b11111110
	rts


SetPkt:	ldx	#CmdPkt		; build the IDE command parameter packet
	ldd	*secnum		; get current sector#
	addd	*secoff+2	; add "disk #" offset LSB
	std	4,x		; set sector number lower 16-bit in selected "disk #"
	ldd	#0		; upper 16-bit always 0 (for DF32 or RF08)
	adcb	*secoff+1	; add "disk #" offset MSB
	adca	*secoff
	std	2,x		; set sector number upper 16-bit in selected "disk #"
	ldd	#IDEbuf
	std	,x		; set buffer address
	rts



;  * wr32wd - write one word to RF08/DF32 Disk File
;
;  call   : [Y]      = core memory address to read word from
;			if [Y] = 0000 -> flush buffer to IDE disk request
;	    [X]      = destination sector number if [Y] <> 0000
;	    [B]      = word offset in sector to write if [Y] <> 0000
;	    <secnum> = current sector in IDE buffer
;	    <secflg> = "not in buffer" flag     [dfNIBF]  (optional)
;		       "IDE buffer dirty" flag  [dfDIRT]  (optional)
;  return : cC = 1 :: request failed
;	    cC = 0 :: request succesfull completed
;	    if [Y] <> 0000
;	    - if requested sector to write the word into is not in the IDE buffer
;		the sector is read, <secnum> is updated, "not in buffer" flag cleared
;	    - word from core memory written into buffer + "dirty" bit set [dfDIRT]
;	    if [Y] == 0000
;	    - buffer in memory written to current sector <secnum>
;	    <Temp1> and <Temp2> destroyed
;
;  pseudo code:
;    if [Y] != 0000
;	if (requested_sector != current_sector)
;	   // before reading requested_sector, write current_sector to disk if dirty
;	   if (sector_flags[dfDIRT] == 1)
;	      write_IDE_sector(current_sector)
;	   current_sector := requested_sector
;	   sector_flags[dfNIBF] := 1
;	if (sector_flags[dfNIBF] == 1)
;	   read_IDE_sector(current_sector)
;	   sector_flags[dfNIBF] := 0
;	write word from (core_memory_address) to (current_sector + word_offset)
;	sector_flags[dfDIRT] := 1
;    else
;       // [Y] == 0000  --> flush current_sector to disk if necessary
;       if (sector_flags[dfDIRT] == 1)
;	   write_IDE_sector(current_sector)
;	   sector_flags[dfDIRT] := 0

wr32wd:	cmpy	#0		    ; flush buffer to IDE disk ?
	beq	wrflsh		    ; y -
	stb	*Temp1		    ; save word offset
	stx	*Temp2
	ldd	*secnum		    ; get current sector number
	cmpd	*Temp2		    ; requested sector# == current sector# ?
	beq	20$		    ; y -
	lda	*secflg		    ; n - is current sector "dirty" ?
	bita	#dfDIRT
	beq	10$		    ;	  n -
				    ;	  y - write the sector to IDE disk
	bsr	SetPkt		    ;	      build the IDE command parameter packet
	jsr	wrDsec		    ;	      write sector
	bcs	wr_nok		    ;	      IDE status: error occurred, report it

10$:	ldd	*Temp2
	std	*secnum		    ; current sector := requested sector
	lda	#dfNIBF		    ; mark current sector not in buffer!
	sta	*secflg

20$:	lda	*secflg		    ; get sector flags
	bita	#dfNIBF		    ; is sector in the IDE buffer ?
	beq	30$		    ; y -
	clr	*secflg		    ; n - clear flags

	bsr	SetPkt		    ; build the IDE command parameter packet
	jsr	rdDsec		    ; read sector
	bcs	wr_nok		    ; check IDE status: error occurred, report it
30$:	ldx	#IDEbuf		    ; get base of IDE buffer
	ldb	*Temp1		    ; get word offset in buffer
	abx
	abx
	ldd	,y		    ; get word from pdp8/e core memory space
	std	,x		    ; store in "DF32 Disk File"
	lda	*secflg
	ora	#dfDIRT		    ; mark buffer "dirty"
	sta	*secflg
	andcc	#0b11111110
	rts
;
wrflsh:	lda	*secflg		; get sector flags
	bita	#dfDIRT		; buffer contents changed ("dirty") ?
	beq	wrflok		; n - all done
	clr	*secflg		; y - clear flags

	bsr	SetPkt		; build the IDE command parameter packet
	jsr	wrDsec		; write sector
	bcc	wrflok		; check IDE status: OK ->
wr_nok:	jsr	DskErr		;       error occurred -> report IDE error
	orcc	#0b00000001	;
	rts
wrflok:	andcc	#0b11111110
	rts


	.if	idetrc

;  * cyc3br - print RF08 DF32 IOT id and the 3-cycle data break option parameters
;
;    call   : [X] = ptr to IOT identification (starting with new line)
;    return : text / data <wc3cdb & ca3cdb> printed, no new line

cyc3br:	jsr	docrlf
	bsr	pdata1		; print IOT identification (DMAR / DMAW)
	ldx	#WCval
	bsr	pdata1
	ldx	#wc3cdb
	jsr	OUT4HS		; print Word Count
	ldx	#CAval
	bsr	pdata1
	ldx	#ca3cdb
	jmp	OUT4HS		; print Current Address and RTS


;  * rw_sts - print RF08/DF32 DMAR/MAW IOT return status
;
;    call   : -
;    return : text / data <df_irq & df_sts> printed, no new line

rw_sts:	ldx	#rw_s1
	bsr	pdata1
	lda	*df_rts		; is return status OK ?
	beq	pdata1		; yes - print OK and RTS
	leax	3,x		; no  - skip OK message, print FAIL and RTS
pdata1:	jmp	pdatao


p_dmar:	.asciz	"DMAR"
p_dmaw:	.asciz	"DMAW"
WCval:	.asciz	"  wc3cdb=$"
CAval:	.asciz	" ca3cdb=$"
rw_s1:	.asciz	" ... df_rts: "
	.asciz	"OK"
	.asciz	"FAIL"
	.endif			; end idetrc debug output

	.else			; no IDE disk implemented: dummy subroutines / IOT's

df_ini:
df_rst:	clra
	rts


	.if	RF08dev

DCMA: DMAR: DMAW: DCIM: DSAC: DIML: DIMA: DFSE:
DFSC: DISK: DMAC: DCXA: DXAL: DXAC: DMMT: NORF08:

	jsr	Errpnt
	.fdb	niRF08

	.else

DCMA: DMAR: DMAW: DCEA: DSAC: DEAL:
DEAC: DFSE: DFSC: DMAC: NODF32:

	jsr	Errpnt
	.fdb	niDF32

	.endif			; RF08 selection or DF32 selection

	.endif			; IDE disk implemented or not






; ===================================================================================
; ##                       RK8-P  Random Access Disk System                        ##
; ##_______________________________________________________________________________##
;
;
;   Introductory information
;  --------------------------
;
; The RK8-P Disk Interface Control can control up to 4 RK01 Disk Drives.  Each Disk
; Drive can hold a removable cartridge which can store up to 831488 12-bit words.
; Write-protection is provided either on sector basis, under program/manual control,
; or on whole disk basis, under manual control.
; Sector numbers can be sequentially located on the disk, or scattered in a track.
; Data transfers are performed using the single-cycle data break facility of the 8/e.
;
;   Specifications
;  ----------------
;
; storage	831488 12-bit words
; tracks	200 + 3
; words/track	4096 (2048 words on #0 and #1 surface)
; sectors	16
; words/sector	256 (data)
; block size	minimum = 256, maximum = 4096
;
; Each disk surface is divided into 203 tracks (200 data and 3 spares). Thus, for 203
; combinations, 8 bits must be allotted to address a track.   Each track is divided
; into 16 sectors with sectors 0-7 (octal) on surface #0, and sectors 10-17 (octal)
; on surface #1.  Four bits (3 for sector and 1 for surface) must be allotted for the
; selection of a sector.  A sector provides storage capacity for:
;     2 12-bit header words
;   256 12-bit data words
;     1 12-bit parity word
; The first header word defines the track and sector address and the second header
; word stores sector protect and status bits. The 12-bit pairty word defines the
; longitudinal parity of the data words.
;
;   RK8-P Disk Interface Control
;  ------------------------------
;
;  - decodes programmed OIT instructions
;  - accepts and stores word count, current address, and command and address words
;  - selects the program-designated RK01 and starts the program-designated operation
;  - jointly with the selected RK01 locates the disk address (track sector & surface)
;  - generates interrupt and break request to initiate a single-cycle data break
;  - buffers the input/output data and performs serial<->parallel conversion
;  - performs housekeeping chores for single-cycle data break transfers (incrementing
;    of word count and current address and providing address for transfers)
;  - generates and checks longitudinal parity for the sectors
;  - provides flags for denoting current conditions and error status
;  - provides logic for aiding in maintenance of an RK8 system
;
; Data transfers between the disk and core memory are implemented using the
; single-cycle data break facility. For this operation, the computer specifies the
; number of words to be transferred (word count), and loads this information into the
; RK8-P word count register by issueing a DLWC instruction. The first core memory
; address involved in the transfer is then loaded into the current address register
; by issueing a DLCA instruction. Next the computer loads a command word defining the
; selected disk, extended memory address (if this option is used), the interrupts it
; will accept, and the type of operation to be performed. This information is loaded
; into the command register by a DLDC instruction. Next the program loads the disk
; address (track, sector, and surface address) into corresponding registers.  This
; information defines disk starting location.
; If a read operation is to be performed, a DLDR instruction is issued to load the
; disk address and start the operation. For a write operation, a DLDW instruction is
; issued.When the correct data track is located, a data break request is generated
; and the data break interface logic assumes control of the transfers. For this phase
; the data break logic controls the loading and conversion of disk data words by
; asserting a B BREAK line to the RK8-P. Information read from the disk is transfered
; to the core memory.
;
; This information is routed via external bus lines DATA00-11 gated by the Data Break
; Interface and supplied to the MB via OMNIBUS lines DATA00-11. For write operations,
; information is accessed at the memory location specified by the current address.
; Data is provided to the RK8-P via the MD00-11 OMNIBUS lines, and supplied to the
; RK8-P via the BMB00-11 external bus lines. The data is loaded into an RK8-P buffer
; under data break control and is shifted out to the disk read/write electronics.
; For each word transfer, the core memory address is specified by the current address
; register, and then the current address and word count registers are incremented.
; When the proper number of words have been transferred, as denoted by word count
; overflow, the RK8-P generates an interrupt request. When the computer honores this
; request, it senses the transfer done flag of the RK8-P with a DSKD instruction.
; If this flag is set, the computer senses the peripheral error flag with a DSKE
; instruction to determine if any errors were recorded during the transfers.
; If not, the transfer is assumed to be valid.
;
;    DLDA [6731] - Load Disk Address (maintenance only)
;	loads the disk address for maintenence or diagnostic functions
;
;    DLDC [6732] - Load Command Register
;	Loads the contents of AC00-11 into the Command Register and clears AC.
;
;	-------------------------------------------------
;	| 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 | 10| 11|   Command Register
;	-------------------------------------------------
;
;	0	- enable change in interrupt status
;	1	- enable program interrupt on transfer done
;	2	- enable interrupt on error
;	3	- unused
;	4	- seek track and surface only
;	5	- enable reading or writing of 2 header words
;	6,7,8	- extended memory address
;	9,10	- disk file number
;	11	- unused
;
;    DLDR [6733] - Load Disk Address and Read
;	Loads track, surface and sector address from AC, the clears AC.
;	Start read from disk if Command Register bit 4 is zero. If bit 4 is a one,
;	instruction is executed only to seek track and surface.
;
;	-------------------------------------------------
;	| 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 | 10| 11|   Track Address Register
;	-------------------------------------------------
;
;	0:7	- track address
;	8	- surface bit
;	9:11	- sector address
;
;    DRDA [6734] - Read Disk Address
;	Clears AC and reads content of Track Address Counter and Surface/Sector
;	counters into AC00-11.
;
;    DLDW [6735] - Load Disk Address and Write
;	Loads track, surface and sector address from AC, then clears AC.
;	Also starts a write operation if Command Register bit 4 is zero. If bit 4
;	is a one, instruction is executed only to seek track and surface.
;
;    DRDC [6736] - Raed Disk Command Register
;	Clears AC, then reads content of Command Register to AC00-11.
;
;    DCHP [6737] - Load Disk Address and Check Parity
;	Loads track, surface and sector address from AC, then clears AC. If bit 4
;	of Command Register is zero, reads data at specified address and checks
;	parity. If bit 4 is one, instruction is executed to seek track and surface.
;
;    DRDS [6741] - Read Disk Status Register
;	Clears AC, then reads content of Status Register into AC.
;
;	-------------------------------------------------
;	| 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 | 10| 11|   Status Register
;	-------------------------------------------------
;
;	0	- error
;	1	- transfer done
;	2	- control busy error
;	3	- time out error
;	4	- parity or timing error
;	5	- data rate error
;	6	- track address error
;	7	- sector no good error
;	8	- write lock error
;	9	- track capacity exceeded error
;	10	- select error
;	11	- busy
;
;    DCLS [6742] - Clear Status Register
;	Clears Status Register.
;
;    DMNT [6743] - Load Maintenance Register
;	Loads the content of AC into Maintenance Register and performs specified
;	operation. Bits remain in Maintenance Register until DMNT is issued with
;	AC bits = 0.
;
;	-------------------------------------------------
;	| 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 | 10| 11|   Maintenance Register
;	-------------------------------------------------
;
;	0	- track addr to track ctr
;	1	- serial reg to data reg
;	2	- data reg to serial reg
;	3	- clear AC, read data reg to AC
;	4	- shift "1" in serial reg
;	5	- shift "0" in serial reg
;	6	- unformatted disk
;	7	- sector pulse
;	8	- index pulse
;	9:11	- not used
;
;    DSKD [6745] - Skip on Disk Done
;	The Disk Transfer Done Flag is sensed; if it is set, the next instruction
;	is skipped.
;
;    DSKE [6747] - Skip on Disk Error
;	The Error Flag is sensed; if it is set, the next instruction is skipped.
;
;    DCLA [6751] - Clear All
;	Clears selected disk to track 000, and clears all control registers and flags
;	except Disk Selection. Transfer Done is set when disk is positioned to
;	track 000.
;
;    DRWC [6752] - Read Word Count Register
;	Clears the AC, the reads contents of WC Register into AC0-11.
;
;    DLWC [6753] - Load Word Count Register
;	Loads the contents of AC into WC Register, the clears AC.
;
;    DLCA [6755] - Load Current Address Register
;	Loads the contents of AC into CA Register, then clears AC.
;
;    DRCA [6757] - Read Current Address Register
;	Clears the AC, then reads the contents of the CA Register, to AC0-11.

	.if	1		; for now, the RK8P does not exist

IOT73:	IOT74:	IOT75:
	cmpa	#7		; reset or initialise ?
	bls	10$		; n - invalid IOT for now
	clra
	rts
10$:	jsr	Errpnt
	.fdb	noRK8P

	.else

IOT73:	ldu	#IOP73		; use an opcode array
	bra	iot7x
IOT74:	ldu	#IOP74
	bra	iot7x
IOT75:	ldu	#IOP75
iot7x:	lsla			; get the IOP code
	jmp	[a,u]


IOP73:	.fdb	No_op		; 0
	.fdb	DLDA		; 1
	.fdb	DLDC		; 2
	.fdb	DLDR		; 3
	.fdb	DRDA		; 4
	.fdb	DLDW		; 5
	.fdb	DRDC		; 6
	.fdb	DCHP		; 7
	.fdb	No_op		; 8
	.fdb	No_op		; 9

IOP74:	.fdb	No_op		; 0
	.fdb	DRDS		; 1
	.fdb	DCLS		; 2
	.fdb	DMNT		; 3
	.fdb	No_op		; 4
	.fdb	DSKD		; 5
	.fdb	No_op		; 6
	.fdb	DSKE		; 7
	.fdb	No_op		; 8
	.fdb	No_op		; 9

IOP75:	.fdb	No_op		; 0
	.fdb	DCLA		; 1
	.fdb	DRWC		; 2
	.fdb	DLWC		; 3
	.fdb	No_op		; 4
	.fdb	DLCA		; 5
	.fdb	No_op		; 6
	.fdb	DRCA		; 7
	.fdb	No_op		; 8 = initialise
	.fdb	No_op		; 9 = reset


DLDA:	DLDC:	DLDR:
DRDA:	DLDW:	DRDC:
DCHP:	DRDS:	DCLS:
DMNT:	DSKD:	DSKE:
DCLA:	DRWC:	DLWC:
DLCA:	DRCA:

	jsr	Errpnt
	.fdb	noRK8P

	.endif



; ===================================================================================



; Binary program loader. Field and origin will be set up to field 0
; and PC=0200.  Data must start and end with leader / trailer code.
; An optional ASCII text before the leader is accepted, and printed.
; The trailer can be followed by nulls.  The first character after the
; trailer/nulls is passed back to the caller. Rub-out character is not
; processed.  An error is flagged by <mumble>.
; Program PC and Field will have the exit values from the loader,
; which serves as a default starting address.
;
;   Description of binary loader format
; ---------------------------------------
; Two characters representing the address (origin) into which the first
; command on the next portion of the tape will be placed. Successive commands
; are placed in memory at addresses: origin+1, origin+2, etc.
; The initial character of the origin has no punch in channel 8, while
; channel 7 is punched. The second character designating the origin has
; no punches in either channel 8 or 7.
; A concluding 2-character group represents the checksum with no punches present
; in channel 8 or channel 7.
; When channel 8 and channel 7 are punched, the character is a "field" setting:
; 1 1 x x x 0 0 0  indicate the memory field into which the following data is
; to be loaded. If, for example "x x x" is "1 0 1", all data following the field
; designator should be loaded into memory field five.
; Contents have no punch in channel 7 or 8.
; The papertape leader and trailer contains any code with channel 8 punched.
; Octal 200 is preferred (several times repeated).
;
;    8 7  6 5 4 s 3 2 1
;   --------------------
;    0 0  x x x . x x x    --> data byte
;    0 1  x x x . x x x    --> origin code, plus next following byte
;    1 0  x x x . x x x    --> leader/trailer byte, [6-1] prefered 0
;    1 1  x x x . 0 0 0    --> field code (single byte)
;
;
;       tape channel        channel 8 and 7    (s = sprocket hole)
;    8 7  6 5 4 s 3 2 1      indicate :
;   ---------------------------------------------------------------------
;    1 0  0 0 0 . 0 0 0     leader code
;
;    0 1  0 0 0 . 0 1 0     origin             octal origin 0200
;    0 0  0 0 0 . 0 0 0
;    0 0  1 1 1 . 0 1 0     contents of 0200   command 7200 or CLA
;    0 0  0 0 0 . 0 0 0
;    0 0  0 1 1 . 0 1 0     contents of 0201   command 3276 or DCA Z 076
;    0 0  1 1 1 . 1 1 0
;    0 0  1 1 1 . 1 0 0     contents of 0202   command 7402 or HLT
;    0 0  0 0 0 . 0 1 0
;    0 0  0 0 0 . 1 0 0     checksum           checksum = 0422
;    0 0  0 1 0 . 0 1 0
;    1 0  0 0 0 . 0 0 0     trailer code
;
; Checksum = 102 + 000 + 072 + 000 + 032 + 076 + 074 + 002.
;
;   Entry to loader.
;   Set up starting defaults and bypass leader code

PDPbin:	jsr	RDRini		; set up interrupt-driven data reception
	ldx	GetFld		; get start of RAM, default field is 0
	clra
	clrb
	tfr	d,y		; clear byte read-in counter
	std	*binsum		; clear checksum
	std	*binsv0		; and temp save locations
	std	*binsv1
	ldd	#0o200		; set a default origin
	std	*PC

;   some paper tapes start with ASCII information, such as
;   - producer and product info
;   - load/start address

_asc0:	jsr	RDRdat		; get byte
	cmpa	#XON		; ^Q ?
	lbeq	_tlr2		; y - no .BIN load needed
	tsta			; msb set ?

	.if	BINldr_txt

	bmi	_ldr0		; y - print "blocks" line first
	cmpa	#Space		; n - check for ASCII printable
	bhs	_asc1		;     ok: print it
	lda	#'.		;     non-printable: output "."
_asc1:	jsr	outch
	bra	_asc0

_ldr0:	pshs	x
	ldx	#RDRms2
	jsr	pdatao		; ASCII header done, print "block" line
	puls	x
	bra	_ldr2
_ldr1:	jsr	Shw_LT		; show leader "." (max 3)
	jsr	RDRdat

	.else

	bpl	_asc0		; "eat" all ASCCI text
	bra	_ldr2		; -> process leader
_ldr1:	jsr	RDRdat		; get 1 byte

	.endif

;   leader section

_ldr2:	tfr	a,b
	andb	#0xC0		; check upper 2 bits, ignore low bits
	cmpb	#0o200		; leader code 02xx / 0200 (pref'd) ?
	beq	_ldr1		; y - get next byte

;   binary data loader - process first (received) byte

_bin0:	sta	*_bin3		; will need a copy
	ldd	*binsv1		; move saved data
	std	*binsv0
	ldb	*_bin3
	std	*binsv1		; save present data
	addd	*binsum		; update checksum
	std	*binsum
	lda	*_bin3
	bita	#0o300		; data, origin, or field ?
	beq	_bin1		; -> data: get 2nd byte of data

    ; field or origin setting

	anda	#0b11000000	; -> check: origin or field ?
	cmpa	#0o300
	beq	_fset
	.if	BINldr_txt
	jsr	Shw_O		; show Origin data received
	.endif
	jsr	binwrd		; origin - read 2nd byte of origin
	std	*PC
	bra	binchk		; ok, continue loading

_fset:	lda	*_bin3		; field format
	tfr	a,b
	andb	#7		; bits 2-1-0 must be zero (spec)
	lbne	lderr1		; -> wrong, not zero!
	lsra			; ok, move field bits [6-5-4] -> [3-2-1]
	lsra
	lsra
	anda	#7		; maximum field number
	.if	BINldr_txt
	jsr	Shw_F		; show Field data received
	.endif
	ldu	#GetFld		; make field pointer
	lsla
	ldx	a,u
	lbeq	lderrf		; -> field does not exist!
	bra	binchk		; ok, continue loading

    ; get 2nd byte of data

_bin1:	bsr	binwrd		; read 2nd byte of data
	.if	BINldr_txt
	jsr	Shw_D		; show data "d" (max 3)
	.endif
	pshs	d
	ldd	*PC
	lslb			; form word offset
	rola
	leau	d,x
	stu	*binptr		; save previous data
	stu	*rc_bus		; show address in BUS (RealConsole)
	ldd	,u
	std	*bindat
	puls	d		; load new data
	std	,u
	std	*MD		; show data in MD (RealConsole)
	ldd	*PC		; update "PC"
	addd	#1
	anda	#0b00001111	; prevent PC overflow
	std	*PC

    ; read 1st byte

binchk:	jsr	RDRdat
	tfr	a,b
	andb	#0xC0		; check upper 2 bits, ignore low bits
	cmpb	#0o200		; trailer code 02xx / 0200 (pref'd) ?
	bne	_bin0		; n -> process data

;   trailer section

	ldd	*binsum		; y -> undo last add
	subd	*binsv0
	subd	*binsv1
	subd	[binptr]	; update checksum
	anda	#0b00001111
	std	*binsum		; must be 0, else
	bne	lderr2		; -> checksum error!
	ldd	*bindat		; restore last written data
	std	[binptr]
	ldd	*PC		; restore PC
	subd	#1
	anda	#Wmask
	std	*PC

    ; process trailer, possibly followed by next part (multi-part tape)
    ; assumption here: trailer becomes leader of next part, next part
    ; starts with a Field or an Origin specification.
    ; Trailer end is detected by (user entry of) the Xon character.
    ; When, after the trailer data is detected, a not-trailer, not-
    ; origin or not-field data byte is read, that byte and all following
    ; bytes are skipped.
    ; Also the byte-read-in counter [Y] is not incremented. This enables
    ; checking after reading where in the file the rubbish data started.

_tlr0:
	.if	BINldr_txt
	jsr	Shw_LT		; show trailer "." (max 3)
	.endif
	jsr	RDRdat
	tfr	a,b
	andb	#0xC0		; check upper 2 bits, ignore low bits
	cmpb	#0o200		; trailer code 02xx / 0200 (pref'd) ?
	beq	_tlr0		; y -
	bitb	#0xC0		; n - Origin or Field specification ?
	lbne	_bin0		;     y - next part ->
	cmpa	#XON		;     ^Q Xon entry ?
	beq	_tlr2		;     y - done reading
_tlr1:	jsr	RDRdat		;     n - read rubbish, no counter update
	cmpa	#XON		;         "eat all" - until Xon (entry)
	bne	_tlr1
_tlr2:	orcc	#0x10		; disable ACIA-IRQ on 6809 CPU
	rts			; Xon is the return code


;   * binwrd - read 2nd byte (data or origin)
;
;     call   : <_bin3> = high 6 bits of 12-bit word
;     return :   [D]   = 12-bit word

binwrd:	jsr	RDRdat		; get low 6 bits
	sta	*_bin4
	bita	#0b11000000	; check for error
	bne	binerr		; error - upper 2 bits <> 0
	ldd	*binsv1
	std	*binsv0
	ldb	*_bin4
	std	*binsv1
	addd	*binsum
	std	*binsum
	lda	*_bin3
	anda	#0b00111111	; drop high bits
	clrb
	lsra
	rorb
	lsra
	rorb
	orb	*_bin4
	rts


    ; errors detected in the binary loader

lderrf:	ldx	#Errmem
	bra	ldrerr		; field does not exist (no memory)
lderr1:	ldx	#Errfld
	bra	ldrerr		; field specification error
lderr2:	ldx	#Errcks
	bra	ldrerr		; checksum failed
binerr:	leas	2,s		; clean-up stack: remove "binwrd" call return adrs
	ldx	#Lderr
;
ldrerr:	pshs	x
	ldx	#ldrcnt		; new line, byte count text
	jsr	pdatao
	sty	*Temp2
	ldx	#Temp2		; print byte counter of received bytes in the loader
	jsr	OUT4HS
	jsr	docrlf
	puls	x
	jsr	pdatao		; print error message
	jsr	docrlf		; print extra blank line
	jmp	docrlf		; and RTS (to "PDPld")




;  * BIN loader - receive data routines (interrupt driven)

RDRini:	ldx	#RDRirq
	stx	.irq		; set IRQ vector
	ldx	#RDRblo
	stx	*RDRin		; set up buffer I/O pointers
	stx	*RDRout

	.if	BINldr_txt
	clr	*RDRddd		; clear block indicator counters
	clr	*RDRltr
	clr	*RDRcnt		; # printer chars on one line
	ldx	#RDRms1
	jsr	pdatao
	.endif

	jmp	RS_ion		; enable RS232 input interrupt , RTS


RDRirq:	pshs	a,x		; save environment [A],[X]
	lda	ACIADA		; get data byte, clear IRQ flag
	ldx	*RDRin
	sta	,x+		; store data in buffer
	cmpx	#RDRbhi		; crossed end of buffer ?
	bne	1$
	ldx	#RDRblo		; y - cycle buffer ptr
1$:	stx	*RDRin
	cmpx	*RDRout		; simple test: buffer overrun *_next_* byte?
	bne	2$
	ldx	#RDRmsx		; *** ugly: text printing inside IRQ code
	jsr	pdatao		; ***       but this should never occur!!
2$:	puls	a,x		; restore environment
	rti


RDRdat:	pshs	x
	ldx	*RDRout		; get retrieve ptr
1$:	cmpx	*RDRin
	beq	1$		; wait for data
	lda	,x+		; get it
	cmpx	#RDRbhi		; crossed end of buffer ?
	bne	2$
	ldx	#RDRblo		; y - cycle buffer ptr
2$:	stx	*RDRout
	leay	1,y		; bump byte counter
	sty	*MQ		; show byte count in MQ (RealConsole)
	puls	x,pc


	.if	BINldr_txt

Shw_O:	pshs	a
	lda	#'O
	bsr	RDRdmp		; show Origin block
	clr	*RDRddd		; clear block counters
	clr	*RDRltr
	puls	a,pc

Shw_F:	pshs	a
	lda	#'F
	bsr	RDRdmp		; show Field block
	clr	*RDRddd		; clear block counters
	clr	*RDRltr
	puls	a,pc

Shw_LT:	pshs	a
	clr	*RDRddd
	lda	*RDRltr		; check # of printed "."
	cmpa	#3
	beq	1$		; 3 '...' printed in this block: skip
	inca
	sta	*RDRltr
	lda	#'.
	bsr	RDRdmp
1$:	puls	a,pc

Shw_D:	pshs	a
	clr	*RDRltr
	lda	*RDRddd		; check # of printed "d"
	cmpa	#3
	beq	1$		; 3 'ddd' printed in this block: skip
	inca
	sta	*RDRddd
	lda	#'d
	bsr	RDRdmp
1$:	puls	a,pc


;  * RDRdmp - dump data indicator
;
;  call   : [A] = print code
;		'.' - for leader or trailer word
;		'd' - for data word
;		'F' - for Field word
;		'O' - for Origin word
;  return : [A] destroyed

RDRdmp:	jsr	outch		; print char
	lda	*RDRcnt
	inca
	cmpa	#60		; end of line ?
	bne	1$		; n -
	pshs	x		; y - new line
	ldx	RDRms3
	jsr	pdatao
	puls	x
	clra			;     clear counter
1$:	sta	*RDRcnt
	rts


RDRms1:	.fcb	CR,LF
	.asciz	"- header : "
RDRms2:	.fcb	CR,LF
	.asciz	"- blocks : "
RDRms3:	.fcb	CR,LF
	.asciz	"           "

	.endif

RDRmsx:	.fcb	CR,LF
	.fcc	"! input buffer overrun"
	.fcb	CR,LF,0




    ; Pntreg - print the PDP-8 registers

Pntreg:	pshs	x
	lda	*Link		; link bit in wrong place
	lsra			; move it
	lsra
	lsra
	lsra
	sta	*_stp2
	lda	*WorkUI		; U in MSB
	lsla			; U to carry and 0 to LSB
	rola			; U to LSB and next (Df) bits = 0
	anda	#0b00000001
	sta	*_stp3
	ldu	#_stp4		; flag set routine
	ldd	*emufnc		; check for ION
	cmpd	#Chkint
	bsr	ClrFLG
	lda	*IOflgs		; check for NoI
	anda	#Jdelay
	bsr	SetFLG
	lda	*IOflgs		; check for Ibuss
	anda	#Jdelay^0xFF
	ora	*IOflgs+1
	bsr	SetFLG
	ldx	#Stpmsg
_stp0:	bsr	pdatao
	ldu	,x++
	lda	,x+
	bsr	PrtOct
	lda	,x		; end of formatted string?
	beq	_stp1		; y -
	cmpa	#CR		; n - CR/LF flag ?
	bne	_stp0
	bsr	otchr		; y - print new line
	lda	#LF
	bsr	otchr
	leax	1,x
	bra	_stp0
_stp1:	puls	x,pc


    ; Set location pointed to by U to 1 if entered with ccZ == 1
    ; location set to 0 otherwise. Exit with U incremented.

ClrFLG:	beq	setf
	bra	clrf

    ; Set location pointed to by U to 1 if entered with ccZ != 0
    ; location set to 0 otherwise. Exit with U incremented.

SetFLG:	beq	clrf
setf:	lda	#1
setfx:	sta	,u+
	rts
clrf:	clra
	bra	setfx



;   Error message print routine.
;   When an error is detected, the code does a JSR to this routine, and
;   the 2 bytes following the JSR is the address of the error message.

Errpnt:
; ---	orcc	#0x50		; disable 6809 FIRQ and IRQ interrupts
	clr	*trcdmp		; disable data dump re-routing (if TRACE BEFORE
				;  is active, output from Pntreg is stored in a
				;  buffer instead of displayed on screen)
	bsr	nwlin
	stx	*_stp2		; preserve [X] !
	ldx	[0,s]		; get message pointer
	bsr	pdatao
	ldx	*_stp2
	bsr	nwlin
	bsr	Pntreg
	.if	RealCons
	leas	2,s		; clean up stack
	clr	*conact		; try to maintain RealConsole control
	bsr	nwlin
	jmp	Emulat
	.else
	bra	.		; PROGRAM HALTED :: loop forever until RESET
	.endif



;  PrtOct - print an octal number
;
;    call   : [A] = number of digits to (1 to 4)
;	      [U] = pointer to data location (2 bytes)
;    return : requested digits printed
;	      [A] and [B] destroyed

PrtOct:	sta	*_octp2		; save digit count
	sta	*_octp3		; also need character count
	ldd	,u
octprd:	std	*_octp4
_octp0:	lda	*_octp5		; get data
	anda	#7		; make into a character
	adda	#'0
	pshs	a		; save on stack
	ldd	*_octp4		; move next digit into place
	lsra
	rorb
	lsra
	rorb
	lsra
	rorb
	std	*_octp4
	dec	*_octp2
	bne	_octp0
_octp1:	puls	a
	bsr	otchr
	dec	*_octp3
	bne	_octp1
	rts


otchr:	tst	*trcdmp		; data dump re-routing active ?
	lbne	stordi		; y - store in buffer (TRACE BEFORE is active)
	jmp	outch		; n - print to screen
nwlin:	jmp	docrlf


;   pdatao - print text string
;
;   call   : [X] = pointer to start of string, ended *NULL*
;   return : [X] - point to byte following the NULL character
;	     [A] - preserved

pdatao:	pshs	a
	lda	,x+	; get char
	beq	pdatax	; done -
pdata:	bsr	otchr
	lda	,x+
	bne	pdata
pdatax:	puls	a,pc



;   Monitor program load.
;   Wait for a binary program to be loaded.
;   This means the terminal data link has been established.
;   A startup message is printed and the routine exits.

PDPld:	ldx	#ldrmss		; print loader start message
	bsr	pdatao
	jsr	PDPbin		; load the binary source
	cmpy	#1		; immediate ^Q (no load needed) ?
	beq	PDPskp		; y - "skipped"
	ldx	#ldrmse		; n - print loader finished message
	bsr	pdatao		;     and ask for another file to load
;
PDPwht:	jsr	inch		; await answer
	anda	#0x5F		; make uppercase
	cmpa	#'N
	beq	PDPldx		; N -> done loading
	cmpa	#'Y
	bne	PDPwht		; not Y -> say what? -- try again
	bsr	otchr		; Y -> print answer
	bra	PDPld		;      go for next file
PDPskp:	ldx	#ldrmsn
	bra	pdatao




;   IDinfo - print start-up identification message
;
;   call : <verbos> and <dsknum> flags set

IDinfo:	ldx	#Sysmsg		; issue program identification message
	bsr	pdatao
	lda	*verbos		; verbose output ?
	beq	IDx		; n -
	ldx	#config
	bsr	pdatao
	lda	*dsknum
	adda	#'0
PDPldx:	bsr	otchr		; print disk ID #
	bra	nwlin		; new line and RTS



;    vbprnt - verbose print string
;
;   call   : [X] = pointer to start of string, ended *NULL*
;	     <verbos> = "verbose" flag
;   return : if <verbos> == 0 --> no text printed
;				  [X] not changed
;	     if <verbos> != 0 --> text printed
;				  [X] point to byte following the NULL character
;	     [A] - preserved

vbprnt:	tst	*verbos		; verbose ?
	bne	pdatao		; y -
IDx:	rts			; n -







;  initRC - initialize RealCons - Core and I/O - hardware
;
;  * must be called -before- the initialization of the pdp8/e and IDE disk 'hardware'
;    initRC also sets a dummy IRQ vector.
;
;    call   : -
;    return : 6821 PIA I/O initialized for I/O Board connection
;	      switch/key entry management storage initialized
;	      FIRQ vector set to process switch/key entry / output LED's
;	      FIRQ enabled on 6809, IRQ, SWI, NMI set to dummy routine
;	      input port switch/toggle mask defined
;	      output ports all cleared

	.if	RealCons
initRC: orcc	#0x40		; make sure 6809-FIRQ is disabled
	ldx	#ircmsg
	bsr	vbprnt		; print progress (if verbose mode)
	lda	#ACIANI
	sta	ACIACS
	sta	*rtssts
	lda	ACIADA		; clear ACIA
	ldx	#PDPFRQ
	stx	.firq		; load FastIRQ destination vector
	lda	#50
	sta	*timcnt		; set flash rate approx 1 Hz.
	ldx	#dumirq
	stx	.nmi		; load NMI destination vector
; ---	stx	.swi		; load SWI destination vector
	stx	.irq		; load IRQ destination vector (ACIA)

	clra			; configure PIA port A & B
	sta	PIACA		; access data direction reg A
	sta	PIADA		; default: port A == input
	ldb	#PACTLJ		; control mask set for direction: input
	stb	PIACA		; access data reg A
	stb	*piadir		; set flag: port A set to "INPUT"

	sta	PIACB		; access data direction reg B
	coma
	sta	PIADB		; always: port B == output
	ldb	#PBCTLO		; control mask set for direction: output
	stb	PIACB		; access data reg B

	ldx	#RCRAM		; start of input port shadow
	lda	#LSTIRX		; # bytes to clear
init0:	clr	,x+
	deca
	bne	init0
	clr	*conact		; no console action response set
	clr	*dlyexm
	clr	*dlydep
	clr	*dlyclr
	clr	*dlyal
	clr	*dlyeal
	clr	*booten		; bootstrap is possible
	ldx	#okmsg		; print progress (if verbose mode)
	bsr	vbprnt
	.if	ide2pdp
	clr	*accide		; access IDE flag: handshake between IDE and FIRQ
	jsr	IDEini		; initialize the IDE disk
	.endif
	andcc	#0b10111111	; enable 6809-FIRQ
	rts


ircmsg: .fcb	CR,LF
	.asciz	"[i] initialising Real Console hardware     ..."




;  ** formatted message string used by Pntreg - print all pdp8/e info
;
;  1st entry - ASCII text string, NULL ended
;  2nd entry - 2-byte start address of variable that holds the data value
;  3rd entry - 1 byte indicating the number of octal digits to be printed
;
;  after a group as described above follows either
;  - the next group with the same lay-out
;  - one byte character CR indicating a new line must be printed
;  - one byte NULL indicating end of the formatted message to print

Stpmsg:	.asciz	" Df="
	.fdb	WorkD-1
	.fcb	1
	.asciz	" AC="
	.fdb	AC
	.fcb	4
	.asciz	" L="
	.fdb	_stp2-1
	.fcb	1
	.asciz	" MQ="
	.fdb	MQ
	.fcb	4
	.asciz	" U="
	.fdb	_stp3-1
	.fcb	1
	.asciz	" ION="
	.fdb	_stp4-1
	.fcb	1
	.asciz	" NoI="
	.fdb	_stp5-1
	.fcb	1

	.asciz	" Ireq="
;	.fdb	_stp6-1
;	.fcb	1
	.fdb	IOflgs
	.fcb	5		; all but JMP/JMS delay

	.fcb	CR		; show with next state

	.asciz	" If="
	.fdb	WorkUI-1
	.fcb	1
	.asciz	" In="
	.fdb	MD
	.fcb	4
	.asciz	" PC="
	.fdb	PC
	.fcb	4

	.fcb	0		; end of message list





  ;  ***********==========    Real Console  FIRQ  routine    ==========***********

  ; The interrupt is based on a timer event that triggers the CA1 pin of the PIA.
  ; The following processes are executed inside the interrupt routine:
  ; 1. the IRQ flag of the PIA is cleared
  ; 2. <timer> is decremented. On reaching $00 the following actions are done:
  ;    - <timer> is reloaded with <timcnt>
  ;    - <rtssts> is tested/changed and the monitor LED (at the RTS pin) is toggled
  ; 3. <SWTDBCn> is decremented. On reaching $00 the following actions are done:
  ;    - <SWTDBCn> is reloaded with a fixed value.
  ;    - the physical input port [n] is read.
  ;    - the physical data is compared with the previous read data.
  ;    - SWITCH bits that are equal are transfered to the shadow input port.
  ; 4. The console switches and toggles are processed for the pdp8/e functionality.
  ; 5. Some storage locations from the simulator are read & displayed on the LEDs.

PDPFRQ:	pshs	d,x
	lda	PIADA		; dummy read: clear IRQ flag
	lda	*timer		; source == PIA [port A] - CA1 (timer)
	deca			; adjust/check monitor LED interval elapsed ?
	sta	*timer
	bne	pirq2
				; y - perform monitor LED update
pirq1:	lda	*timcnt		; reload interval time
	sta	*timer
	ldb	*rtssts		; toggle monitor LED on/off
	bitb	#RTSBIT		;    ** preserve IRQ-enable bit **
	beq	pirq1a
	andb	#RTSBIT^0xFF	; RTS:=1, LED on
	bra	pirq1b
pirq1a:	orb	#RTSBIT		; RTS:=0, LED off
pirq1b:	stb	ACIACS
	stb	*rtssts

  ;  change data direction of input ports to "READ"
  ; ------------------------------------------------

pirq2:	ldx	#RCRAM		; set pointer to base of array
	clr	*procin		; start at input port #0, I/O Board #0
	lda	*piadir		; set PIA port A direction to "INPUT"
	bita	#CA2BIT		; was direction already "INPUT"?
	bne	pirq2a		; y - no action needed
	anda	#0xFF-DIRBIT	; n - change direction to INPUT
	sta	PIACA		;     1. access data direction reg A
	clra
	sta	PIADA		;     2. set port A == input
	lda	*piadir		;     3. now set buffer-74245 to input
	ora	#CA2BIT+DIRBIT	;	 (prevent short-circuit moment
	sta	PIACA		;         between buffer-74245 and PIA)
	sta	*piadir		;     4. set flag: port A set to "INPUT"

  ;  loop through all defined active input ports and debounce switch contacts
  ; --------------------------------------------------------------------------

pirq2a:	lda	SWTDBC,x	; get SwitchToggleDebounce for this port
	deca			; SwitchToggleDebounce--
	bne	pirq2c
				; if SwitchToggleDebounce == 0
	lda	#BOUNCE
	sta	SWTDBC,x	; reload SwitchToggleDebounce
	ldb	*procin		; get current processing port number
	orb	#0xFF-IOADDR-RDWR ; negate DIS3S control
;	andb	#0xFF-RDWR	  ; direction: READ: enable "read" LS138
				  ; read physical port data
	stb	PIADB		; > setup correct input latch "address"
	andb	#0xFF-DIS3S
	stb	PIADB		; > enable LS138 decoders
	nop			; > hardware settle delay
	lda	PIADA		; > read switch/toggle selected input port
	orb	#DIS3S
	stb	PIADB		; > disable LS138 decoders

	cmpa	PREVIN,x	; if physical_data == previous_data
	bne	pirq2b
	sta	INPORT,x	; set new key_data
	bra	pirq3
				; else
pirq2b:	sta	PREVIN,x	; previous_data = physical_data
	lda	#BOUNCE		; reload SwitchDebounce
				; else
pirq2c:	sta	SWTDBC,x	; store updated SwitchToggleDebounce
pirq3:	ldb	*procin
	incb			; move to next input port
	stb	*procin
	inx
	cmpb	#CONIHI		; all input ports processed ?
	bls	pirq2a		; n - read next input port


  ;  process console switches / toggles
  ; ------------------------------------

  ; * 1. always read Switch Register and store in SR

	lda	*INBASE+RD8e_1	; read SR[0:3]
	ldb	*INBASE+RD8e_0	; read SR[4:11]
	anda	#SW_swh		; isolate SR[0:3]
	std	*SR		; set Switch Register


  ; * 2. check status of HALT and S/STEP keys // set <conact> flag if necessary
  ;
  ;   After power-up [conact] == $00, *any* state of the switches is possible!
  ;   HALT   = off  --> [conact] not changed, process all toggles
  ;			CONT activated  --> [conact] := ACTrun or ACTtrc
  ;   S/STEP = off  --> [conact] not changed, process all toggles
  ;			CONT activated  --> [conact] := ACTrun or ACTtrc
  ;   HALT   = on   --> [conact] := $00, process all toggles
  ;			CONT activated  --> [conact] := ACTstp
  ;   S/STEP = on   --> [conact] := $00, process all toggles
  ;			CONT activated  --> [conact] := ACTcyc

	lda	*INBASE+RD8e_1	; check if HALT or S/STEP switch is active
	anda	#SW_hlt+SW_stp	; note: in active state the bit is 0 !
	cmpa	#SW_hlt+SW_stp
	beq	procsw		; both are NOT active: [conact] not changed
	clr	*conact		; any of the two is active: [conact] := $00
				;                   - clear ACTrun / ACTtrc

  ; * 3. process CONTINUE toggle (action depends on state of HALT and S/STEP)

procsw:	ldb	*INBASE+RD8e_3	; read toggles
; 	bitb	#SW_cnt		; CONT active (bit is 1) ?
;	beq	15$		; n -
	bpl	15$		; n - (performance: CONT at bit 7)

  ;   # CONT active (bit is 1) - check state of HALT and S/STEP switches
  ;	   HALT = on	--> [conact] := ACTstp
  ;	   S/STEP = on	--> [conact] := ACTcyc
  ;	   both  off	--> [conact] := ACTrun or ACTtrc

10$:	lda	#ACTrun		; assume both off, and trace disabled
	ldb	*trcflg		; is trace option disabled ?
	beq	11$		; y - set to "run" state
	lda	#ACTtrc		; n - set to "trace" state
11$:	sta	*conact
	ldb	*INBASE+RD8e_1	; read HALT and S/STEP switch
	lda	#ACTcyc		; [conact] value if S/STEP active (bit is 0)
	bitb	#SW_stp		;   (cycle instructions not yet implemented)
	beq	13$		;   (HALT is don't care if S/STEP is active)
12$:	bitb	#SW_hlt
	bne	15$
	lda	#ACTstp		; HALT active :: step instructions
13$:	sta	*conact
	bra	20$


  ; * 4. process other toggles (ADDR-LOAD, EXTD-ADDR-LOAD, EXAM, DEP, CLEAR)
  ;	 Activation of these is only honored if [conact] != ACTrun

15$:	lda	*conact		; accept toggles ? [ACTrun] - (bit7) is clear
	lbmi	99$		; n -

   ; ## process ADDRESS LOAD toggle

20$:	ldb	*INBASE+RD8e_3
	bitb	#SW_al		; ADDR-LOAD active (bit is 1) ?
	bne	24$		; y -
	lda	*dlyal		; n - ADDR-LOAD at least TFRDLY inactive ?
	lbeq	30$
	deca
	jmp	29$
24$:	lda	*dlyal		; was ADDR-LOAD inactive long enough ?
	bne	28$		; n - reload wait time for ADDR-LOAD toggle
	ldd	*SR		; y - copy SR to Program Counter
	std	*PC
	lda	*INBASE+RD8e_1	;     get HALT and SW switch status
	bita	#SW_hlt		;     HALT active ?
	bne	27$		;     n - no need to check SW
	bita	#SW_sw		;     y - SW active ?
	beq	27$		;	  n -

;   * load bootstrap code
;     the following conditions are true:
;	1. LOAD ADDR momentary switch is pushed
;	2. HALT switch is active
;	3. if <booten> flag is 1 -> no actions
;	   if <booten> flag is 0 -> bootstrap code is loaded into core
;     NOTE: the current implementation automatically loads the bootstrap code
;	    in Field number 0. If the boostrap is not loaded into Field #0,
;	    you must also use the "EXT ADDR LOAD" *before* CONT (and LOAD ADDR)
;	    is pushed!

	lda	*booten		; bootstrap code load possible ?
	bne	249$		; n -
	inca
	sta	*booten		; y - disable bootstrap code load from now on
	pshs	y
	ldy	*boottb		; valid bootstrap code ?
	bne	241$
	puls	y		; n - issue message, that's all
	ldx	#noboot
	jsr	vbprnt		;     only if "verbose mode" is choosen
	bra	249$
;
241$:	ldd	*PC		; first calculate the real memory address
	aslb
	rola
	addd	GetFld		; always in Field 0
	tfr	d,x
;
	ldb	#6		; copy the first 3 instructions == 6 bytes
242$:	lda	,y+		;   CLA
	sta	,x+		;   DMAR
	decb			;   DFSC
	bne	242$
	ldb	*PC+1		; get base of current page for relocation
	andb	#0b01111111
	clra
	pshs	d
	addd	,y++		;   JMP   .-1
	std	,x++
	puls	d
	addd	,y++		;   JMP I .+1
	std	,x++
	ldd	,y++		;   start address of bootstrap
	std	,x++
	ldd	,y++
	std	wc3cdb		; load Word Count
	ldd	,y++
	std	ca3cdb		; load Current Address
	tfr	y,x		; ptr to OS identification text
	puls	y
;
	lda	*verbos		; verbose active ?
	beq	249$		; n -
	jsr	docrlf		; y - new line
	jsr	pdatao		;     print selected OS (DMS or OS/8 for now)
	ldx	#bootok
	jsr	pdatao		;     print " bootstrap loaded at "
	lda	#4
	sta	_octp2		;     print 4 digit bootstrap start PC
	sta	_octp3
	ldd	*PC
	jsr	octprd
	jsr	docrlf
249$:
27$:	ldb	*INBASE+RD8e_3
28$:	lda	#TFRDLY		; reload wait time for ADDR-LOAD toggle inactive
29$:	sta	*dlyal

   ; ## process EXTENDED ADDRESS LOAD toggle

30$:	bitb	#SW_eal		; EXTD-ADDR-LOAD active (bit is 1) ?
	bne	34$		; y -
	lda	*dlyeal		; n - EXTD-ADDR-LOAD at least TFRDLY inactive ?
	beq	40$
	deca
	bra	39$
34$:	lda	*dlyeal		; was EXTD-ADDR-LOAD inactive long enough ?
	bne	38$		; n - reload wait time for ADDR-LOAD toggle
	lda	*SR		; y - read SR4-SR11
	anda	#Dbits		;     --> D-field (SR9-10-11)
	sta	*WorkD
	sta	*SaveD
	lda	*SR		;     read SR4-SR11
	asra
	asra
	asra
	anda	#Fbits		;     --> I-field (SR6-7-8)
	sta	*FIRQsv		;	  save I-field
	tst	*WorkUI
	bpl	31$
	ora	#Ubit		;	  preserve U bit
31$:	sta	*WorkUI
	lda	*FIRQsv		;	  restore I-field
	tst	*SaveUI
	bpl	32$
	ora	#Ubit
32$:	sta	*SaveUI
	lda	*FIRQsv		;	  restore I-field
	tst	*BuffUI
	bpl	33$
	ora	#Ubit
33$:	sta	*BuffUI
38$:	lda	#TFRDLY		; reload wait time for EXTD-ADDR-LOAD toggle inactive
39$:	sta	*dlyeal

   ; ## process EXAMINE toggle

40$:	bitb	#SW_exm		; EXAM active (bit is 1) ?
	bne	44$		; y -
	lda	*dlyexm		; n - EXAM at least TFRDLY inactive ?
	beq	50$
	deca
	bra	49$

   ; 1)	Get current 'active' address, NXM is allowed.
   ;    Since the emulator code loops at "Doinst" while HALT or S/STEP is active,
   ;    MD (Memory Data) holds automatically the correct information.
   ; 2)	Increment 'active' address.
   ; ** NOT (yet) CORRECT: first EXAMINE request increments PC ...

44$:	lda	*dlyexm		; was EXAM inactive long enough ?
	bne	48$		; n - reload wait time for EXAM toggle

   ; actually reading the data from [PC] and writing that data into [MD]
   ; is not necessary, as the main loop (Doinst) in the simulator does this!

;	lda	*WorkUI		; y - accept EXAM toggle - get I field
;	lsla
;	ldx	#GetFld
;	ldx	a,x		; start address in memory of this Field
;	bne	46$
;	ldd	#0		; NXM - Non eXistent Memory
;	std	*MD		; --> data read = 0000
;	bra	47$
; 46$:	ldd	*PC
;	lslb			; *2 :: memory -> core address in Field
;	rola
;	ldd	d,x		; get data from physical core address
;	std	*MD		; show in MD on console
47$:	ldd	*PC
	addd	#1		; auto-increment Program Counter
	std	*PC
	ldb	*INBASE+RD8e_3
48$:	lda	#TFRDLY		; reload wait time for EXAM toggle inactive
49$:	sta	*dlyexm

   ; ## process DEPOSIT toggle

50$:	bitb	#SW_dep		; DEP active (bit is 1) ?
	bne	54$		; y -
	lda	*dlydep		; n - DEP at least TFRDLY inactive ?
	beq	60$
	deca
	bra	59$

   ; 1)	Get current 'active' address, NXM is allowed.
   ; 2)	Copy data in [SR] to the 'active' address.
   ;    Current implementation: if address is NXM, deposit action not done.
   ;	??? The contents of [SR] is displayed in the pdp8 register MD (???)
   ; 3)	Increment 'active' address.

54$:	lda	*dlydep		; was DEP inactive long enough ?
	bne	58$		; n - reload wait time for DEP toggle
	lda	*WorkUI		; get I field
	lsla
	ldx	#GetFld
	ldx	a,x		; start address in memory of this Field
	beq	56$		; NXM - Non eXistent Memory: no deposit!
	ldd	*PC
	lslb			; *2 :: memory -> core address in Field
	rola
	leax	d,x		; physical core address
	ldd	*SR
	std	,x		; store data from SR
56$:	ldd	*PC
	addd	#1		; auto-increment Program Counter
	std	*PC
	ldb	*INBASE+RD8e_3
58$:	lda	#TFRDLY		; reload wait time for DEP toggle inactive
59$:	sta	*dlydep

   ; ## process CLEAR toggle

60$:	bitb	#SW_clr		; check CLEAR active (bit is 1)
	bne	64$		; y -
	lda	*dlyclr		; n - CLEAR at least TFRDLY inactive ?
	beq	99$
	deca
	bra	69$
64$:	lda	*dlyclr		; was CLEAR inactive long enough ?
	bne	68$		; n - reload wait time for CLEAR toggle
	jsr	Clear		; activated: reset the pdp8 environment
68$:	lda	#TFRDLY		; reload wait time for CLEAR toggle inactive
69$:	sta	*dlyclr


  ;  output current state to LED's
  ; -------------------------------

99$:	lda	*piadir		; set data output port == OUTPUT
; --	bita	#CA2BIT		; was direction already "OUTPUT" ?
; --	beq	999$		; y - no action needed (we know, action is IS needed!)
	anda	#0xFF-CA2BIT	; n - change directio to OUTPUT
	sta	PIACA		;     1. only the 245 buffer (prevent short-circuit)
	sta	*piadir		;	 set flag: port A set to "OUTPUT"
	anda	#0xFF-DIRBIT	;     2. access data direction reg A
	sta	PIACA
	lda	#0xFF
	sta	PIADA		;     3. set port A == output
	lda	*piadir
	sta	PIACA		;     4. access data output reg A
999$:
	.if	ide2pdp
				; process the handshake between IDE and FIRQ
	ldb	*rc_ste+1	; get STATE[7:0] bits
	lda	*accide		; get IDE/FIRQ shared variable
	beq	dskacc		; == $00 ==> access complete ?
	bpl	dskacc		; == $80 --> IDE access requested ?
	lda	#1		; y - change to $01 --> signal IDE "request accepted"
	sta	*accide
dskacc:	tsta			; check if LEDs "BRK"/"BRK PROG" must be ON or OFF
	beq	10$
	orb	#LD_brk+LD_bkp	    ; turn both ON
	bra	20$
10$:	andb	#255-LD_brk-LD_bkp  ; turn both OFF
20$:	stb	*rc_ste+1
	.endif

ledout:	lda	*WorkUI		; output the current (pdp8/e) address
	lsla
	lsla
	lsla
	lsla
	lsla
	ora	*PC		; EMA[210] (?) , A[11-8]
	ldb	#WR8e_1		; output upper 7 bits
	coma			; invert for negative logic
	sta	PIADA		; load data pattern
	orb	#0xFF-IOADDR	; negate DIS3S control bit
;	orb	#RDWR		; direction: WRITE: enable "write" LS138
	stb	PIADB		; setup correct output latch "address"
	andb	#0xFF-DIS3S
	stb	PIADB		; assert DIS3S: enable LS138 decoders
	nop			; hardware settle delay
	orb	#DIS3S
	stb	PIADB		; negate DIS3S: disable LS138 decoders

	lda	*PC+1		; A[7-0]
	ldb	#WR8e_0		; output lower 8 bits
	coma			; invert for negative logic
	sta	PIADA		; load data pattern
	orb	#0xFF-IOADDR	; negate DIS3S control bit
;	orb	#RDWR		; direction: WRITE: enable "write" LS138
	stb	PIADB		; setup correct output latch "address"
	andb	#0xFF-DIS3S
	stb	PIADB		; assert DIS3S: enable LS138 decoders
	nop			; hardware settle delay
	orb	#DIS3S
	stb	PIADB		; negate DIS3S: disable LS138 decoders

    ; check the Indicator Selector Switch (ISS)

	lda	*INBASE+RD8e_2	; read ISS bits - get correct data
	ldx	#rc_ste
	bita	#ISSste
	beq	_shw1		; -> STATE
	ldx	#AC
	bita	#ISSac
	beq	_shw1		; -> AC
	ldx	#MD
	bita	#ISSmd
	beq	_shw1		; -> MD
	ldx	#MQ
	bita	#ISSmq
	beq	_shw1		; -> MQ
	ldx	#rc_bus
	bita	#ISSbus
	beq	_shw1		; -> BUS
	jsr	ISSSTS		; must be STATUS (special handling)
_shw1:	ldd	,x
	stb	*FIRQsv		; save lower 8-bit data
	ldb	*conact		; status == ACTrun or ACTtrc ?
	bitb	#ACTrun+ACTtrc
	beq	no_run
	adda	#RUNled		; add RUN status LED = ON

no_run:	ldb	#WR8e_3		; output upper 4 bits
	coma			; invert for negative logic
	sta	PIADA		; load data pattern
	orb	#0xFF-IOADDR	; negate DIS3S control bit
;	orb	#RDWR		; direction: WRITE: enable "write" LS138
	stb	PIADB		; setup correct output latch "address"
	andb	#0xFF-DIS3S
	stb	PIADB		; assert DIS3S: enable LS138 decoders
	nop			; hardware settle delay
	orb	#DIS3S
	stb	PIADB		; negate DIS3S: disable LS138 decoders

	ldb	#WR8e_2		; output lower 8 bits
	lda	*FIRQsv
	coma			; invert for negative logic
	sta	PIADA		; load data pattern
	orb	#0xFF-IOADDR	; negate DIS3S control bit
;	orb	#RDWR		; direction: WRITE: enable "write" LS138
	stb	PIADB		; setup correct output latch "address"
	andb	#0xFF-DIS3S
	stb	PIADB		; assert DIS3S: enable LS138 decoders
	nop			; hardware settle delay
	orb	#DIS3S
	stb	PIADB		; negate DIS3S: disable LS138 decoders

	puls	d,x
dumirq:	rti



;------------------------------------------------------------------------------------
;  RF08/DF32 bootstrap code for DMS - according to DIGITAL doc
;
;		   *0200
;     0000  6603   DMAR		/ read first block
;     0001  6622   DFSC		/ wait for I/O completion
;     0002  5201   JMP    .-1
;     0003  5604   JMP  I .+1
;     0004  7600   7600		/ execute loaded bootcode
;
;     7750	   *7750	/ 3-cycle break addresses
;     7750  7576   7576		/ load Word Count
;     7751  7576   7576		/ load Current Address
;
;     WordCount and Current Address must be loaded with 0o7576.
;     To load the bootstrap from the start address in the SR (PC), the two JMP
;     instructions must be relocated. The bootstrap is always loaded in Field 0.
;     From the start PC [SR] all 5 words must fit inside the 128-word selected page.
;
;
;  RF08/DF32 bootstrap code for OS/8 - according to DIGITAL doc
;
;     The OS/8 bootstrap code for DF/RF (OS/8 System Generation Notes, table 7) is:
;
;		   *7750
;     7750  7600   0o7600	/ CLA			   also Word Count !!
;     7751  6603   DMAR		/ read first block	   also Current Address !!
;     7752  6622   DFSC		/ wait for I/O completion  also boot start address
;     7753  5352   JMP    .-1
;     7754  5752   JMP  I .-2
;
;     WordCount and Current Address are also the first two instructions!
;     The bootstrap must be startet from address 0o7750. The CLA instruction also
;     serves as Word Count, thus 0o0200 (128 decimal) words are read from disk.
;     The DMAR instruction also serves as Current Address, thus the bootcode is
;     loaded into memory from 0o6603 (locations 0o6603-0o7003).
;     The "JMP .-1" instruction is the wait loop for read data transfer completion.
;     The "JMP  I .-2" instruction jumps back to address 0o7752, where the contents
;     (the DFSC instruction) is used as the destination of the indirect jump address!
;     So the start of OS/8 is at o6622 - o6603 = o17 words offset from the address
;     where the code is loaded.
;
;  ----
;
;  To get *both* bootstraps in a table, the OS/8 clever constructed bootstrap code
;  must be changed into a DMS "style" OS/8 bootcode. Then both bootstraps can be
;  stored in a table which is selected by the command line switch "-b".
;
;     The Word Count register is loaded with 0o7600.
;     The Current Address register is loaded with 0o6603. The bootstrap becomes:
;
;		   *adr		/ any start address set by [SR]
;     0000  7600   CLA		/ AC := 0 --> disk (read) address
;     0001  6603   DMAR		/ read first block
;     0002  6622   DFSC		/ wait for I/O completion
;     0003  5202   JMP    .-1
;     0004  5605   JMP  I .+1
;     0005  7600   6603		/ execute loaded bootcode
;
;     7750	   *7750	/ 3-cycle break addresses
;     7750  7600   7600		/ load Word Count
;     7751  6622   6622		/ load Current Address
;
;  To enable the bootstrap code to be loaded anywhere, the two JMP instructions
;  must be relocated, just as the two instructions in the DMS bootstrap code.
;  The first instruction (CLA) is added to the DMS bootstrap code.
;
;
;		  bootstrap code and data structure
;		=====================================
;
;    ___________________________________
;   |   _____________________________	|
;   |  |			     |	|
;   |  |			     |	|	- 6 words bootstrap code
;   |  |           boot code	     |	|	    (4th and 5th word need relocation)
;   |  |			     |	|
;   |  |-----------------------------|	|
;   |  |	  Word Count	     |	|	- 1 word: Word Count
;   |  |       Current Address	     |	|	- 1 word: Current Address
;   |  |-----------------------------|	|
;   |  |	    OS id	     |	|	- ASCII string, NULL terminated
;   |  |_____________________________|	|
;   |___________________________________|
;
;
;
; ____________________________###  DMS bootstrap  ###________________________________
; ===================================================================================

				;	*0000		/ start address DF/RF boot
btDMSt:	.fdb	0o7600		;	CLA
	.fdb	0o6603		;	DMAR		/ read first block
	.fdb	0o6622		;	DFSC		/ wait for I/O completion
	.fdb	0o5202		;	JMP    .-1
	.fdb	0o5605		;	JMP  I .+1
	.fdb	0o7600		;	7600		/ execute loaded bootcode
				;	*7750		/ 3-cycle break addresses
	.fdb	0o7576		;	7576		/ load Word Count
	.fdb	0o7576		;	7576		/ load Current Address
	.asciz	"DMS"		;			/ OS identification


;
; ____________________________###  OS/8 bootstrap  ###_______________________________
; ===================================================================================

				;	*0000		/ start address DF/RF boot
btOS8t:	.fdb	0o7600		;	CLA
	.fdb	0o6603		;	DMAR		/ read first block
	.fdb	0o6622		;	DFSC		/ wait for I/O completion
	.fdb	0o5202		;	JMP    .-1
	.fdb	0o5605		;	JMP  I .+1
	.fdb	0o6622		;	6622		/ execute loaded bootcode
				;	*7750		/ 3-cycle break addresses
	.fdb	0o7600		;	7600		/ load Word Count
	.fdb	0o6603		;	6603		/ load Current Address
	.asciz	"OS/8"		;			/ OS identification



bootok:	.asciz	" bootstrap loaded at "
noboot:	.fcb	CR,LF
	.fcc	"no OS selected"
	.fcb	CR,LF,0


	.else


;  getoct - get octal value
;
;    call   : [B] = number of octal digits (1 to 4)
;    return : V=1     - entry aborted
;		        [D] = invalid
;	      V=0 C=1 - *first* entry is CR (used to accept "default")
;		        [D] = invalid
;	      V=0 C=0 - valid entry, value in [D]

getoct:	clra
	tfr	d,x		; set loop counter for requested digits
	pshs	x		; save for 1st char == CR check
	clrb
	std	*read8		; init result
getoc1:	jsr	inch		; get char, no echo
	cmpa	#CR
	beq	getocd		; - accept default if 1st entry
	cmpa	#CtrlC
	beq	getocc		; - ^C : abort entry
	cmpa	#'0
	blo	getoc1		; accept 0-7 only
	cmpa	#'7
	bhi	getoc1
	jsr	outch		; print entry
	suba	#'0		; base-0
	pshs	a		; put octal digit on stack
	ldd	*read8		; get current value
	lslb			; make room for new digit
	rola
	lslb
	rola
	lslb
	rola
	orb	,s+		; add new digit from stack
	std	*read8		; update result
	leax	-1,x
	bne	getoc1
getoce:	andcc	#0xFC		; V=0 C=0 - result value in [D]
	puls	x,pc		; clean up stack
;
getocd:	cmpx	,s		; was CR the first entry?
	bne	getoce		; n - return accumulated value
	orcc	#1		; y - V=0 C=1
	puls	x,pc
;
getocc:	orcc	#2		; V=1 - abort entry
	puls	x,pc

	.endif





;  * ISSSTS - build Indicator Selector Switch STATUS data
;
;    call   : ### also called from FIRQ routine if ISS at "STATUS" ###
;    return : [X] - ptr to collected bit mask (word)
;    note   : Greater Than flag not (yet) implemented
;
;    ----------------------------------------------------------------
;   | LNK  GT   INT   NO  ION   UM  IF0  IF1   IF2  DF0   DF1   DF2  |
;   |           BUS  INT                                             |
;   |=====----===========+===========================================|
;   |  0    1    2    3    4    5    6    7    8    9    10     11   |
;    --------------------+-------------------------------------------

ISSSTS:	clra
	ldb	*Link
	bitb	#Lmask
	beq	stshi2
	lda	#LD_lnk		; Link
stshi2:				; *** Greater Than flag ***
stshi1:	ldb	*IOflgs		; test Jdelay (msb)
	bpl	stslo5
	ora	#LD_nia		; No Interrupt Allowed
stshi0:	andb	#Jdelay^0xFF
	orb	*IOflgs+1
	beq	stslo5
	ora	#LD_irb		; Interrupt Bus
stslo5:	sta	*issts2
	lda	*WorkD
	anda	#Dbits		; Data Field 0-1-2
	ldb	*WorkUI
	bpl	stslo6
	ora	#LD_umd		; User Mode
stslo6:	andb	#Fbits
	lslb
	lslb
	lslb			; Instruction Field 0-1-2
	stb	*issts2+1
	adda	*issts2+1
	ldx	*emufnc
	cmpx	#Chkint
	bne	stslo7
	ora	#LD_ion		; Interrupt On
stslo7:	sta	*issts2+1
	ldx	#issts2
	rts



	.if	RealCons
	.else

;  ShwReg - show all register information (incl. ASCII artwork)
;
;    call   : [A] = requested register to display
;    return : print "lights" based on ISS selection

ShwReg:	ldx	#line1
	jsr	pdatao
	cmpa	#ISSste		; check selected display line
	bne	ISS_5
;     ______________________________________________________________
;    |  F    D    E   IR0  IR1  IR2  MD   DATA  SW PAUSE  BRK  BRK  |
;    |                               DIR  CONT            PROG      |
;    |==============================================================|
;    |  ##   ##   ##   ##   ##   ##   ##   ##   ##   ##   ##   ##   | STATE
;    ----------------------------------------------------------------

ISS_6:	ldd	*rc_ste		; get STATE data
	ldy	#stenam		;     and ID,
	ldx	#stelin		;     1st line STATE format, ASCII art
	bra	ISSDPY

ISS_5:	cmpa	#ISSsts
	bne	ISS_4
;     ______________________________________________________________
;    | LNK   GT  INT   NO  ION   UM  IF0  IF1  IF2  DF0  DF1  DF2   |
;    |           BUS  INT                                           |
;    |==============================================================|
;    |  ##   ##   ##   ##   ##   ##   ##   ##   ##   ##   ##   ##   | STATUS
;    ----------------------------------------------------------------

	bsr	ISSSTS		; collect STATUS flag bits
	ldd	,x		; load the flag bits in [D]
	ldy	#stsnam
	ldx	#stslin
	bra	ISSDPY
;     ______________________________________________________________
;    |  0    1    2    3    4    5    6    7    8    9    10   11   |
;    |==============================================================|
;    |  ##   ##   ##   ##   ##   ##   ##   ##   ##   ##   ##   ##   | AC/MD/MQ/BUS
;    ----------------------------------------------------------------

ISS_4:	ldx	#numlin
	cmpa	#ISSbus
	bne	ISS_3
	ldd	#0		; "BUS" (to be implemented ...)
	ldy	#busnam
	bra	ISSDPY

ISS_3:	cmpa	#ISSmd
	bne	ISS_2
	ldd	*MD
	ldy	#mdnam
	bra	ISSDPY

ISS_2:	cmpa	#ISSmq
	bne	ISS_1
	ldd	*MQ
	ldy	#mqnam
	bra	ISSDPY

ISS_1:	ldd	*AC		; must be "AC"
	ldy	#acnam

   ; [D] - light data
   ;       for every bit [11-0] set "##" printed, else "  "
   ; [Y] - register ID text string
   ; [X] - pointer to 1st line format, ASCII art

ISSDPY:	pshs	d
	jsr	pdatao		; print register ASCII art line(s)
	lda	#'|
	jsr	outch
	lda	#'=
	ldb	#62
	jsr	OUTABT		; lights separator line
	ldx	#line2
	jsr	pdatao		; goto "lights" line
	puls	d
	pshs	b
	lsla			; move to 1st bit in msb
	lsla
	lsla
	lsla
	ldb	#4		; print 4 bits [11-8]
	bsr	shwlgt		; show the 4 "lights"
	puls	a
	ldb	#8		; print 8 bits [7-0]
	bsr	shwlgt
	tfr	y,x
	jsr	pdatao		; print register name
	jsr	docrlf
	lda	#'-
	ldb	#64
	jsr	OUTABT		; lights separator line
	jmp	docrlf



;  * shwlgt - show "lights"
;
;    call   : [A] = "light" bits
;	      [B] = # lights to print
;    return : [A], [B], [X] destroyed

shwlgt:	ldx	#OFF2sp		; assume bit = clear
	lsla
	bcc	shwlg1
	ldx	#ON3sp		; nope,  bit = set
shwlg1:	jsr	pdatao
	decb
	bne	shwlgt
	rts



   ; ASCII "art" register display

line1:	.fcb	CR,LF
	.fcc	" ______________________________________________________________"
	.fcb	CR,LF,0
stenam:	.asciz	"| STATE"
stelin:	.fcc	"|  F    D    E   IR0  IR1  IR2  MD   DATA  SW PAUSE  BRK  BRK  |"
	.fcb	CR,LF
	.fcc	"|                               DIR  CONT            PROG      |"
	.fcb	CR,LF,0
stsnam:	.asciz	"| STATUS"
stslin:	.fcc	"| LNK   GT  INT   NO  ION   UM  IF0  IF1  IF2  DF0  DF1  DF2   |"
	.fcb	CR,LF
	.fcc	"|           BUS  INT                                           |"
	.fcb	CR,LF,0
numlin:	.fcc	"|  0    1    2    3    4    5    6    7    8    9    10   11   |"
	.fcb	CR,LF,0
line2:	.fcc	"|"
	.fcb	CR,LF
	.asciz	"|  "
acnam:	.asciz	"| AC"
mdnam:	.asciz	"| MD"
mqnam:	.asciz	"| MQ"
busnam:	.asciz	"| BUS"
ON3sp:	.asciz	"##   "
OFF2sp:	.asciz	"     "

SSmsg:	.fcb	CR,LF,CR,LF
	.fcc	"   <SPC>:single step   <CR>:run   <K>:keyboard input"
	.fcb	CR,LF
	.fcc	" <1>:STATE  <2>:STATUS  <3>:AC  <4>:MD  <5>:MQ  <6>:BUS"
	.fcb	CR,LF,CR,LF,0

Kbdmsg:	.fcb	CR,LF
	.asciz	"keyboard entry: "

askpc:	.fcb	CR,LF
	.asciz	"enter start PC [0200] 4-digit octal ->"

	.endif


;====================================================================================
;====================================================================================


   ; simulator messages

Sysmsg:	.fcb	CR,LF
	.fcc	"*** pdp8/e  SIMULATOR  v1.5 - 24aug2005 ***"
	.fcb	CR,LF,0

clrcor:	.fcb	CR,LF
	.asciz	"[i] initialising pdp8/e core memory fields ..."
okmsg:	.asciz	" ok"

ldrmss:	.fcb	CR,LF
	.asciz	"    download binary file ->  end with ^Q   ... "
ldrmsn:	.fcc	"skipped"
	.fcb	CR,LF,0
ldrmse:	.fcc	"completed"
	.fcb	CR,LF
	.fcc	"    load another .BIN  [Y/N]"
dotdot:	.asciz	" : "
Lderr:	.asciz	"Error loading binary file"
ldrcnt:	.fcb	CR,LF
	.asciz	"! at file byte count 0x"
Errfld:	.asciz	"Field specification invalid"
Errmem:	.asciz	"Field does not exist"
Errcks:	.asciz	"Checksum failed"
;NoXON:	.asciz	"No XON received as a start character after load"

Insterr:	.asciz	"Illegal instruction"
KM8Eerr:	.asciz	"Unimplemented KM8-E instruction"
PC02_IOP:	.asciz	"illegal IOP on device 01 (PC02)"
PC03_IOP:	.asciz	"illegal IOP on device 02 (PC03)"
TTI_3_7:	.asciz	"IOP #3 or #7 on device 03 (TTI)"
TTO_3_4:	.asciz	"IOP #3 or #7 on device 04 (TTO)"

	.if	RF08dev
niRF08:		.asciz  "IOT for RF08 Disk File not implemented"
	.else
niDF32:		.asciz  "IOT for DF32 Disk File not implemented"
	.endif

noRK8P:		.asciz  "IOT for RK8-P device not implemented"
;Bitmsg:	.asciz	"Bits set in locations that should be 0"

Clrmsg:	.asciz	"Error in Clear subroutine (from IOset call)"
Hltmsg:	.asciz	"Program halt (HLT microcode in OPR instruction)"
NoIerr:	.asciz	"Unimplemented interrupt IOP (IOT-00 #6)"
NoFLD:	.asciz	"No memory allocated for this field"


config:	.fcb	CR,LF
	.fcc	"PDP-8/e  12 kW core"
	.fcb	CR,LF
	.fcc	"  KM8-E  memory extension and timeshare control"
	.fcb	CR,LF
	.fcc	"  KE8-E  EAE (extended arithmetic element)"
	.fcb	CR,LF
	.fcc	"  KL8-E  Teletype keyboard/printer"
	.fcb	CR,LF
	.fcc	"  PC8-E  high-speed paper tape reader and punch"
	.fcb	CR,LF
	.if	RF08dev
	.fcc	"  RF08   Disk File, 4 RS08 disks"
	.else
	.fcc	"  DF32   DECdisk File, 4 DS32 Disk Extenders"
	.endif
	.fcc	" -- ID=#"
	.fcb	0



;====================================================================================
;====================================================================================

	.if	ide2pdp

useLBA	=	1		; 1 = LBA mode  //  0 = C/H/S mode
				; IDEHdO: 111*  //  IDEHdO: 101*



;  IDEini - initialize the IDE disk
;
;  call   : I/O Board - 6821 PIA must be initialized
;  return : (iodatd) initialised = IDE I/O data direction
;	    (idectl) initialised
;	    IDE disk reset

IDEini:	ldx	#idemsg		; print progress (if verbose mode)
	jsr	vbprnt
	lda	#HDdirR		; data direction := INPUT
	sta	*iodatd		; set I/O data direction = read
	ldb	#WRide_x	; get port with DDIR bit
	jsr	IO_WRT		; and set I/O data direction

	ldd	#SPC
	std	*seccyl		; set up #Sectors per Cylinder
	ldd	#SPT
	std	*sectrk		; set up #Sectors per track

	lda	#DCNop		; set control: all signals inactive
	sta	*idectl		; store active IDE control byte
	ldb	#WRide_c	; get IDE control output port #
	jsr	IO_WRT		; write control byte to IDE control
	lda	#DCNop-DCRes
	jsr	IO_WRT		; assert RESET*
	clra
10$:	deca			; pulse width > 25 uSec (spec)
	bne	10$
	lda	#DCNop		; negate RESET*
	jsr	IO_WRT

	lda	#CmdDiag	; execute drive diagnostics
	jsr	wrtCmd
	jsr	WtNbsy		; wait for drive not busy status
	jsr	WtDrdy		; wait for drive ready status

    ; ATA spec requires that SoftReset bit must be written to 0

	lda	#IDEDvc
	jsr	SetAdr		; select device control register
	lda	#IDEIrq		; [IRQenable] = disabled
;	anda	#IDERes^0xFF	; [SoftReset] = 0
	jsr	wrbyte
	jsr	WtNbsy		; wait for drive not busy status
	jsr	WtDrdy		; wait for drive ready status

    ; select the disk device, I use "select head number #0" (if C/H/S mode)

	lda	#IDEHd
	jsr	SetAdr
	lda	#IDEHdO		; use LBA or C/H/S format, drive #0, (head #0)
	jsr	wrbyte
	jsr	WtNbsy		; wait for disk not busy
	jsr	WtDrdy		; wait for drive ready status

	lda	#CmdRecal
	jsr	wrtCmd		; issue the recalibrate command
	jsr	WtNbsy		; and wait until command executed
	jsr	WtDrdy		; wait for drive ready status
	ldx	#okmsg		; print progress (if verbose mode)
	jmp	vbprnt		; and RTS


idemsg:	.fcb	CR,LF
	.asciz	"[i] initialising IDE hard disk interface   ..."





; ###################################################################################
; ##										   ##
; ##		Support commands : spin disk up / spin disk down		   ##
; ##		Status reporting : Disk error formatting and display		   ##
; ##		Disk Identification summary display				   ##
; ##										   ##
; ###################################################################################


	.if	0		; Spin up/down Compact Flash card ?
				; no need for!!

;  SpuDsk - spin up the disk
;
;  call   : -
;  return : -

SpuDsk:	jsr	WtNbsy		; wait till disk is ready
	lda	#CmdStrt
	jmp	wrtCmd		; issue the command & RTS



;  SpdDsk - spin down the disk
;
;  call   : -
;  return : -

SpdDsk:	jsr	WtNbsy		; wait till disk is ready
	lda	#CmdStop
	jmp	wrtCmd		; issue the command & RTS

	.endif			; Spin up/down




;  DskErr - give extensive error reporting for disk errors
;
;  call   : [X] - pointer to command packet
;		[0:1] = buffer (data information) address
;  return : error messages about what went wrong with a disk access
;	    (idesv2) used for temporary local storage

DskErr:	pshs	d
	stx	*idesv2		; save pointer to LBA
	ldx	#ErrDsk		; print general error message & error status text
	jsr	pdatao
	lda	#IDESts		; - IDE status register
	jsr	rdDreg
	jsr	OUTB2O
	jsr	pdatao		; print error bits text
	lda	#IDEErr		; - error bits
	jsr	rdDreg
	jsr	OUTB2O
	jsr	pdatao		; print requested LBA where error occurred
	ldx	*idesv2
	ldd	LBA3,x
	jsr	OUTB2O
	tfr	b,a
	jsr	OUTB2O
	ldd	LBA1,x
	jsr	OUTB2O
	tfr	b,a
	jsr	OUTB2O

	.if	useLBA
	ldx	#ErrLBA		; print disk LBA where it happened
	jsr	pdatao
	lda	#IDEHd
	jsr	rdDreg
	anda	#0b00001111	; LBA[27:24]
	jsr	OUTB2O
	lda	#IDEcylH
	jsr	rdDreg
	jsr	OUTB2O		; LBA[23:16]
	lda	#IDEcylL
	jsr	rdDreg
	jsr	OUTB2O		; LBA[15:8]
	lda	#IDEsec
	jsr	rdDreg
	jsr	OUTB2O		; LBA[7:0]
	jsr	docrlf
	puls	d,pc

	.else
	ldx	#ErrCHS		; print disk CHS where it happened
	jsr	pdatao
	lda	#IDEcylH
	jsr	rdDreg
	jsr	OUTB2O
	lda	#IDEcylL
	jsr	rdDreg
	jsr	OUTB2O
	bsr	slash
	lda	#IDEHd
	jsr	rdDreg
	jsr	OUTB2O
	bsr	slash
	lda	#IDEsec
	jsr	rdDreg
	jsr	OUTB2O
	jsr	docrlf
	puls	d,pc

slash:	lda	#'/
	jmp	outch		; print "/" and RTS

	.endif




; ###################################################################################
; ##										   ##
; ##		LEVEL 2  -  read and write ONE sector (512) bytes		   ##
; ##			    translate LBA to CHS and load parameters		   ##
; ##										   ##
; ###################################################################################



;  rdDsec - reads one sector from the IDE disk
;
;  call   : [X] - pointer to command packet
;		[0:1] = buffer (data destination) address
;		[2:5] = LBA3:LBA0
;  return : cC=1 -> error occurred
;	    cC=0 -> data read into the supplied buffer

rdDsec:	jsr	WtNbsy		; wait for busy bit cleared (disk ready)
	lda	#IDEHd		; select disk 0
	jsr	SetAdr
	lda	#IDEHdO		; format: use LBA or C/H/S, drive #0
;	anda	#IDEHdA		; select head #0
	jsr	wrbyte
				; load parameters
	jsr	SetLBA		; set address
	lda	#CmdRead	; issue read command
	jsr	wrtCmd
	jsr	WtNbsy		; and wait for busy gone
				; check for errors
	jsr	rdDsts		; get status byte
	bita	#StsErr		; test error bit
	beq	rdsec1
	orcc	#0x01		; error -> cC=1
	rts
rdsec1: jsr	WtDdrq		; wait for DRQ
	pshs	x		; transport the data block to destination
	ldx	SDA,x		; get destination address
	bsr	rddata		; get the data
	puls	x
	andcc	#0xFE		; cC=0 -> all ok
	rts



;  rddata - read one block from the IDE disk
;
;  call   : [X] - pointer to 256 word buffer
;	    all disk parameters (C/H/S) loaded into drive's registers
;	    Data_Request is active (been awaited by routine that calls this function)
;  return : supplied buffer holds 256 words from specified (preloaded) sector
;	    local variable (idesv1) = $00

rddata:	pshs	d,x
	lda	#IDEData
	jsr	SetAdr		; select Data Register
	jsr	SetIn
	clr	*idesv1		; set 256 read operations counter
rdblk:	ldb	#WRide_c	; set IDE control OUTPUT port #
	lda	*idectl		; get current active IDE control byte
	anda	#DCIor^0xFF	; assert RD*
	jsr	IO_WRT
	ldb	#RDide_dl	;    set IDE data INPUT port #
	jsr	IO_RD		;    get data byte
	sta	,x+
	ldb	#RDide_du	;    set IDE data INPUT port #
	jsr	IO_RD		;    get data byte
	sta	,x+
	ldb	#WRide_c	; set IDE control OUTPUT port #
	lda	*idectl		; negate RD*
	jsr	IO_WRT
	dec	*idesv1
	bne	rdblk		; until 256 read operations done
	puls	d,x,pc



;  wrDsec - writes one sector to the IDE disk
;
;  call   : [X] - pointer to command packet
;		[0:1] = buffer (data source) address
;		[2:4] = LBA High, Middle, Low
;  return : cC=1 -> error occurred
;	    cC=0 -> data written to the disk OK

wrDsec:	jsr	WtNbsy		; wait for busy bit cleared (disk ready)
	lda	#IDEHd		; select disk 0
	jsr	SetAdr
	lda	#IDEHdO		; format: use LBA or C/H/S, drive #0
;	anda	#IDEHdA		; select head # 0
	jsr	wrbyte
	bsr	SetLBA		; set address
	lda	#CmdWrite	; issue write command
	jsr	wrtCmd
	jsr	WtNbsy		; wait till command ready
				; check for errors
	jsr	rdDsts		; get status byte
	bita	#StsErr		; test error bit
	beq	wrsec1		; no error -> continue
	orcc	#0x01		; error -> cC=1
	rts
wrsec1:	jsr	WtDdrq		; wait for DRQ
	pshs	x		; transport the data block
	ldx	SDA,x		; get source address
	bsr	wrdata		; save the data to disk
	puls	x
	andcc	#0xFE		; cC=0 -> all ok / enable FIRQ
	rts



;  wrdata - write one block of data to the IDE disk
;
;  call   : [X] - pointer to 256 word data buffer
;	    all disk parameters (C/H/S) loaded into drive's registers
;	    Data_Request is active (been awaited by routine that calls this function)
;  return : supplied buffer written to specified (preloaded) sector
;	    local variable (idesv1) = $00

wrdata:	pshs	d,x
	jsr	SetOut		; set I/O databus to output
	lda	#IDEData
	jsr	SetAdr		; select Data Register
	clr	*idesv1		; set 256 write operations counter
wrblk:	ldb	#WRide_dl	; set IDE data OUTPUT port #
	lda	,x+
	jsr	IO_WRT		; write data byte
	ldb	#WRide_du	; set IDE data OUTPUT port #
	lda	,x+
	jsr	IO_WRT		; write data byte
	ldb	#WRide_c	; set IDE control OUTPUT port #
	lda	*idectl		; get current active IDE control byte
	anda	#DCIow^0xFF	; assert WR*
	jsr	IO_WRT
	lda	*idectl		; negate WR*
	jsr	IO_WRT
	dec	*idesv1
	bne	wrblk		; until 256 write operations done
	puls	d,x,pc





;  SetLBA - sets the LBA for the next transfer
;
;  call   : [X] - pointer to command packet
;		[1:0] = buffer address (not used, not changed)
;		[2:5] = LBA3:LBA0
;  return : registers of the IDE disk loaded
;  remark : only -ONE- sector is processed
;
; In LBA modus the sectors are simply numbered from 0 to -big-.
; The lowest byte of the LBA sector number is written into the sector number
; register, the middle 16 bits of the LBA sector number are written in the cylinder
; number registers (low and high, all 16 bits are used). The highest 4 bits of the
; LBA sector number are written in the head and device register. That gives you 28
; bits of LBA sector number.


	.if	useLBA		; LBA mode
SetLBA:	lda	#IDEsec
	bsr	SetAdr
	lda	LBA0,x		; LBA[7:0]
	bsr	wrbyte
	lda	#IDEcylL
	bsr	SetAdr
	lda	LBA1,x
	bsr	wrbyte		; LBA[15:8]
	lda	#IDEcylH
	bsr	SetAdr
	lda	LBA2,x
	bsr	wrbyte		; LBA[23:16]
	lda	#IDEHd
	bsr	SetAdr
	lda	LBA3,x
	anda	#IDEHdA		; isolate low nibble == most significant 4 bits
	ora	#IDEHdO		; format: use LBA, drive #0
	bsr	wrbyte		; LBA[27:24], drive #0
	lda	#IDEnum
	bsr	SetAdr
	lda	#1		; set up for ONE sector transfer
	bra	wrbyte		; and RTS

	.else			; C/H/S mode

;		a stack frame is built to make the conversion to C/H/S
;		these are the offsets in the stack frame (after tsx, relative to X)
;
LBS3	=	0	; stack frame LBAS
LBS2	=	1	;
LBS1	=	2	;
LBS0	=	3	; cylinder high
;

SetLBA:	pshs	x		; save X, and load input data onto stack
	lda	LBA0,x
	pshs	a		; - 3
	lda	LBA1,x
	pshs	a		; - 2
	lda	LBA2,x
	pshs	a		; - 1
	lda	LBA3,x
	pshs	a		; - 0
	tfr	s,x		; setup [X] as stack frame pointer

    ; init for divide by sectors/cylinder to get the cylinder number.
    ; Result will fit in 16 bits, so 16 bits result is enough.

	ldb	#16		; 16 divide loops
	stb	*idesv1
lbac1:	asl	LBS0,x		; shift divident
	rol	LBS1,x
	rol	LBS2,x
	rol	LBS3,x
	ldd	LBS3,x		; get high word
	subd	*seccyl		; test if divisor subtraction fits
	bcs	lbac2		; n - next loop again
	std	LBS3,x		; y - set lower cylinder bit, store result
	inc	LBS0,x
lbac2:	dec	*idesv1
	bne	lbac1
				; write result to disk registers
	lda	#IDEcylL
	bsr	SetAdr
	lda	LBS0,x
	bsr	wrbyte		; 8 low bits of Cylinder #
	lda	#IDEcylH
	bsr	SetAdr
	lda	LBS1,x
	bsr	wrbyte		; 2 high bits of Cylinder #
				; see about Head number
	lda	#0		; simple substract will do the work
	sta	LBS0,x		; max #Head times ...
	ldd	LBS3,x		; get remaining blocks count
				; test if one head more
lbac3:	subd	*sectrk		; minus one track's sectors
	bcs	lbac4		; result negative ?
	inc	LBS0,x		; n - one more head
	bra	lbac3
lbac4:	addd	*sectrk		; y - but gone one too far
	addd	#1		; sector count starts at 1 !
	std	LBS3,x		; save Sector#
				; write to disk registers
	lda	#IDEHd		; Head is special
	bsr	SetAdr
	lda	LBS0,x
	anda	#IDEHdA		; isolate the HS3,HS2,HS1,HS0 bits
	ora	#IDEHdO		; format: use C/H/S, drive #0
	bsr	wrbyte		; write Head #
	lda	#IDEsec
	bsr	SetAdr
	lda	LBS2,x		; low byte only as Sector #
	bsr	wrbyte
	lda	#IDEnum		; set address for sector count
	bsr	SetAdr
	lda	#1		; set up for ONE sector transfer
	bsr	wrbyte
	leas	4,s		; clean up the stack - 4 bytes LBA
	puls	x,pc
	.endif





; ###################################################################################
; ##										   ##
; ##		LEVEL 1  -  write command / write byte				   ##
; ##			    set device register address				   ##
; ##			    read status / read register / read byte		   ##
; ##			    test/check READY, BUSY, DATA-REQ status		   ##
; ##										   ##
; ###################################################################################



;  wrtCmd - write one command to the IDE disk
;
;  call   : [A] - command
;  return : command written in command register of IDE drive
;	    no other registers changed
;	    current I/O (IDE) databus direction = WRITE

wrtCmd:	pshs	a
	lda	#IDECmd
	bsr	SetAdr		; select Command Register
	puls	a
;	bra	wrbyte		; issue command and RTS <fall through >



;  wrbyte - write one byte to the IDE disk
;
;  call   : [A] - byte to write
;	    command to IDE drive has been issued (idectl)
;  return : no other registers changed
;	    current I/O (IDE) databus direction = WRITE

wrbyte:	pshs	d
	bsr	SetOut		; set I/O databus to output
	ldb	#WRide_dl	; set IDE data OUTPUT port #
	bsr	IO_WRT		; write data byte
	ldb	#WRide_c	; set IDE control OUTPUT port #
	lda	*idectl		; get current active IDE control byte
	anda	#DCIow^0xFF	; assert WR*
	bsr	IO_WRT
	lda	*idectl		; negate WR*, latch data into drive's register
	bsr	IO_WRT
	puls	d,pc



;  SetAdr - set an address on the IDE disk
;
;  call   : [A] - device register address [CS1*,CS0*,A2,A1,A0]
;  return : IDE control bit CS1*,CS0*,A2,A1,A0 written
;			bit RST*,RD*,WR* set (inactive state)
;	    (idectl) - current written IDE control byte
;	    no other registers changed
;	    current I/O (IDE) databus direction not changed

SetAdr:	sta	*idectl		; save active IDE control byte
	ldb	#WRide_c	; set IDE control output port #
	bra	IO_WRT		; write to IDE control and RTS



;  rdDreg - read register (byte) from the IDE disk
;
;  call   : [A] - IDE device register address
;  return : [A] - data read from specified register
;	    no other registers changed
;	    current I/O (IDE) databus direction = READ

rdDreg:	bsr	SetAdr		; setup requested register
	bra	rdbyte		; read byte of register and RTS



;  rdDsts - get status of the IDE disk
;
;  call   : -
;  return : [A] - IDE drive status register info
;	    no other registers changed
;	    current I/O (IDE) databus direction = READ

rdDsts:	lda	#IDESts		; address = status register
	bsr	SetAdr
;	bra	rdbyte		; get the status and RTS <fall through>



;  rdbyte - read one byte from the IDE disk
;
;  call   : command to IDE drive has been issued (idectl)
;  return : [A] = byte read
;	    no other registers changed
;	    current I/O (IDE) databus direction = READ

rdbyte:	pshs	b
	bsr	SetIn		; set I/O databus to input
	ldb	#WRide_c	; set IDE control OUTPUT port #
	lda	*idectl		; get current active IDE control byte
	anda	#DCIor^0xFF	; assert RD*
	bsr	IO_WRT
	ldb	#RDide_dl	;    set IDE data INPUT port #
	bsr	IO_RD		;    get data byte
	pshs	a
	ldb	#WRide_c	; set IDE control OUTPUT port #
	lda	*idectl		; negate RD*, latch drive's data register
	bsr	IO_WRT
	puls	a
	puls	b,pc



;  WtDrdy - wait till the drive indicates ready status
;
;  call   : -
;  return : when RDY bit of IDE drive is set
;	    no registers changed
;	    current I/O (IDE) databus direction = READ

WtDrdy:	pshs	d
10$:	bsr	rdDsts		; get status
	anda	#StsRdy		; ready ?
	beq	10$		; n - BUSY WAIT
	puls	d,pc



;  WtDdrq - wait till the drive indicates "ready for data" status
;
;  call   : -
;  return : when DRQ bit of IDE drive is set
;	    no registers changed
;	    current I/O (IDE) databus direction = READ

WtDdrq:	pshs	d
10$:	bsr	rdDsts		; get status
	anda	#StsDrq		; ready for data ?
	beq	10$		; n - BUSY WAIT
	puls	d,pc



;  WtNbsy - wait till the drive indicates ready status
;
;  call   : -
;  return : when BSY bit of IDE drive is clear
;	    no registers changed
;	    current I/O (IDE) databus direction = READ

WtNbsy:	pshs	d
10$:	bsr	rdDsts		; get status
	anda	#StsBsy		; still busy ?
	bne	10$		; y - BUSY WAIT
	puls	d,pc




;  SetIn - set data I/O port to INPUT (read data) from the IDE disk
;
;  call   : -
;  return : *all* bits of port WRide_x written
;	    no registers changed
;	    (iodatd) reflects current I/O (IDE) databus direction

SetIn:	pshs	d
	lda	#HDdirR		; data direction := INPUT
	bra	setdir



;  SetOut - set data D[7:0] I/O port to OUTPUT (write data) to the IDE disk
;
;  call   : -
;  return : *all* bits of port WRide_x written
;	    no registers changed
;	    (iodatd) reflects current I/O (IDE) databus direction

SetOut:	pshs	d
	lda	#HDdirW		; data direction := OUTPUT
setdir:	cmpa	*iodatd		; current I/O data direction == requested ?
	beq	10$
	ldb	#WRide_x	; n - port with DDIR bit
	sta	*iodatd		;
	bsr	IO_WRT		;     set requested I/O data direction
10$:	puls	d,pc







; ###################################################################################
; ##										   ##
; ##		LEVEL 0  -  access to physical I/O Board ports			   ##
; ##										   ##
; ###################################################################################


;  IO_WRT - write to physical output port
;
;  call   - [B] = I/O Board number , port number
;           [A] = data byte
;  return - no registers changed
;	    (piadir) reflects current I/O - Core Board communication direction

IO_WRT:	pshs	a		; set data output port == OUTPUT (if needed)
	lda	*piadir		; was direction already "OUTPUT" ?
	bita	#CA2BIT
	beq	iowrt		; y - no action needed
	anda	#0xFF-CA2BIT	; n - switch to "OUTPUT"
	sta	PIACA		; 1. only the 245 buffer (prevent short-circuit)
	sta	*piadir		; set flag: port A set to "OUTPUT"
	anda	#0xFF-DIRBIT	; 2. access data direction reg A
	sta	PIACA
	lda	#0xFF
	sta	PIADA		;    and set port A == output
	lda	*piadir
	sta	PIACA		; 3. access data output reg A
iowrt:	puls	a
	sta	PIADA		; load data
	nop
	orb	#0xFF-IOADDR	; negate DIS3S control bit
;	orb	#RDWR		; direction: WRITE: enable "write" LS138
	stb	PIADB		; setup correct output latch "address"
	andb	#0xFF-DIS3S
	stb	PIADB		; assert DIS3S: enable LS138 decoders
	nop			; hardware settle delay
	orb	#DIS3S
	stb	PIADB		; negate DIS3S: disable LS138 decoders
	andb	#IOADDR		; restore original I/O Board & port number
	rts



;  IO_RD - read from physical input port
;
;  call   - [B] = I/O Board number , port number
;  return - [A] = data byte
;           all other registers not changed
;	    (piadir) reflects current I/O - Core Board communication direction

IO_RD:	lda	*piadir		; was direction already "INPUT" ?
	bita	#CA2BIT
	bne	iord		; y - no action needed
	anda	#0xFF-DIRBIT	; n - switch to "INPUT"
	sta	PIACA		;     1. access data direction reg A
	clra
	sta	PIADA		;        and set port A == input
	lda	*piadir		;     2. now set buffer-74245 to input
	ora	#CA2BIT+DIRBIT	;        (prevent short-circuit moment
	sta	PIACA		;         between buffer-74245 and PIA)
	sta	*piadir		;     3. set flag: port A set to "INPUT"
iord:	orb	#0xFF-IOADDR-RDWR  ; negate DIS3S control
;	andb	#0xFF-RDWR	   ; direction: READ: enable "read" LS138
				   ; read physical port data
	stb	PIADB		   ; > setup correct input latch "address"
	andb	#0xFF-DIS3S
	stb	PIADB		   ; > enable LS138 decoders
	nop			   ; > hardware settle delay
	lda	PIADA		   ; > read switch/toggle selected input port
	orb	#DIS3S
	stb	PIADB		   ; > disable LS138 decoders
	andb	#IOADDR		   ; restore original I/O Board & port number
	rts



    ; disk error message strings
    ; formatted text: keep in this order!

ErrDsk:	.fcb	CR,LF
	.fcc	">>> Disk Error Occurred <<<"
	.fcb	CR,LF
	.asciz	"   Status :"
	.fcb	CR,LF
	.asciz	"   Error  :"
	.fcb	CR,LF
	.asciz	"   LBA req:"
	.if	useLBA
ErrLBA:	.fcb	CR,LF
	.asciz	"   LBA dsk:"
	.else
ErrCHS:	.fcb	CR,LF
	.asciz	"   CHS dsk:"
	.endif

	.endif



;________________________________________________________________________________
;#------------------------------------------------------------------------------#
;#										#
;#		*  command line switch processing				#
;#		*  emulation command processing (direct response)		#
;#		*  trace facility (part of emulation command set)		#
;#______________________________________________________________________________#
;--------------------------------------------------------------------------------


; strtup - start-up: scan command line for options
;
;     syntax: pdp8 [-v] [-c] [-t] [-d#] [-bXXX]
;
;     "-v"	indicates verbose output at start up
;     "-c"	indicates the use of CORE "memory" (not erased at start-up)
;     "-t"	indicates trace facility wanted from the start
;     "-d#"	indicates the "disk number" to use in the simulation
;		   where # can be a number between 0 and 9 (both inclusive)
;     "-bXXX"	indicates which OS bootstrap must be selected
;		   where 'xxx' can be either 'DMS' or 'OS8' (must be UPPERCASE)
;     The order is not important, however only 1 space is allowed.
;     If the "-v", "-c" or "t" option is specified more than once, the result
;     is always that the option is turned on (not toggle on/off function).
;     If the "-d" option is specified more than once, the last "-d" is active.
;     If the "-b" option is specified more than once, the last "-b" is active.
;     Options recognized before a syntax error are set, all options after the
;     syntax error are not processed.
;
;     All CPU registers can be used freely, only DP has been initialised.

strtup:	clr	*dsknum		; default "disk" #0
	clra
	clrb
	std	*secoff		;	  "disk" #0 has sector offset 0
	std	*secoff+2
	clr	*core		;	  MOS memory (erase at start-up)
	clr	*verbos		;	  brief identification output
	std	*boottb		;	  no OS bootstrap selected
	clr	*trcflg		;	  no trace enabled
	clr	*trcsts
	clr	*trcdmp		;	  trace dump re-routing disabled
	clr	*rf8inf		;	  RF08 "register" dump OFF

	.if	DEFER_IO
	clr	*defer		; global "deferred I/O" active flag := FALSE
	ldx	#deftab
	clrb			; for ( i=0 to MAXDEV )
1$:	tfr	b,a
	lsla
	lsla			; {
	clr	a,x		;   deftab[i].active := FALSE
	incb			; }
	cmpb	#MAXDEV
	bne	1$
	.endif

	ldx	linptr
	ldb	#5		; max # arguments
10$:	pshs	b
	jsr	getopt		; pick up argument
	puls	b
	beq	11$		; - no args or bad syntax: done
	decb
	bne	10$
11$:	lda	*trcflg		; was "-t" specified ?
	bne	trprmp		; y - sollicit for simulation control command
	rts			; n - all done

  ; process the -t commandline option: query for trace specification
  ;  -> output result from trace spec is at "Stptrc" during running simulation

trprmp:	tfr	cc,a		; get current ACIA interrupt flag status
	sta	*trcirq
	orcc	#0x10		; disable (pdp8) ACIA interrupt
tpromp:	ldx	#trclst-1	; initialize to trace command lists
	stx	comadr
	ldx	#trcms1		; sollicit for trace specification
	jsr	pdatao
	inc	bolflg		; set 'beginning of line' flag
	jsr	getlin		; get line of input
	cmpb	#3		; abort line on a control-c
	beq	tpromp
	ldx	bufbeg		; set syntax scanning pointer
	stx	synptr		; to beginning of buffer/line
	lda	1,x		; get first char
	cmpa	#CR		; reprompt on an empty line (CR only)
	beq	tpromp
	clr	*trcext
	clr	*trcsts
	clr	*trcdmp
				; get the trigger specifier
	lda	#5		; use list 5 to match
	jsr	comand		; search for a match
	beq	tpromp		; try again if just a CR was typed
	bmi	synnok		; syntax not ok
	asla			; ok - process trace specifier
	ldx	#20$-2
	jmp	[a,x]		; jump to processing routine

20$:	.word	trace		; TRACE
	.word	tr_rlt		; REALTIME
	.word	tr_run		; RUN
	.word	tr_stp		; STEP
	.word	tr_dsp		; DISPLAY
	.word	tr_lod		; LOAD


trace:				; get the trace type specifier
	lda	#1		; use list 1 to match
	jsr	comand		; search for a match
	beq	tpromp		; try again if just a CR was typed
	bmi	synnok		; syntax not ok
	asla			; ok - process trace specifier
	ldx	#30$-2
	jmp	[a,x]		; jump to processing routine

30$:	.word	tr_aft
	.word	tr_unt
	.word	tr_bef
	.word	tr_onl
	.word	tr_alw



  ; -------------------------------------------------
  ; SIMULATION CONTROL [STEP, RUN, REALTIME] commands
  ; -------------------------------------------------

tr_stp:	lda	#7		; STEP - use list 7 to match
	jsr	comand		; search for match
	beq	tpromp		; try again if just a CR was typed
	bmi	synnok
	deca			; "COUNT" or "UNTIL" ?
	bne	20$
	jsr	datfmt		; STEP COUNT - get count value (any number base)
	ble	synnok		; illegal number in active base, or no entry
	ldd	nbrhi		; get the entered value
	beq	synnok		; reject count == 0 (this indicates STEP UNTIL)
	std	*trccnt
	lda	#TRCENA+TRCACT+TRC_ST	; enable + active + step
	bra	30$
20$:	clra
	clrb
	std	*trccnt		; STEP UNTIL ==> set count = zero
	jsr	getcdx		; get condition  ( .REG  or  <memory> )
	beq	synnok		; - syntax error
	bvs	synnok		; - not .REG nor <memory>
					; flag STEP trigger command specified
	ora	#TRCENA+TRCACT+TRC_ST	; enable + active + step
30$:	sta	*trcsts			; store the trigger specifier
	bra	nomore


synnok: ldx	bufbeg		; get start of line
10$:	cpx	linptr		; move cursor to the error in syntax
	beq	20$		; at error position?
	jsr 	OUTS		; n - output a space
	inx
	bra	10$
20$:	lda	#Space
	ldb	#11
	jsr	OUTABT		; y - skip trigger sollicit message
	lda	#'^		;     and print up-arrow
	ldb	#CR
	jsr	OUT2CH
	jmp	tpromp		;     try again


  ; ---------------------------------------------------
  ; TRACE [AFTER, UNTIL, BEFORE, ONLY, ALWAYS] commands
  ; ---------------------------------------------------

tr_alw:	lda	#TRCENA+TRCACT+TRC_AL	; enable + always + active
	sta	*trcsts
	bra	nomore

tr_unt:	lda	#TRCENA+TRCACT+TRC_UN	; enable + until + active
	bra	trac01
tr_bef:	lda	#TRCENA+TRCACT+TRC_BF	; enable + before + active
	ldx	#TRCblo
	stx	*trcbip			; init trace data buffer pointers
	stx	*trcbop
	sta	*trcdmp			; non-zero flag: re-route register dump
	bra	trac01
tr_aft:	lda	#TRCENA+TRC_AF		; enable + after
	bra	trac01
tr_onl:	lda	#TRCENA+TRC_OY		; enable + only
trac01:	sta	*trcsts			; store the trace specifier

  ; following the TRACE type specifier is either
  ; 1. a register name (for example .PC, .AC, .IF, etc.)
  ; 2. an address with optional a leading format keyword (HEX, OCT, DEC, BIN)

	bsr	getcdx		; get condition  ( .REG  or  <memory> )
	beq	synnok		; - syntax error
	bvs	synnok		; - not .REG nor <memory>

  ; check if TRACE command was AFTER, UNTIL, ONLY or BEFORE
  ; these ones allows the "OCCUR" keyword as an extra specification option

	lda	*trcsts
	anda	#TRCTYP
	cmpa	#TRC_AF
	beq	occur
	cmpa	#TRC_UN
	beq	occur
	cmpa	#TRC_OY
	beq	occur
	cmpa	#TRC_BF
	bne	nomore
occur:	lda	#11		; look for "OCCUR" keyword
	jsr	comand
	beq	nomore		; nope -
	bmi	synnok		; garbage -
	jsr	datfmt		; get occurrence value (any number base)
	ble	synnok		; illegal number in active base, or no entry
	ldd	nbrhi		; get the entered value
	beq	synnok		; reject count == 0
	std	*trcocc
	lda	*trcext
	ora	#TREOCC		; flag "OCCUR" specified
	sta	*trcext

  ; there should be no more characters on the input line (except delimiters)

nomore:	jsr	skpdlm		; check for garbage at the end
	bcc	synnok		; ccC=0 garbage follows -> try again
	ldx	#comlst-1	; restore CORESYS-09 command table pointer
	stx	comadr		; to allow debugging of simulator itself!
; ---	jsr	docrlf		; blank line before resume (for readability)
	lda	*trcirq
	tfr	a,cc		; restore ACIA irq flag status
	rts



tr_run:	lda	#6		; RUN - use list 6 to match (only UNTIL allowed)
	jsr	comand		; search for match
	lbeq	tpromp		; try again if just a CR was typed
	bmi	runnok
	bsr	getcdx		; get condition  ( .REG  or  <memory> )
	beq	runnok		; - syntax error
	bvs	runnok		; - not .REG nor <memory>
					; flag RUN trigger command specified
	ora	#TRCENA+TRC_RN	; enable + run
	sta	*trcsts		; store the trigger specifier
	bra	occur		; and check for "OCCUR" keyword
runnok:	jmp	synnok


tr_rlt:				; REALTIME
	lda	#TRCENA+TRC_RT	; enable + run
	sta	*trcsts		; store the trigger specifier
	bra	nomore




    ; getcdx - get condition specification
    ;	       the condition is either a .REGISTER  or a  <memmory> specification
    ;	       for call/return see "r_spec" and "m_spec" routines below

getcdx:	bsr	r_spec		; try for .REGISTER command
	beq	10$		; - syntax error
	bvc	10$		; - .REGISTER processed succesfull
	bsr	m_spec		; try for <MEMORY> command
10$:	rts



    ; r_spec - see if .REGISTER is specified (and process it)
    ;
    ;  call   : command line entered and variables valid
    ;  return : [ccZ] = 1 :: syntax error
    ;		[ccZ] = 0
    ;			[ccV] = 1 :: not .REGISTER trigger keyword
    ;			[ccV] = 0 :: .REGISTER trigger spec processed succesfull
    ;				     <trcsts> updated: TRCREG[5] and TRCACC[4:3]
    ;				     [A] = updated <trcsts> data
    ;				     <trcext> updated if "NOT" specified

r_spec:	lda	#2
	jsr	comand		; try search for a register name
	bmi	30$		; not a register
	beq	rm_syn		; end of line - some info must follow!
	deca
	asla			; register ID value * 2
	sta	*trcmem		; store register ID as number, base-0
	lda	#8		; look for CHANGED
	jsr	comand
	beq	rm_syn		; - premature end of line
	bmi	20$		; nope, not "CHANGED"; check for a value
	lda	*trcsts
	ora	#TRCCHG+TRCREG	; access type = REGISTER / CHANGED
	sta	*trcsts
	ldx	#0xFFFF		; set impossible starting "previous" value
	stx	*trcinf		; so any change at all will trigger the trace
10$:	andcc	#0xF9		; - OK, Z=0 / V=0
	rts

20$:	bsr	datnot		; get value (optional "NOT" and number base)
	ble	rm_syn		; illegal number in active base, or no entry
	ldd	nbrhi		; get the entered value
	std	*trctrg
	lda	*trcsts
; --	ora	#TRCREG		; TRCREG bit = 0
; --	sta	*trcsts
	bra	10$
30$:	andcc	#0xFB		; not .REGISTER :: Z=0
	orcc	#0x02		;		   V=1
	rts
rm_syn:	clra			; syntax error  :: Z=1
	rts



    ; datfmt - get a value with optional a leading number base specifier
    ; datnot - as "datfmt", but a leading "NOT" keyword is allowed and processed
    ;
    ;  call   : command line entered and variables valid
    ;  return : [ccZ] = 1 :: syntax error, or end of line detected
    ;		[ccZ] = 0
    ;			[ccN] = 1 :: error in value entry (thus syntax error)
    ;			[ccN] = 0 :: valid value, optional number based included
    ;				     <nbrhi> holds 16-bit number
    ;				     for "datnot" : if "NOT" specified
    ;						    <trcext[TREIDC]> is set

datnot: lda	#9		; look for optional "NOT" keyword
	jsr	comand
	beq	fmtnot		; premature end of line
	bmi	datfmt		; n - it is not "NOT", something else is there
	lda	*trcext		; y - set Invert Datamatch Condition flag
	ora	#TREIDC
	sta	*trcext
datfmt:	lda	#4		; look for optional HEX, OCT, DEC or BIN
	jsr	comand
	beq	fmtnot		; end of line - there must be something!
	bpl	10$		; set specified base
	lda	#3		; not recognized, must be value, set OCTAL base
10$:	sta	ibcode		; set base code
	jsr	mnumber		; get number
fmtnot:	rts



    ; m_spec - see if <MEMORY> is specified (and process it)
    ;
    ;  call   : command line entered and variables valid
    ;  return : [ccZ] = 1 :: syntax error
    ;		[ccZ] = 0
    ;			[ccV] = 1 :: not <memory> trigger "keyword"
    ;			[ccV] = 0 :: <memory> trigger spec processed succesfull
    ;				     <trcsts> updated: TRCMEM[5] and TRCACC[4:3]
    ;				     [A] = updated <trcsts> data
    ;				     <trcext> updated if "NOT" specified

m_spec:	bsr	datfmt		; get address (optional number base)
	ble	rm_syn		; bad character in value or no address given
	ldd	nbrhi		; entered value (12-bit, 15-bit if I-field <> 0)
	anda	#Wmask		; isolate 12-bit address
	std	*Temp2
	lda	nbrhi		; get hi-byte entered value again
	bsr	c8to09		; convert pdp8 to 6809 memory space
	beq	rm_syn		; - error
	std	*trcmem		; store address in 6809 memory space
	lda	#3
	jsr	comand		; find the access type (DATA/READ/WRITE/CHANGED)
	ble	rm_syn		; - premature end
	cmpa	#4		; "CHANGED" ?
	bne	20$
	ldx	#0xFFFF		;    y - start with impossible "previous" value
	stx	*trcinf		;	 so any change will trigger the trace
20$:	deca
	tfr	a,b		; save access type for "DATA" check
	lsla			; move access type bits to correct position
	lsla
	lsla
	ora	#TRCMEM		; trace memory contents
	ora	*trcsts		; store access type
	sta	*trcsts
	tstb			; is access type == DATA ?
	bne	40$		; n - (READ, WRITE, CHANGED) :: done
	bsr	datnot		; get value (optional "NOT" and number base)
	ble	rm_syn		; bad value or end of line
	ldd	nbrhi		; get the entered value
	std	*trctrg
	lda	*trcsts
40$:	andcc	#0xF9		; - OK, Z=0 / V=0
	rts



    ; c8to09 - convert pdp8 field and 12-bit address into 6809 real address
    ;
    ;  call   :	[A] = high byte of an entered address
    ;		      (high nibble contains field number [-fff])
    ;		<Temp2> = 12-bit address
    ;  return : ccZ = 1 -> not existent field (error)
    ;		ccZ = 0 -> [D] = address in 6809 memory space

c8to09:	anda	#0b11110000	; isolate I/D field bits
	bmi	20$		; ? ==> I/D field higher than 7 ?!
	lsra			; move "field" bits to low nibble
	lsra
	lsra
	ldx	#GetFld
	leax	a,x		; base address (6809 space) of I/D field
	ldd	,x		; does field exists ?
	beq	10$		; n - ccZ = 1
	ldd	*Temp2
	lslb			; translate pdp8 to 6809 memory space
	rola
	addd	,x		; add base address of specified I/D field
10$:	rts
20$:	clra
	rts




  ; ---------------------------------------------------------------
  ; SIMULATION CONTROL immediate response commands [LOAD - DISPLAY]
  ; ---------------------------------------------------------------

    ; LOAD INFORMATION

tr_lod:	lda	#2
	jsr	comand		; try search for a register name
	bmi	lodmem		; not a register
	beq	lodsyn		; end of line - some info must follow!
	deca
	asla			; register ID value * 2
	ldx	#lodreg
	ldx	a,x
	stx	*trcmem		; store load register action routine address
	bsr	datfmt		; get register value
	ble	lodsyn		; bad character in value, or no entry at all
	jsr	skpdlm
	bcc	lodsyn		; garbage follows -
	ldd	nbrhi		; get the entered value
	jsr	[trcmem]	; try to set the register
	bne	lodsyn		; illegal value for specified register
loddon:	jmp	tpromp
lodsyn:	jmp	synnok

lodmem:	jsr	datfmt		; get address (optional number base)
	ble	lodsyn		; bad character in value or no address given
	ldd	nbrhi		; entered value (12-bit, 15-bit if I-field <> 0)
	std	*trcinf
	anda	#Wmask		; isolate 12-bit address
	std	*Temp2
	lda	*trcinf		; get hi-byte entered value again
	bsr	c8to09		; convert pdp8 to 6809 memory space
	beq	lodsyn		; - error
	std	*trctrg		; store address in 6809 memory space
	clr	*trctmp		; assume single address: set flag to '0'
	lda	#10
	jsr	comand		; search for "THRU" keyword
	ble	40$		; - not "THRU"
;
	lda	#1
	sta	*trctmp		; - "THRU" -> set flag: load address block
	jsr	datfmt		; get load (end) address in any format base
	ble	lodsyn		; bad character in value or no address given
	ldd	nbrhi		; entered value (12-bit, 15-bit if I-field <> 0)
	std	*trcmem		; store entry last address
	cmpd	*trcinf		; is it higher than the first address?
	blo	lodsyn		; n -
	anda	#Wmask		; isolate 12-bit address
	std	*Temp2
	lda	*trcmem		; get hi-byte entered value again
	bsr	c8to09		; convert pdp8 to 6809 memory space
	beq	lodsyn		; - error
	std	*trcmem		; store address in 6809 memory space
;
40$:	jsr	datfmt		; get load value
	ble	lodsyn		; bad character in value, or no entry at all
	jsr	skpdlm
	bcc	lodsyn		; garbage follows -
	ldd	nbrhi		; get the entered value
	ldx	*trctrg		; pointer to (first) memory location
	tst	*trctmp		; single address or block of addresses ?
	bne	50$
	std	,x		; store value in specified address
	bra	loddon
50$:	std	,x++		; store value in specified address block
	cpx	*trcmem
	blo	50$
	bra	loddon


    ; DISPLAY INFORMATION

tr_dsp:	clr	*trctmp		; assume "DISPLAY REGISTERS"
	jsr	datfmt		; get display (start) address in any format base
	beq	32$		; no entry - EOL ?
	bmi	dsprf8		; bad character in value, is it "RF08" ?
	ldd	nbrhi		; entered value (12-bit, 15-bit if I-field <> 0)
	std	*trcinf		; store entry first address
	anda	#Wmask		; isolate 12-bit address
	std	*Temp2
	lda	*trcinf		; get hi-byte entered value again
	jsr	c8to09		; convert pdp8 to 6809 memory space
	beq	dspsyn		; - error
	std	*trctrg		; store address in 6809 memory space
	com	*trctmp		; assume EOL -> set flag: display single address
	lda	#10
	jsr	comand		; search for "THRU" keyword
	ble	20$		; - not "THRU"
;
	lda	#1
	sta	*trctmp		; - "THRU" -> set flag: display address list
	jsr	datfmt		; get display (end) address in any format base
	ble	dspsyn		; bad character in value or no address given
	ldd	nbrhi		; entered value (12-bit, 15-bit if I-field <> 0)
	std	*trcmem		; store entry last address
	cmpd	*trcinf		; is it higher than the first address?
	blo	dspsyn		; n -
	anda	#Wmask		; isolate 12-bit address
	std	*Temp2
	lda	*trcmem		; get hi-byte entered value again
	anda	#0b11110000	; isolate I/D field bits
	ldb	*trcinf		; get hi-byte entered start address again
	andb	#0b11110000	; isolate I/D field bits
	cba			; start and end address in same field ?
	bne	dspsyn
	lda	*trcmem
	jsr	c8to09		; convert pdp8 to 6809 memory space
	beq	dspsyn		; - error
	std	*trcmem		; store address in 6809 memory space
;
20$:	lda	#4		; look for optional display data format
	jsr	comand
	bmi	dspsyn		; something is there, but it is garbage
	beq	30$		; end of line - set default octal display output
	bpl	31$		; set specified base
30$:	lda	#3		; not recognized, must be value, set OCTAL base
31$:	sta	dbcode		; set base code to display the memory data
32$:	jsr	skpdlm		; check for garbage at the end
	bcs	dspcor
dspsyn:	jmp	synnok		; syntax error / garbage :: Z=1

dsprf8:	lda	#12
	jsr	comand		; check for "RF08"
	bmi	dspsyn		; nope -
	beq	dspsyn
	lda	#13		; ok - next is either "OFF" or "ON"
	jsr	comand
	bmi	dspsyn
	beq	dspsyn
	deca
	pshs	a		; save; OFF=0 / ON=1
	jsr	skpdlm		; garbage at the end?
	puls	a
	bcc	dspsyn		; y -
	sta	*rf8inf		; n - set RF08 "register" display as specified
	jmp	tpromp
	

  ; command validated
  ; data display output base is set
  ; <trctmp> == -1 -> display data of a single address <trcinf>
  ;		      <trctrg> == physical address <trcinf> in 6809 memory space
  ;	     ==  0 -> display registers/status of pdp8 simulator
  ;	     == +1 -> display data of an address list <trcinf> - <trcmem>
  ;		      <trctrg> == physical address <trcinf> in 6809 memory space
  ;		      note: <trcmem> is already in 6809 memory space

dspcor:
; ---	jsr	docrlf		; blank line for readability
	lda	*trctmp
	bmi	dsp1		; - single address
	bne	dsplst		; - address block
	jsr	Pntreg		; - registers and info
	bra	dspdon

  ; display contents of a single address
  ; print address in octal followed by real 6809 address in brackets and then
  ; the data of that address word in the requested format (hex, oct, dec, bin)

dsp1:	lda	dbcode		; save requested display format
	pshs	a
	lda	#3		; force address display in octal
	sta	dbcode
	ldd	*trcinf
	anda	#Wmask		; mask (optional) fieldnumber bits
	std	*Temp2
	ldx	#Temp2		; print address in field itself in octal
	bsr	out2b
	ldx	#trcms2
	jsr	pdatao		; print "in field number"
	lda	*trcinf
	jsr	OUTHLO		; print upper 4 bits hex (should be 0-7 max)
; --	ldx	#trcms3
	jsr	pdatao		; print " ["

	lda	#1
	sta	dbcode		; force output in hex
	ldx	#trctrg
	bsr	out2b		; display real (6809) address
	ldx	#trcms4
	jsr	pdatao		; print "] = "
	puls	a
	sta	dbcode		; restore requested display format
	ldx	*trctrg
	bsr	out2b		; and print the memory contents
dspdon:	bsr	crlf		; extra blank line for readability
	jmp	tpromp		; request new simulation command


  ; display contents of an address list
  ; begin address = <trcinf> in pdp8 space (and in <trctrg> in 6809 memory space)
  ; end   address = <trcmem> in 6809 memory space

dsplst:	ldb	#0b00000111	; assume octal mode, set mask = low 3 bits
	lda	dbcode		; check the requested data display format
	cmpa	#3		; octal - output nice header
	beq	10$
	cmpa	#1
	bne	dspdb
	ldb	#0b00001111	; nope, it is hex, set mask = low 4 bits
10$:	stb	*Temp1		; set mask
	jsr	prthdr		; print the correct header lines
	jsr	prtstl		; print start of (new) data dump line

   ; print 1st dump line, move to start position (offset) given in start address

	lda	*trcinf+1
	anda	*Temp1		; get offset in 8/16 word block
	beq	20$		; no offset, just start of line
	ldb	*trctmp+1	; get # spaces between data fields
	incb			; correct the (-1)
	mul			; number of spaces to offset position on screen
	lda	#Space
	bsr	outabt		; move to correct data field position

    ; print the data in specified format

20$:	bsr	outspc		; hex output: 1 leading space
	ldb	*Temp1
	cmpb	#0b00000111
	bne	21$
	bsr	outspc		; octal output: 2 leading spaces, print 1 more
21$:	bsr	prtwrd		; print 1 data word
	beq	dspdon		; - all data printed
	andb	*Temp1
	bne	20$
	bsr	crlf		; y - indeed
	bsr	prtstl
	bra	20$

out2b:	jmp	out2by


    ; output format is either decimal or binary (in dbcode)
    ; dump start address (of this line) and 10 (dec) or 4 (bin) locations
    ; begin address = <trcinf> in pdp8 space (and in <trctrg> in 6809 memory space)
    ; end   address = <trcmem> in 6809 memory space

dspdb:	ldb	#10		; assume DEC format
	lda	dbcode
	pshs	a		; save requested data format
	cmpa	#2
	beq	101$
	ldb	#4		; nope, it is BIN !
101$:	stb	*trctmp		; output count per line
	lda	#3
	sta	dbcode
	ldx	#trcinf
	bsr	out2b		; print address itself in octal
	ldx	#dotdot
	jsr	pdatao		; print separator " : "
	puls	a
	sta	dbcode		; set requested data display format
110$:	bsr	prtwrd		; print data word
	beq	dspdon		; - all data printed
	dec	*trctmp		; check if new line is needed
	bne	120$
	bsr	crlf		; y - indeed
	bra	dspdb
120$:	bsr	outspc		; n - move to next, print space
	bra	110$

crlf:	jmp	docrlf



    ; prtwrd - print a word of data
    ;
    ;   call : <trctrg> points to data word location
    ;	       <trcmem> points to last location
    ;			(used for end-of-print condition test)	
    ; return : <trcinf> incremented by 1 (next address in pdp8 memory space)
    ;	       <trctrg> incremented by 2 (next address in 6809 memory space)
    ;	       ccZ = 1 : all locations printed (trctrg == trcmem)
    ;	       ccZ = 0 : ok, [D] = <trcinf>, updated pdp8 address

prtwrd:	ldx	*trctrg		; point to data
	bsr	out2b		; print it
	ldd	*trctrg		; all requested data printed ?
	cmpd	*trcmem
	beq	prtrts		; y -
	addd	#2		; n - move to next (6809) address
	std	*trctrg
	ldd	*trcinf		; check if new line is needed
	addd	#1
	std	*trcinf
	andcc	#0xFB		; ccZ := 1
prtrts:	rts

outspc:	jmp	OUTS
outabt:	jmp	OUTABT



    ; prtstl - print start of data dump line
    ;
    ;   call : <Temp1> = display mode mask (0F=hex, 07=octal)
    ;	       <trcinf> = memory address
    ; return : - for octal mode, base address of 8-word block is printed
    ;	       - for hex mode, there is no base address (no room for on screen)

prtstl:	lda	*Temp1
	cmpa	#0b00000111
	bne	prtrts		; hex - done
	ldd	*trcinf		; oct - form base address of 8-word block
	andb	#0b11111000
	std	*Temp2
	ldx	#Temp2
	bsr	out2b		; print base address
	bsr	outspc		; and 2 spaces
	bra	outspc



    ; prthdr - print header
    ;
    ;   call : [A] = display mode (1=hex, 2=dec, 3=octal, 4=bin)
    ; return : header, separator line and newline printed
    ;		- for octal mode, header is 0 thru 7
    ;		  each output line starts with the base address
    ;		- for hex mode, header is 0 thru F
    ;	       <trctmp+1> == # spaces between data fields (-1)
    ;	       [X] and <Temp2> destroyed

prthdr:	ldx	#dashes-1	; set pointer to # dashes for header line
	leax	a,x
	clr	*trctmp		; set last digit (for header) to 0
	cmpa	#1
	beq	10$		; - hex output: no address (not enough room)
	lda	#Space
	ldb	#8		; - octal
	bsr	outabt		;   skip address field
	lda	#7		;   set # spaces between octal data fields
	ldb	#8		;   set highest (+1) offset digit
	bra	20$
10$:	lda	#4		; - hex set # spaces between hex data fields
	ldb	#16		;	set highest (+1) offset digit
20$:	stb	*Temp2		; store highest (+1) offset digit
	sta	*trctmp+1	; store data field skip space count
30$:	lda	#Space
	ldb	*trctmp+1
	bsr	outabt		; skip data field
	lda	*trctmp
	jsr	OUTHRO		; print offset digit in 8/16 word block
	lda	*trctmp
	inca
	sta	*trctmp
	cmpa	*Temp2		; printed all 7/15 offset digits ?
	bne	30$
	bsr	crlf
	ldb	,x		; print separator line
	lda	#'-
	bsr	outabt
	bra	crlf		; new line and RTS

dashes:	.fcb	80		; hex output header line # dashes
	.fcb	68		; dec
	.fcb	72		; oct
	.fcb	75		; bin




    ; getopt - process one option from the command input line
    ;
    ; call   : [X] : ptr to command line
    ; return : ccZ = 1 --> end of command line or syntax error (abort)
    ;	       ccZ = 0 --> command token processed
    ;			   <verbos>, <core>, <trcflg>, <dsknum> or <boottb> set
    ;			   [X] : updated to next character

getopt:	lda	,x+		; get next character
	cmpa	#CR		; no arguments ?
	beq	10$		; y -
	cmpa	#Space		; n - a space must follow
	bne	9$
	lda	,x+		; get token indicator
	cmpa	#'-		; must be minus sign
	bne	9$
	lda	,x+		; get flag character
	anda	#0x5F		; force uppercase
	cmpa	#'V
	beq	20$		; -> verbose
	cmpa	#'C
	beq	30$		; -> CORE memory
	cmpa	#'T
	beq	35$		; -> TRACE enable
	cmpa	#'D
	beq	40$		; -> disk number
	cmpa	#'B
	beq	50$		; -> OS bootstrap
9$:	clra			; ?? -> ccZ := 1
10$:	rts
				;
20$:	inc	*verbos		; turn "verbose" on
	bra	99$
30$:	inc	*core		; non-zero: CORE "installed"
	bra	99$
35$:	inc	*trcflg		; enable "trace" on
	bra	99$

40$:	lda	,x+		; get "disk" number
	cmpa	#'0
	blo	9$		; < 0
	cmpa	#'9
	bhi	9$		; > 9
	suba	#'0		; ASCII -> binary [0-9]
	sta	*dsknum
	ldb	#<Sec4dsk	; get LSB multiplicant
	mul			; multiply LSB's
	std	*secoff+2
	lda	*dsknum		; get multiplier
	ldb	#>Sec4dsk	; get MSB multiplicant
	mul			; multiply MSB's
	addb	*secoff+2	; add LSB's to MSB's of previous partial product
	adca	#0		; add carry to MSB's
	std	*secoff+1
	clr	*secoff		; 8bit * 16bit = 24bit result, bits [24:31] always 0
	bra	99$
50$:	tfr	x,y		; save command line ptr
	ldx	#OStab		; - ptr to lookup table
51$:	ldb	#3		; - # characters to match
	tfr	y,u		; - ptr to command line 'xxx'
52$:	lda	,u+
	cmpa	,x+		; character match ?
	bne	54$
	decb			; y - all characters checked ?
	bne	52$		;     n -
	ldd	,x		;     y - set ptr to bootcode table
	std	*boottb
	tfr	u,x		;	  update command line ptr
	bra	99$
54$:	incb			; n - to next lookup entry
	abx
	tst	,x		;     end of lookup table ?
	bne	51$		;     n -
	bra	9$		;     y - abort: OS not recognized
				;
99$:	andcc	#0xFB		; all OK -> ccZ := 0
	rts


OStab:	.fcc	"DMS"
	.fdb	btDMSt		; bootstrap code for DMS
	.fcc	"OS8"
	.fdb	btOS8t		; bootstrap code for OS/8
	.fcb	0



    ; trcchk - check specified trace condition
    ;
    ;	call   : [X] - simulator environment! must be restored at exit
    ;	return : ccZ = 1 --> specified trace condition == FALSE
    ;		 ccZ = 0 --> specified trace condition == TRUE
    ;
    ;	There are a few special cases, easy to detect from <trcsts>.
    ;
    ;	 1. new request	 --> TRCNEW (TRCENA is clear)
    ;	 2. REALTIME	 --> TRCENA+TRC_RT
    ;	 3. TRACE ALWAYS --> TRCENA+TRCACT+TRC_AL
    ;	 4. STEP COUNT	 --> TRCENA+TRCACT+TRC_ST  *and*  <trccnt> =/= $0000

trcchk:	lda	*trcsts
	bita	#TRCENA		; is trace facility enabled ?
	bne	10$		;
	cmpa	#TRCNEW		; n - new trace request wanted ?
	bne	15$
	pshs	x		;     y - sollicit for (new) trace specification
	jsr	docrlf		;	  first blank line for readability
	jsr	trprmp
	puls	x
	bra	trcchk		;	  and do a trace condition check now !

    ; check the simple cases first :: "REALTIME", "TRACE ALWAYS" and "STEP COUNT"

10$:	cmpa	#TRCENA+TRC_RT		; # go "REALTIME" ?
	bne	20$
	clr	*trcflg			;   disable trace facility
	clr	*trcsts			;   CONT from console required to run
15$:	clra
	rts
20$:	cmpa	#TRCENA+TRCACT+TRC_AL	; # "TRACE ALWAYS" ?
	beq	31$
	cmpa	#TRCENA+TRCACT+TRC_ST	; # "STEP" ?
	bne	40$
	ldd	*trccnt			;   y - check if counter <> 0000
	beq	40$			;	== 0  ==> *not* "STEP COUNT" !
	subd	#1			;	<> 0  ==> decrement counter
	std	*trccnt
	beq	trcc_s			;	   now 0  ==> stop and sollicit
31$:	lda	#1			;   set return condition = TRUE
	rts

    ; check the specified trace condition
    ; the result is either "condition met" or "condition not met"
    ; the returned boolean (indicates whether the register dump must be printed)
    ; depends on the condition result AND the specified trigger command
    ;  - TRACE [AFTER, UNTIL, ONLY, BEFORE]
    ;  - RUN UNTIL
    ;  - STEP UNTIL

40$:	jsr	trccdx
	bne	trcc_T

    ; check needed response based on trace condition is FALSE

trcc_F:	lda	*trcsts
	anda	#TRCACT		; is trace _already_ active ?
	rts			; return TRUE / FALSE

    ; check needed response based on trace condition is TRUE
    ; and take "OCCUR" counter into account

trcc_T:	lda	*trcsts
	tfr	a,b
	anda	#TRCTYP		; isolate trace specifier (AFTER, UNTIL, ONLY)
	cmpa	#TRC_AF
	beq	trcc_a		; -> AFTER
	cmpa	#TRC_UN
	beq	trcc_u		; -> UNTIL
	cmpa	#TRC_BF
	beq	trcc_b		; -> BEFORE
	cmpa	#TRC_OY
	beq	trcc_o		; -> ONLY
	cmpa	#TRC_RN
	beq	trcc_r		; -> RUN UNTIL
	cmpa	#TRC_ST
	beq	trcc_s		; -> STEP UNTIL
	clra			; trace specifier unknown: condition always FALSE
	rts			; if you get here, there is actually a BUG !

  ; when the TRACE AFTER condition is met, you want to follow the execution of
  ; pdp8 instructions, so after the TRUE condition the simulator keeps running

trcc_a:	bitb	#TRCACT		; is TRACE AFTER _already_ active ?
	bne	20$		; y - return TRUE
	bsr	chkocc		; n - check OCCURrence counter specified
	bvs	10$		;     OCCUR not specified -
	bne	30$		;     OCCUR count not expired -
10$:	orb	#TRCACT		;     - set trace is active from now on
	stb	*trcsts
20$:	rts
30$:	clra
	rts

				; process "ONLY" condition == TRUE
trcc_o:	bsr	chkocc		; check OCCURrence counter specified
	bvs	10$		; OCCUR not specified -
	beq	10$		; OCCUR count not expired -
	clra
	rts
10$:	lda	#1		; return TRUE
	rts


  ; when the TRACE UNTIL or BEFORE condition is met, you want to examine what has
  ; happened up till now, so after the TRUE condition the simulator stops running

trcc_b:	bitb	#TRCACT		; is TRACE BEFORE _still_ active ?
	beq	30$		; n - return FALSE
	bsr	chkocc		; y - check OCCURrence counter specified
	bvs	10$		;     OCCUR not specified -
	bne	20$		;     OCCUR count not expired -
10$:	clr	*trcdmp		;     disable data dump re-routing
	jsr	dispdi		;     and display "trace BEFORE" stored data
	bra	trcc_s		;     set trace stopped from now on
20$:	clra
30$:	rts

trcc_u:	bitb	#TRCACT		; is TRACE UNTIL _still_ active ?
	beq	10$		; n - return FALSE
	bsr	chkocc		; y - check OCCURrence counter specified
	bvs	trcc_s		;     OCCUR not specified -
	beq	trcc_s		;     OCCUR count expired -
	clra
10$:	rts

  ; when the STEP UNTIL or RUN UNTIL condition is met, the simulator must stop
  ; and a request must be issued for a new emulation command
  ; Note. "RUN UNTIL" also supports the "OCCUR" keyword

				; process "RUN UNTIL" condition == TRUE
trcc_r:	bsr	chkocc		; check OCCURrence counter specified
	bvs	trcc_s		; OCCUR not specified -
	beq	trcc_s		; OCCUR count expired -
	clra
	rts

trcc_s:	lda	*conact		; process "STEP UNTIL" condition == TRUE
	anda	#255-ACTrun	; terminate RUN or STEP state
	sta	*conact
	lda	#TRCNEW		; request for new trace specification
	sta	*trcsts		; (clears [TRCACT] bit)
	lda	#1		; return TRUE
	rts



    ; chkocc - check occurrence event specified and occurrence counter
    ;
    ; call   : -
    ; return : ccV = 1 --> OCCUR event not specified
    ;	       ccV = 0 - ccZ = 0 --> OCCUR count not expired
    ;			 ccZ = 1 --> OCCUR count is expired

chkocc:	pshs	d
	lda	*trcext
	bita	#TREOCC		; OCCUR event specified?
	bne	10$
	orcc	#2		; n - ccV := 1
	puls	d,pc
10$:	ldd	*trcocc		; y - check/update count
	beq	20$		;     - already expired
	subd	#1		;     - update
	std	*trcocc
	andcc	#0xF9		; not (yet) expired: ccV :=0 / ccZ := 0
	puls	d,pc
20$:	andcc	#0xFD		; expired: ccV := 0 / ccZ := 1
	puls	d,pc



    ; trccdx - check trace condition
    ;
    ; call   : [X] - simulator environment! must be restored at exit
    ; return : ccZ = 1 --> trace condition == FALSE
    ;	       ccZ = 0 --> trace condition == TRUE
    ;
    ; trick  : check the trace spec (AFTER, UNTIL, ONLY) first in combination
    ;	       with the TRCACT flag. The logical output combination can make
    ;	       the actual check not needed.

trccdx:	lda	*trcsts
	tfr	a,b		; keep a copy of the trace status byte
	anda	#TRCTYP		; isolate trace specifier (AFTER, UNTIL, ONLY)
	cmpa	#TRC_OY
	beq	40$		; "TRACE ONLY" :: check always needed
	cmpa	#TRC_RN
	beq	40$		; "RUN UNTIL"  :: check always needed
	cmpa	#TRC_ST
	bne	10$		; "STEP ?????" :: check if trace_count == 0000
	ldd	*trccnt
	bne	30$		;		  n - "STEP COUNT"
	ldb	*trcsts
	bra	40$		;		  y - "STEP UNTIL" -> check needed
10$:	cmpa	#TRC_AF
	bne	20$
	bitb	#TRCACT		; "TRACE AFTER" : check if trace already active
	beq	40$		;	not yet active :: check needed
	bra	30$
20$:	cmpa	#TRC_UN
	beq	22$
	cmpa	#TRC_BF		; "TRACE BEFORE"
	bne	30$		;   and
22$:	bitb	#TRCACT		; "TRACE UNTIL" : check if trace still active
	bne	40$		;	still active :: check needed
30$:	clra			; => always FALSE condition
	rts

40$:	pshs	x		; check specified condition
	bitb	#TRCMEM		; access register or memory ?
	bne	t_mem		; -> memory

    ;	____________________________________
    ;	|				   |
    ;	|  condition target is a register  |
    ;   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    ;
    ;  possible target conditions:
    ;	 - [register] CHANGED
    ;	 - [register] <value>

t_reg:	ldb	*trcmem		; get register ID
	ldx	#t_regs
	abx
	clra			; init content value = 0000
	clrb			; (make sure non-used bits are '0')
	jsr	[,x]		; get the current register contents in [D]
	std	*trctmp
	ldb	*trcsts		; trace specifier == CHANGED ?
	andb	#TRCACC
	cmpb	#TRCCHG
	bne	t_reg1
	ldd	*trctmp		; y - compare with previous value
chgchk:	cmpd	*trcinf
	beq	dat_ff		;     no change => trace condition FALSE
	std	*trcinf		;     changed -> update and return TRUE
	bra	dat_tt
t_reg1:	ldd	*trctmp		; n - trace specifier is a specific value

    ; include "NOT" keyword  ( trcext[TREIDC] )

datchk:	cmpd	*trctrg
	beq	dat_t
	lda	*trcext		; no data match, invert to TRUE ?
dat_f:	bmi	dat_tt		; y -
dat_ff:	clra			; return condition == FALSE
	puls	x,pc
dat_t:	lda	*trcext		; data match, invert to FALSE ?
	bmi	dat_ff		; y -
dat_tt:	lda	#1		; return condition == TRUE
	puls	x,pc


    ;	___________________________________________
    ;	|					  |
    ;	|  condition target is a memory location  |
    ;   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    ;
    ;  possible target conditions:
    ;	 - <address> DATA <value>
    ;	 - <address> CHANGED | READ | WRITE
    ;
    ;	 Note 1. memory address is already translated into 6809 memory space
    ;	 Note 2. pdp8 address is 12 bit, 3 leading bits for I/D field is included

t_mem:	lda	*trcsts		; access specifier == "CHANGED" ?
	anda	#TRCACC
	cmpa	#TRCCHG
	bne	10$
	ldd	[trcmem]	; y - compare memory with previous value
	bra	chgchk

10$:	cmpa	#TRCDAT		; access specifier == "DATA" ?
	bne	20$
	ldd	[trcmem]	; y - compare memory with specified value
	bra	datchk

    ; access specifier == "READ" or "WRITE"
    ;	-> requires examination of the *-JUST-* executed pdp8 instruction!
    ;	   note. As the trace facility is executed AFTER the pdp8 instruction is
    ;		 executed, there are some limitations, for example auto-index fails.
    ;		 Further, IOT and OPR instructions are not monitored at all.
    ;
    ;       11  10  9   8   7   6   5   4   3   2   1   0
    ;       ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ 
    ;      |   |   |   |   |   |   |   |   |   |   |   |   |
    ;      |_ _|__ |__ |__ |__ |___|___|__ |__ |__ |__ |__ |
    ;      :           :   :   :                           :
    ;      :  opcode   : i : z :           address         :
    ;
    ; notes.
    ;        address - 12-bit word in current page
    ;           i    - indirect bit  [0=direct, 1=indirect]
    ;           z    - page bit      [0=page zero, 1=current page]
    ;
    ;     000 = AND  - logical AND of AC with memory
    ;     001 = TAD  - 2-complement add of memory to <L,AC> (13-bit value)
    ;     010 = ISZ  - increment memory and skip next instruction if result is 0
    ;     011 = DCA  - deposit AC in memory and clear AC
    ;     100 = JMS  - jump to subroutine. PC stored in subroutine start address.
    ;     101 = JMP  - jump unconditional
    ;     110 = IOT  - input/output transfer
    ;     111 = OPR  - microcoded operations, 3 groups
    ;
    ;
    ; * READ: check the instruction field of the following pdp8 instructions
    ;	  - AND (direct address or indirect address)
    ;	  - TAD (direct address OR indirect address)
    ;	  - JMS (indirect address only)
    ;	  - JMP (indirect address only)
    ;
    ; * WRITE: check the instruction field of the following pdp8 instructions
    ;	  - ISZ (direct address or indirect address)
    ;	  - DCA (direct address or indirect address)
    ;	  - JMS (direct address or indirect address)

20$:	ldb	*MD		; get the just executed instruction
	andb	#Opmask		; isolate opcode
	ldx	#ins_rd		; assume READ
	cmpa	#TRCRD		; trace specifier == READ ?
	beq	21$		; y - got the correct lookup table
	ldx	#ins_wr		; n - get WRITE lookup table
21$:	abx
	ldd	*MD		; get the instruction
	jsr	[,x]		; get effective address from instruction field
	bcs	30$		; instruction not relevant for Read Access check
	cmpd	*trcmem		; effective address == specified address ?
	beq	dat_tt		; y - match !
30$:	clra			; n - => condition result is FALSE
	puls	x,pc



t_regs:	.fdb	200$		; 0 - PC
	.fdb	210$		; 1 - AC
	.fdb	220$		; 2 - MQ
	.fdb	230$		; 3 - instruction word
	.fdb	240$		; 4 - IF
	.fdb	250$		; 5 - DF
	.fdb	260$		; 6 - IOflgs (interrupt mask)
	.fdb	270$		; 7 - Link bit
	.fdb	280$		; 8 - User bit

200$:	ldd	*PC
	rts
210$:	ldd	*AC
	rts
220$:	ldd	*MQ
	rts
230$:	ldd	*MD
	rts
240$:	ldb	*WorkUI
	andb	#Fbits
	rts
250$:	ldb	*WorkD
	andb	#Dbits
	rts
260$:	ldd	*IOflgs
	rts
270$:	ldb	*Link
	andb	#Lmask
	beq	272$
271$:	ldb	#1
	rts
272$:	clrb
	rts
280$:	ldb	*WorkUI
	bmi	271$
	clrb
	rts



    ; load register routines
    ;   call  : [D] = entered value
    ;   return: [ccZ] = 1 --> OK, value stored in register
    ;			0 --> illegal value

lodreg:	.fdb	200$		; 0 - PC
	.fdb	210$		; 1 - AC
	.fdb	220$		; 2 - MQ
	.fdb	230$		; 3 - instruction word
	.fdb	240$		; 4 - IF
	.fdb	250$		; 5 - DF
	.fdb	260$		; 6 - IOflgs (interrupt mask)
	.fdb	270$		; 7 - Link bit
	.fdb	280$		; 8 - User bit

200$:	bita	#0b11110000
	bne	201$
	std	*PC
	clra
201$:	rts

210$:	bita	#0b11110000
	bne	211$
	std	*AC
	clra
211$:	rts

220$:	bita	#0b11110000
	bne	221$
	std	*MQ
	clra
221$:	rts

230$:	bita	#0b11110000
	bne	231$
	std	*MD
	clra
231$:	rts

240$:	tsta
	bne	241$
	bitb	#0b11111000
	bne	241$
	lda	*WorkUI
	anda	#0b11111000	; preserve U bit
	aba
	sta	*WorkUI
	clra
241$:	rts

250$:	tsta
	bne	251$
	bitb	#0b11111000
	bne	241$
	stb	*WorkD
	clra
251$:	rts

260$:	std	*IOflgs		; NO CHECK !
	clra
	rts

270$:	tsta
	bne	271$
	bitb	#0b11111110
	bne	271$
	lda	#Lmask		; assume "1"
	asrb
	bcs	27$
	clra
27$:	sta	*Link
	clra
271$:	rts

280$:	tsta
	bne	281$
	bitb	#0b11111110
	bne	281$
	lda	*WorkUI
	ora	#Ubit		; assume "1" : set U bit
	asrb
	bcs	28$
	anda	#255-Ubit	; nope it's "0" : clear U bit
28$:	sta	*WorkUI
	clra
281$:	rts



ins_rd:	.fdb	oprrd0		; AND
	.fdb	oprrd0		; TAD
	.fdb	no_cdx		; ISZ
	.fdb	no_cdx		; DCA
	.fdb	oprrd1		; JMS (check for indirect mode)
	.fdb	oprrd1		; JMP (check for indirect mode)
	.fdb	no_cdx		; IOT
	.fdb	no_cdx		; OPR

ins_wr:	.fdb	no_cdx		; AND
	.fdb	no_cdx		; TAD
	.fdb	opr_wr		; ISZ
	.fdb	opr_wr		; DCA
	.fdb	opr_wr		; JMS
	.fdb	no_cdx		; JMP
	.fdb	no_cdx		; IOT
	.fdb	no_cdx		; OPR




    ; Memory Access effective address calculation routines
    ;
    ;   call   : [D] = instruction
    ;   return : [ccC] = 1 -> not relevant (trace condition always FALSE)
    ;		 [ccC] = 0 -> [D] = effective 15-bit address
    ;				    (I-field or D-field taken into account)

opr_wr:	
oprrd0:	bita	#1		; check Indirect bit of AND or TAD instruction
	bne	oprind
	bsr	eadir0		; get inst. referenced address in 6809 space
	ldb	*WorkUI		; now include ** Instruction ** field
	bsr	basfld		; get base address of field in 6809 memory space
	beq	no_cdx		; - not allocated
	addd	*Temp2		; referenced direct address (6809 space)
	andcc	#255-1
	rts

oprrd1:	bita	#1		; check Indirect bit of JMP or JMS instruction
	bne	oprind		; ok, indirect mode
no_cdx:	orcc	#1		; not indirect mode -> address not relevant
	rts

oprind:	bsr	eadir0		; get instr. referenced address in 6809 space
	ldb	*WorkD		; now include ** Data ** field
	bsr	basfld		; get base address of field in 6809 memory space
	beq	no_cdx		; - not allocated
	pshs	d		; save 6809 start address of this field
	addd	*Temp2		; indirection address
	tfr	d,x
	ldd	,x		; referenced indirect address (pdp8 space!)
	lslb
	rola			; referenced indirect address (6809 space)
	addb	1,s		; add field base in 6809 space
	adca	,s++
	andcc	#255-1
	rts



    ; basfld - get base address in 6809 memory space of the specified field number
    ;
    ;   call   : [B] = bits [2:0] is field number (from WorkUI or WorkD)
    ;   return : [ccZ] = 1 -> requested field does not exist
    ;		 [ccC] = 0 -> [D] = base address of field in 6809 memory space

basfld:	lslb			; field number x2 for lookup table index
	ldx	#GetFld		; base of allocated fields table
	ldd	b,x		; get base of specified field in 6809 memory space
	rts



    ; eadir0 - get instruction referenced address (field 0)  in 6809 space
    ;
    ;   call   : [D] = instruction
    ;   return : [D] = instruction field address in 6809 memory space
    ;			(I-field or D-field NOT taken into account)
    ;		<Temp2> == [D]

eadir0:	clra			; calculate effective address
	tstb			; in zero page ?
	bpl	10$		; y - [D] is base address
	andb	#Offset		; n - strip current_page bit
	lda	*PC
10$:	stb	*Temp1		; save address in page
	ldb	*PC+1		; get current page #
	andb	#LoPage
	orb	*Temp1		; add instruction address field
	lslb			; translate to 6809 memory space
	rola
	std	*Temp2		; address in field 0
	rts



; =============================================================================
;
;  TRACE BEFORE requires that the data *before* the (condition == TRUE) is met,
;  is remembered. Due to limited storage space, not everything can be stored.
;  "stordi" is called when a character must be stored (to be printed later).
;  "stordi" maintains a cyclic buffer that holds the most recent data.
;  When the (condition == TRUE) is met for the first time, the saved trace data
;  is displayed on the screen with the "dispdi" routine. The oldest stored data
;  is of course presented first, the most recent data information last.
;
;  Note. The register dump is as follows:  <text_part1><CR><text_part2>
;	 <text_part1> is from the previous instruction execution, <text_part2>
;	 is from the next instruction execution. When the buffer overruns, it
;	 is correct to only move the output pointer to the <CR> between the two
;	 text string sections. When the information is displayed the dump starts
;	 with the <text_part2> of the oldest stored register dump, which is OK !

   ; stordi - store register dump data information
   ;
   ;   call   : [A] = character to store in  buffer
   ;		buffer pointers <trcbip/trcbop> are initialised
   ;   return : character stored in buffer
   ;		buffer adminstration updated
   ;		oldest complete line removed if buffer is full, LF stripped
   ;   note   :	preserve [X] register!
   ;
   ;   pseudo code ::
   ;		if (char == LF)  exit;		/* LF not stored */
   ;		buf[ip] := char;
   ;		update ip;			/* cyclic buffer */
   ;		if (ip == op) then
   ;		   update op;			/* OP to nxt msg */
   ;		   while (buf[op] <> CR)
   ;			update op;
   ;		   update op;
   ;		else
   ;		   exit;

stordi:	pshs	x
	cmpa	#LF		; LF character ?
	beq	stor_x		; y - not stored
	ldx	*trcbip
	sta	,x		; store char in buffer
	ldd	*trcbip
	bsr	bump_p		; update pointer in cyclic buffer
	std	*trcbip
	cmpd	*trcbop		; INptr == OUTptr ?
	bne	stor_x		; n - all done
10$:	ldd	*trcbop		; y - buffer overrun -> move OUTptr
	bsr	bump_p		;     to the start of the next message
	std	*trcbop		;     effectively removing the oldest message
	ldx	*trcbop
	lda	,x		;     is OUTptr on CR character of oldest msg ?
	cmpa	#CR
	bne	10$		;     n - keep looking for CR
	ldd	*trcbop		;     y - set OUTptr to start of next message
	bsr	bump_p
	std	*trcbop
stor_x:	puls	x,pc



   ; bump_p - bump pointer +1 , maintaining a cyclic buffer
   ;
   ;   call   : [D] - pointer
   ;   return : [D] - updated pointer

bump_p:	addd	#1
	cmpd	#TRCbhi		; is pointer past highest buffer address ?
	bls	10$
	ldd	#TRCblo		; y - set back to start of buffer
10$:	rts




   ; dispdi - display stored register dump data information
   ;
   ;   call   : -
   ;   return : all complete lines in buffer displayed (LIFO)
   ;   note   :	preserve [X] register!
   ;
   ;   pseudo code ::
   ;		print newline;
   ;		while (op <> ip)
   ;		   char = buf[op];
   ;		   display (char);
   ;		   if (char == CR)  display (LF);
   ;		   update op;
   ;		exit;

dispdi:	pshs	x
	jsr	docrlf		; blank line for readability
10$:	ldx	*trcbop
	cpx	*trcbip		; OUTptr == INptr ?
	beq	stor_x		; y - buffer empty, or all printed
	lda	,x		; n - get char from buffer
	jsr	outch		;     print char
	cmpa	#CR		;     was char CR ?
	bne	20$		;     n -
	lda	#LF		;     y - print LF too
	jsr	outch
20$:	ldd	*trcbop
	bsr	bump_p		;     update OUTptr
	std	*trcbop
	bra	10$



  ; rf8reg - display RF08 "registers"
  ;
  ;	when the emulation command "DISPLAY RF08 ON" is specified, every physical
  ;	data transfer (that is DMAR and DMAW) from/to the RF08 is printed before
  ;	the actual data transfer starts.

rf8reg:	pshs	x,y
	ldb	#3
	stb	dbcode		; make sure output is in octal format
	ldy	#RFregs		; point to format output
	ldb	cplcnt		; check char/line count: output at start of line?
	bne	1$		; n - start with a CR,LF
	leay	2,y		; y - skip CR,LF
1$:	bsr	pdatay		; "DMA"
	jsr	outchr		; 'R/W'
	ldb	#5		; print 5 data items
2$:	bsr	pdatay		; print text id of "register"
	ldx	,y++		; get address of variable
	jsr	[,y++]		; call correct print routine
	decb
	bne	2$
	bsr	pdatay		; end with new line
	puls	x,y,pc


pdatay:	pshs	a
	lda	,y+	; get char
	beq	2$	; done -
1$:	jsr	outchr
	lda	,y+
	bne	1$
2$:	puls	a,pc


RFregs:	.fcb	CR,LF
	.asciz	"  DMA"
	.asciz	"  wc="
	.fdb	wc3cdb
	.fdb	out2by
	.asciz	" ca="
	.fdb	ca3cdb
	.fdb	out2by
	.asciz	" df_dma="
	.fdb	df_dma
	.fdb	out1by
	.asciz	" df_dmx="
	.fdb	df_dmx
	.fdb	out2by
	.asciz	" df_dma="
	.fdb	df_dma
	.fdb	out2by
	.fcb	CR,LF,0




  ;  trigger specification request/response messages and command lookup tables
  ;  ** keep trcms2-trcms3-trcms4 in this order **

trcms1:	.asciz	"  command: "
trcms2:	.asciz	" in field number "
trcms3:	.asciz	" ["
trcms4:	.asciz	"] = "



  ; simulation control commands  //  trace facility trigger specification syntax
  ;
  ; TRACE  ALWAYS  {cr}
  ;
  ;                                 --                       --
  ;        --      --               |        --   --          |
  ;        | AFTER  |               |        | OCT |          |
  ; TRACE  | UNTIL  |  <condition>  | OCCUR  | HEX |  <count> |  {cr}
  ;        | BEFORE |               |        | DEC |          |
  ;        | ONLY   |               |        | BIN |          |
  ;        --      --               |        --   --          |
  ;                                 --                       --
  ;
  ;             --   --
  ;             | OCT |
  ; STEP COUNT  | HEX |  <count_value>  {cr}
  ;             | DEC |
  ;             | BIN |
  ;             --   --
  ;
  ; STEP UNTIL <condition>  {cr}
  ;
  ;                         --                       --
  ;                         |        --   --          |
  ;                         |        | OCT |          |
  ; RUN  UNTIL <condition>  | OCCUR  | HEX |  <count> |  {cr}
  ;                         |        | DEC |          |
  ;                         |        | BIN |          |
  ;                         |        --   --          |
  ;                         --                       --
  ;
  ; REALTIME  {cr}
  ;
  ;         --                                                          --
  ;         |                     --                        --           |
  ;         | --   --             |      --   --             |   --   -- |
  ;         | | OCT |             |      | OCT |             |   | OCT | |
  ; DISPLAY | | HEX | <address_1> | THRU | HEX | <address_2> |   | HEX | |  {cr}
  ;         | | DEC |             |      | DEC |             |   | DEC | |
  ;         | | BIN |             |      | BIN |             |   | BIN | |
  ;         | --   --             |      --   --             |   --   -- |
  ;         |                     --                        --           |
  ;         --                                                          --
  ;
  ;              --   --
  ;              | OFF |
  ; DISPLAY RF08 | ON  |  {cr}
  ;              --   --
  ; 
  ;
  ;      --                                                --
  ;      | .REGISTER_ID                                     |
  ;      |                     --                        -- |
  ;      | --   --             |      --   --             | |  --   --
  ;      | | OCT |             |      | OCT |             | |  | OCT |
  ; load | | HEX | <address_1> | THRU | HEX | <address_2> | |  | HEX | <value> {cr}
  ;      | | DEC |             |      | DEC |             | |  | DEC |
  ;      | | BIN |             |      | BIN |             | |  | BIN |
  ;      | --   --             |      --   --             | |  --   --
  ;      |                     --                        -- |
  ;      --                                                --
  ;
  ; -------------------------------  <condition>  -------------------------------
  ;
  ;  (1) - register
  ;
  ;                       --                            --
  ;                       |  CHANGED                     |
  ;	                  |            --            --  |
  ;	                  |            | OCT  <value> |  |
  ;	  <.REGISTER_ID>  |  --   --   | HEX  <value> |  |
  ;	                  |  | NOT |   | DEC  <value> |  |
  ;	                  |  --   --   | BIN  <value> |  |
  ;                       |            --            --  |
  ;                       --                            --
  ;
  ;
  ;  (2) - memory location
  ;
  ;                           --                                 --
  ;                           |  CHANGED                          |
  ;       --   --             |  READ                             |
  ;	  | OCT |             |  WRITE                            |
  ;	  | HEX |             |                 --            --  |
  ;	  | DEC |  <address>  |        --   --  | OCT  <value> |  |
  ;	  | BIN |             |  DATA  | NOT |  | HEX  <value> |  |
  ;	  --   --             |        --   --  | DEC  <value> |  |
  ;                           |                 | BIN  >value> |  |
  ;                           |                 --            --  |
  ;                           --                                 --
  ;
  ;  Note.
  ;  If no data format (HEX, OCT, DEC, BIN) is specified, OCTAL is assumed.



  ; ## Command group lists
  ;    - carriage return indicates end-of-command string
  ;    - line feed indicates end-of-command group

trclst:	.ascii	/AFTER/
	.byte	CR
	.ascii	/UNTIL/
	.byte	CR
	.ascii	/BEFORE/
	.byte	CR
	.ascii	/ONLY/
	.byte	CR
	.ascii	/ALWAYS/
	.byte	CR
	.byte	LF		; end of list 1

    ; register names

	.ascii	/.PC/		; PC
	.byte	CR
	.ascii	/.AC/		; AC
	.byte	CR
	.ascii	/.MQ/		; MQ
	.byte	CR
	.ascii	/.MD/		; instruction word
	.byte	CR
	.ascii	/.IF/		; IF
	.byte	CR
	.ascii	/.DF/		; DF
	.byte	CR
	.ascii	/.IO/		; IOflgs (interrupt mask)
	.byte	CR
	.ascii	/.L/		; Link bit
	.byte	CR
	.ascii	/.U/		; User bit
	.byte	CR
	.byte	LF		; end of list 2

    ; access type trace condition

	.ascii	/DATA/
	.byte	CR
	.ascii	/READ/
	.byte	CR
	.ascii	/WRITE/
	.byte	CR
	.ascii	/CHANGED/
	.byte	CR
	.byte	LF		; end of list 3

    ; number base specifiers

	.ascii	/HEX/
	.byte	CR
	.ascii	/DEC/
	.byte	CR
	.ascii	/OCT/
	.byte	CR
	.ascii	/BIN/
	.byte	CR
	.byte	LF		; end of list 4

    ; simulation control command keywords

	.ascii	/TRACE/
	.byte	CR
	.ascii	/REALTIME/
	.byte	CR
	.ascii	/RUN/
	.byte	CR
	.ascii	/STEP/
	.byte	CR
	.ascii	/DISPLAY/
	.byte	CR
	.ascii	/LOAD/
	.byte	CR
	.byte	LF		; end of list 5

    ; the RUN command only allows UNTIL as next keyword

	.ascii	/UNTIL/
	.byte	CR
	.byte	LF		; end of list 6

    ; the STEP command options

	.ascii	/COUNT/
	.byte	CR
	.ascii	/UNTIL/
	.byte	CR
	.byte	LF		; end of list 7

    ; the REGISTER access target only allows CHANGED (besides a value)

	.ascii	/CHANGED/
	.byte	CR
	.byte	LF		; end of list 8

    ; the DATA <value> has optionally the inversion keyword "NOT"

	.ascii	/NOT/
	.byte	CR
	.byte	LF		; end of list 9

    ; the DISPLAY command has optionally the "THRU" keyword

	.ascii	/THRU/
	.byte	CR
	.byte	LF		; end of list 10

    ; the TRACE AFTER/UNTIL/ONLY/BEFORE commands
    ; and RUN UNTIL have optionally the "OCCUR" keyword

	.ascii	/OCCUR/
	.byte	CR
	.byte	LF		; end of list 11

    ; the DISPLAY command has an extra option: "RF08"
    ; which can be followed by "OFF" or "ON" to display the RF08 "registers"

	.ascii	/RF08/
	.byte	CR
	.byte	LF		; end of list 12

	.ascii	/OFF/
	.byte	CR
	.ascii	/ON/
	.byte	CR
	.byte	LF		; end of list 13


	.end